-- ==============================================================
-- FILE : HE_SDRAM.VHD  -  DO NOT EDIT THIS FILE
-- ==============================================================
--
-- This file is part of the Hardware Interface Layer (H.I.L) of
-- your project when using your HERON-FPGA7 module.
--
-- This file must NOT be modified by users of the Module.
--
-- ==============================================================
--
--  Module      : HE_SDRAM
--  Date        : 24/03/2003
--  Author      : R. Williams - HUNT ENGINEERING
--  Description : 256Mbyte SDRAM interface
--
-- ==============================================================
--
--  Ver     Modified By      Date      Changes
--  ---     -----------      ----      -------
--  1.0     R. Williams    24-03-03    First Written
--
-- ==============================================================


library IEEE;
  use IEEE.std_logic_1164.all;
  use IEEE.std_logic_arith.all;
  use IEEE.std_logic_unsigned.all;

-- synopsys translate_off
library UNISIM;
  use UNISIM.vcomponents.all;
-- synopsys translate_on

entity HE_SDRAM is
  port (
    -- External PADS
    SDRAM_A        : out std_logic_vector(12 downto 0);
    SDRAM_BA       : out std_logic_vector(1 downto 0);
    SDRAM_CS       : out std_logic_vector(1 downto 0);
    SDRAM_CKE      : out std_logic_vector(1 downto 0);
    SDRAM_DQM      : out std_logic;
    SDRAM_RAS      : out std_logic;
    SDRAM_CAS      : out std_logic;
    SDRAM_WE       : out std_logic;
    -- Module I/Os 
    OSC            : in  std_logic;
    xSDCLK         : in  std_logic;
    CLKFX_133MHZ   : out std_logic;
    SD_OEN         : out std_logic;
    SD_DIN         : in  std_logic_vector(31 downto 0);
    SD_DOUT        : out std_logic_vector(31 downto 0);
    -- SDRAM Clock
    X_SDCLK        : out  std_logic;
    -- SDRAM Write Port
    SDRAM_WR_DATA  : in  std_logic_vector(31 downto 0);
    SDRAM_WR_ADDR  : in  std_logic_vector(25 downto 0);
    SDRAM_WR_AEN   : in  std_logic;
    SDRAM_WR_READY : in  std_logic;
    SDRAM_WR_BURST : in  std_logic;
    SDRAM_WR_DONE  : out std_logic;
    -- SDRAM Read Port
    SDRAM_RD_DATA  : out std_logic_vector(31 downto 0);
    SDRAM_RD_ADDR  : in  std_logic_vector(25 downto 0);
    SDRAM_RD_AEN   : in  std_logic;
    SDRAM_RD_READY : in  std_logic;
    SDRAM_RD_BURST : in  std_logic;
    SDRAM_RD_DONE  : out std_logic;
    SDRAM_RD_DVAL  : out std_logic
  );
end HE_SDRAM;


architecture RTL of HE_SDRAM is

  component BUFG
    port (
      I : in  std_logic;
      O : out std_logic );
  end component;

  component FD
    port(
      Q   : out std_logic;
      D   : in  std_logic;
      C   : in  std_logic );
  end component;

  component FDC
    port(
      Q   : out std_logic;
      D   : in  std_logic;
      C   : in  std_logic;
      CLR : in  std_logic );
  end component;

  component FDCE
    port(
      Q   : out std_logic;
      D   : in  std_logic;
      C   : in  std_logic;
      CE  : in  std_logic;
      CLR : in  std_logic );
  end component;

  component DCM
    generic (
      -- Required for DLL operation
      DLL_FREQUENCY_MODE    : string  := "LOW";
      DUTY_CYCLE_CORRECTION : boolean := TRUE;
      -- Required for DFS operation
      DFS_FREQUENCY_MODE : string  := "LOW";
      CLKFX_DIVIDE       : integer := 1;
      CLKFX_MULTIPLY     : integer := 2;
      CLK_FEEDBACK       : string  := "NONE";
      -- Required
      STARTUP_WAIT       : boolean := FALSE
    );
    port (
      CLKIN     : in  std_logic;
      CLKFB     : in  std_logic;
      DSSEN     : in  std_logic;
      PSINCDEC  : in  std_logic;
      PSEN      : in  std_logic;
      PSCLK     : in  std_logic;
      RST       : in  std_logic;
      CLK0      : out std_logic;
      CLK90     : out std_logic;
      CLK180    : out std_logic;
      CLK270    : out std_logic;
      CLK2X     : out std_logic;
      CLK2X180  : out std_logic;
      CLKDV     : out std_logic;
      CLKFX     : out std_logic;
      CLKFX180  : out std_logic;
      LOCKED    : out std_logic;
      PSDONE    : out std_logic;
      STATUS    : out std_logic_vector(7 downto 0) );
  end component;

  component HE_SD_CNTR
    port (
      SDCLK    : in  std_logic;
      RESET    : in  std_logic;
      COUNT_EN : in  std_logic;
      ADDR_EN  : in  std_logic;
      ADDR_IN  : in  std_logic_vector(25 downto 0);
      ROW_ADDR : out std_logic_vector(13 downto 0);
      COL_ADDR : out std_logic_vector(9 downto 0);
      BANK     : out std_logic_vector(1 downto 0);
      TC_FN_A  : out std_logic;
      TC_FN_B  : out std_logic;
      TC_FN_C  : out std_logic
    );
  end component;

  -- Note : attributes on instances are supported by all synthesis tools.

  attribute DLL_FREQUENCY_MODE    : string;
  attribute DUTY_CYCLE_CORRECTION : string;
  attribute DFS_FREQUENCY_MODE    : string;
  attribute CLKFX_DIVIDE          : integer;
  attribute CLKFX_MULTIPLY        : integer;
  attribute STARTUP_WAIT          : string;
  attribute CLK_FEEDBACK          : string;

  attribute DLL_FREQUENCY_MODE    of iDFS_DCM : label is "LOW";
  attribute DUTY_CYCLE_CORRECTION of iDFS_DCM : label is "TRUE";
  attribute DFS_FREQUENCY_MODE    of iDFS_DCM : label is "LOW";
  attribute CLKFX_DIVIDE          of iDFS_DCM : label is 3;
  attribute CLKFX_MULTIPLY        of iDFS_DCM : label is 4;
  attribute STARTUP_WAIT          of iDFS_DCM : label is "FALSE";
  attribute CLK_FEEDBACK          of iDFS_DCM : label is "NONE";

  attribute DLL_FREQUENCY_MODE    of iDLL_DCM : label is "LOW"; 
  attribute DUTY_CYCLE_CORRECTION of iDLL_DCM : label is "TRUE"; 
  attribute DFS_FREQUENCY_MODE    of iDLL_DCM : label is "LOW";
  attribute STARTUP_WAIT          of iDLL_DCM : label is "FALSE";

  attribute LOC : string;
  attribute IOB : string;

  -- SDRAM_WE
  attribute LOC of iOFD0 : label is "E22";
  -- SDRAM_CAS
  attribute LOC of iOFD1 : label is "E21";
  -- SDRAM_RAS
  attribute LOC of iOFD2 : label is "C23";
  -- SDRAM_CS(0)
  attribute LOC of iOFD3 : label is "H21";
  -- SDRAM_CS(1)
--attribute LOC of iOFD4 : label is "H20";
  -- SDRAM_DQM
  attribute LOC of iOFD5 : label is "J21";
  -- SDRAM_BA(0)
  attribute LOC of iOFD6 : label is "K21";
  -- SDRAM_BA(1)
  attribute LOC of iOFD7 : label is "A24";

  attribute LOC of iAD0  : label is "F21";
  attribute LOC of iAD1  : label is "F22";
  attribute LOC of iAD2  : label is "G23";
  attribute LOC of iAD3  : label is "G22";
  attribute LOC of iAD4  : label is "B24";
  attribute LOC of iAD5  : label is "B23";
  attribute LOC of iAD6  : label is "D23";
  attribute LOC of iAD7  : label is "D22";
  attribute LOC of iAD8  : label is "D24";
  attribute LOC of iAD9  : label is "C24";
  attribute LOC of iAD10 : label is "B27";
  attribute LOC of iAD11 : label is "A26";
  attribute LOC of iAD12 : label is "A29";

  attribute LOC of iDI0  : label is "J22";
  attribute LOC of iDI1  : label is "B28";
  attribute LOC of iDI2  : label is "F23";
  attribute LOC of iDI3  : label is "H23";
  attribute LOC of iDI4  : label is "D25";
  attribute LOC of iDI5  : label is "E25";
  attribute LOC of iDI6  : label is "G24";
  attribute LOC of iDI7  : label is "A30";

  attribute LOC of iDI8  : label is "B29";
  attribute LOC of iDI9  : label is "F24";
  attribute LOC of iDI10 : label is "H22";
  attribute LOC of iDI11 : label is "D26";
  attribute LOC of iDI12 : label is "E24";
  attribute LOC of iDI13 : label is "G25";
  attribute LOC of iDI14 : label is "A31";
  attribute LOC of iDI15 : label is "D27";

  attribute LOC of iDI16 : label is "C26";
  attribute LOC of iDI17 : label is "K23";
  attribute LOC of iDI18 : label is "B30";
  attribute LOC of iDI19 : label is "C27";
  attribute LOC of iDI20 : label is "J24";
  attribute LOC of iDI21 : label is "B32";
  attribute LOC of iDI22 : label is "E26";
  attribute LOC of iDI23 : label is "H25";

  attribute LOC of iDI24 : label is "K22";
  attribute LOC of iDI25 : label is "B31";
  attribute LOC of iDI26 : label is "C28";
  attribute LOC of iDI27 : label is "J23";
  attribute LOC of iDI28 : label is "C33";
  attribute LOC of iDI29 : label is "F27";
  attribute LOC of iDI30 : label is "H24";
  attribute LOC of iDI31 : label is "F26";

  attribute LOC of iDO0  : label is "J22";
  attribute LOC of iDO1  : label is "B28";
  attribute LOC of iDO2  : label is "F23";
  attribute LOC of iDO3  : label is "H23";
  attribute LOC of iDO4  : label is "D25";
  attribute LOC of iDO5  : label is "E25";
  attribute LOC of iDO6  : label is "G24";
  attribute LOC of iDO7  : label is "A30";

  attribute LOC of iDO8  : label is "B29";
  attribute LOC of iDO9  : label is "F24";
  attribute LOC of iDO10 : label is "H22";
  attribute LOC of iDO11 : label is "D26";
  attribute LOC of iDO12 : label is "E24";
  attribute LOC of iDO13 : label is "G25";
  attribute LOC of iDO14 : label is "A31";
  attribute LOC of iDO15 : label is "D27";

  attribute LOC of iDO16 : label is "C26";
  attribute LOC of iDO17 : label is "K23";
  attribute LOC of iDO18 : label is "B30";
  attribute LOC of iDO19 : label is "C27";
  attribute LOC of iDO20 : label is "J24";
  attribute LOC of iDO21 : label is "B32";
  attribute LOC of iDO22 : label is "E26";
  attribute LOC of iDO23 : label is "H25";

  attribute LOC of iDO24 : label is "K22";
  attribute LOC of iDO25 : label is "B31";
  attribute LOC of iDO26 : label is "C28";
  attribute LOC of iDO27 : label is "J23";
  attribute LOC of iDO28 : label is "C33";
  attribute LOC of iDO29 : label is "F27";
  attribute LOC of iDO30 : label is "H24";
  attribute LOC of iDO31 : label is "F26";

  attribute IOB of iOFD0 : label is "TRUE";
  attribute IOB of iOFD1 : label is "TRUE";
  attribute IOB of iOFD2 : label is "TRUE";
  attribute IOB of iOFD3 : label is "TRUE";
--attribute IOB of iOFD4 : label is "TRUE";
  attribute IOB of iOFD5 : label is "TRUE";
  attribute IOB of iOFD6 : label is "TRUE";
  attribute IOB of iOFD7 : label is "TRUE";

  attribute IOB of iAD0  : label is "TRUE";
  attribute IOB of iAD1  : label is "TRUE";
  attribute IOB of iAD2  : label is "TRUE";
  attribute IOB of iAD3  : label is "TRUE";
  attribute IOB of iAD4  : label is "TRUE";
  attribute IOB of iAD5  : label is "TRUE";
  attribute IOB of iAD6  : label is "TRUE";
  attribute IOB of iAD7  : label is "TRUE";
  attribute IOB of iAD8  : label is "TRUE";
  attribute IOB of iAD9  : label is "TRUE";
  attribute IOB of iAD10 : label is "TRUE";
  attribute IOB of iAD11 : label is "TRUE";
  attribute IOB of iAD12 : label is "TRUE";

  attribute IOB of iDI0  : label is "TRUE";
  attribute IOB of iDI1  : label is "TRUE";
  attribute IOB of iDI2  : label is "TRUE";
  attribute IOB of iDI3  : label is "TRUE";
  attribute IOB of iDI4  : label is "TRUE";
  attribute IOB of iDI5  : label is "TRUE";
  attribute IOB of iDI6  : label is "TRUE";
  attribute IOB of iDI7  : label is "TRUE";
  attribute IOB of iDI8  : label is "TRUE";
  attribute IOB of iDI9  : label is "TRUE";
  attribute IOB of iDI10 : label is "TRUE";
  attribute IOB of iDI11 : label is "TRUE";
  attribute IOB of iDI12 : label is "TRUE";
  attribute IOB of iDI13 : label is "TRUE";
  attribute IOB of iDI14 : label is "TRUE";
  attribute IOB of iDI15 : label is "TRUE";
  attribute IOB of iDI16 : label is "TRUE";
  attribute IOB of iDI17 : label is "TRUE";
  attribute IOB of iDI18 : label is "TRUE";
  attribute IOB of iDI19 : label is "TRUE";
  attribute IOB of iDI20 : label is "TRUE";
  attribute IOB of iDI21 : label is "TRUE";
  attribute IOB of iDI22 : label is "TRUE";
  attribute IOB of iDI23 : label is "TRUE";
  attribute IOB of iDI24 : label is "TRUE";
  attribute IOB of iDI25 : label is "TRUE";
  attribute IOB of iDI26 : label is "TRUE";
  attribute IOB of iDI27 : label is "TRUE";
  attribute IOB of iDI28 : label is "TRUE";
  attribute IOB of iDI29 : label is "TRUE";
  attribute IOB of iDI30 : label is "TRUE";
  attribute IOB of iDI31 : label is "TRUE";
  attribute IOB of iDO0  : label is "TRUE";
  attribute IOB of iDO1  : label is "TRUE";
  attribute IOB of iDO2  : label is "TRUE";
  attribute IOB of iDO3  : label is "TRUE";
  attribute IOB of iDO4  : label is "TRUE";
  attribute IOB of iDO5  : label is "TRUE";
  attribute IOB of iDO6  : label is "TRUE";
  attribute IOB of iDO7  : label is "TRUE";
  attribute IOB of iDO8  : label is "TRUE";
  attribute IOB of iDO9  : label is "TRUE";
  attribute IOB of iDO10 : label is "TRUE";
  attribute IOB of iDO11 : label is "TRUE";
  attribute IOB of iDO12 : label is "TRUE";
  attribute IOB of iDO13 : label is "TRUE";
  attribute IOB of iDO14 : label is "TRUE";
  attribute IOB of iDO15 : label is "TRUE";
  attribute IOB of iDO16 : label is "TRUE";
  attribute IOB of iDO17 : label is "TRUE";
  attribute IOB of iDO18 : label is "TRUE";
  attribute IOB of iDO19 : label is "TRUE";
  attribute IOB of iDO20 : label is "TRUE";
  attribute IOB of iDO21 : label is "TRUE";
  attribute IOB of iDO22 : label is "TRUE";
  attribute IOB of iDO23 : label is "TRUE";
  attribute IOB of iDO24 : label is "TRUE";
  attribute IOB of iDO25 : label is "TRUE";
  attribute IOB of iDO26 : label is "TRUE";
  attribute IOB of iDO27 : label is "TRUE";
  attribute IOB of iDO28 : label is "TRUE";
  attribute IOB of iDO29 : label is "TRUE";
  attribute IOB of iDO30 : label is "TRUE";
  attribute IOB of iDO31 : label is "TRUE";

  attribute NODELAY : string;

  attribute NODELAY of iDI0  : label is "TRUE";
  attribute NODELAY of iDI1  : label is "TRUE";
  attribute NODELAY of iDI2  : label is "TRUE";
  attribute NODELAY of iDI3  : label is "TRUE";
  attribute NODELAY of iDI4  : label is "TRUE";
  attribute NODELAY of iDI5  : label is "TRUE";
  attribute NODELAY of iDI6  : label is "TRUE";
  attribute NODELAY of iDI7  : label is "TRUE";
  attribute NODELAY of iDI8  : label is "TRUE";
  attribute NODELAY of iDI9  : label is "TRUE";
  attribute NODELAY of iDI10 : label is "TRUE";
  attribute NODELAY of iDI11 : label is "TRUE";
  attribute NODELAY of iDI12 : label is "TRUE";
  attribute NODELAY of iDI13 : label is "TRUE";
  attribute NODELAY of iDI14 : label is "TRUE";
  attribute NODELAY of iDI15 : label is "TRUE";
  attribute NODELAY of iDI16 : label is "TRUE";
  attribute NODELAY of iDI17 : label is "TRUE";
  attribute NODELAY of iDI18 : label is "TRUE";
  attribute NODELAY of iDI19 : label is "TRUE";
  attribute NODELAY of iDI20 : label is "TRUE";
  attribute NODELAY of iDI21 : label is "TRUE";
  attribute NODELAY of iDI22 : label is "TRUE";
  attribute NODELAY of iDI23 : label is "TRUE";
  attribute NODELAY of iDI24 : label is "TRUE";
  attribute NODELAY of iDI25 : label is "TRUE";
  attribute NODELAY of iDI26 : label is "TRUE";
  attribute NODELAY of iDI27 : label is "TRUE";
  attribute NODELAY of iDI28 : label is "TRUE";
  attribute NODELAY of iDI29 : label is "TRUE";
  attribute NODELAY of iDI30 : label is "TRUE";
  attribute NODELAY of iDI31 : label is "TRUE";

  constant dfs_dcm_m : integer := 4;
  constant dfs_dcm_d : integer := 3;

  constant dll_dcm_m : integer := 1;
  constant dll_dcm_d : integer := 1;

  type state_type1 is (power_on, do_precharge, set_refresh, do_mr_set, idle);

  signal GND           : std_logic;
  -- Power On State Machine and DLL control signals
  signal POWER_ON_sm   : std_logic_vector(2 downto 0);
  signal POWER_ON_smc  : std_logic_vector(2 downto 0);
  signal LOCK_A        : std_logic;
  signal LOCK_B        : std_logic;
  signal R1            : std_logic;
  signal R2            : std_logic;
  signal R3            : std_logic;
  signal R4            : std_logic;
  signal RST_DLL       : std_logic;
  signal RST_SDRAM     : std_logic;
  signal SDCLK_DCM     : std_logic;
  signal SDCLK         : std_logic;
  signal SDCLK_LOCKED  : std_logic;
  signal SDCLK_LOCKEDd : std_logic;
  signal CLKFX_LOCKED  : std_logic;
  signal CLKFX_LOCKEDd : std_logic;
  -- Refresh signals
  signal REFRESHcount  : std_logic_vector(10 downto 0);
  signal REFRESHtc     : std_logic_vector(10 downto 0);
  signal REFRESHto_do  : std_logic_vector(1 downto 0);
  signal REFRESHdone   : std_logic;
  signal REFRESHen     : std_logic;
  signal REFRESHce     : std_logic;
  signal REFRESHdir    : std_logic;
  signal REFRESHhold   : std_logic;
  signal REFRESHset    : std_logic;
  signal STARTUPcount  : std_logic_vector(4 downto 0);
  signal STARTUPtc     : std_logic_vector(4 downto 0);
  signal STARTUPen     : std_logic;
  -- Startup Control State Machine signals
  signal do_first_pc   : std_logic;
  signal do_mrs        : std_logic;
  signal do_refresh    : std_logic;
  signal proceed       : std_logic;
  signal state1        : state_type1;
  -- Banks A, B, C and D Timing Control signals
  signal BANK_A_sm     : std_logic_vector(2 downto 0);
  signal BANK_A_smc    : std_logic_vector(2 downto 0);
  signal BANK_A_active : std_logic;
  signal BANK_A_rw     : std_logic;
  signal BANK_A_pc     : std_logic;
  signal BANK_A_idle   : std_logic;
  signal BANK_B_sm     : std_logic_vector(2 downto 0);
  signal BANK_B_smc    : std_logic_vector(2 downto 0);
  signal BANK_B_active : std_logic;
  signal BANK_B_rw     : std_logic;
  signal BANK_B_pc     : std_logic;
  signal BANK_B_idle   : std_logic;
  signal BANK_C_sm     : std_logic_vector(2 downto 0);
  signal BANK_C_smc    : std_logic_vector(2 downto 0);
  signal BANK_C_active : std_logic;
  signal BANK_C_rw     : std_logic;
  signal BANK_C_pc     : std_logic;
  signal BANK_C_idle   : std_logic;
  signal BANK_D_sm     : std_logic_vector(2 downto 0);
  signal BANK_D_smc    : std_logic_vector(2 downto 0);
  signal BANK_D_active : std_logic;
  signal BANK_D_rw     : std_logic;
  signal BANK_D_pc     : std_logic;
  signal BANK_D_idle   : std_logic;
  -- Main State Machine and Control Logic signals
  signal BANK_SELECT   : std_logic_vector(1 downto 0);
  signal ACTIVEchoice  : std_logic_vector(1 downto 0);
  -- 1 Port Version
--signal PAGE_SELECT   : std_logic;
  signal READ_BANK     : std_logic_vector(1 downto 0);
  signal WRITE_BANK    : std_logic_vector(1 downto 0);
  -- 4 Port Version
--signal PAGE_SELECT   : std_logic_vector(1 downto 0);
--signal READchoice    : std_logic_vector(1 downto 0);
--signal WRITEchoice   : std_logic_vector(1 downto 0);
--signal LASTread      : std_logic_vector(1 downto 0);
--signal LASTwrite     : std_logic_vector(1 downto 0);
  signal PCchoice      : std_logic_vector(1 downto 0);
  signal SAMEchoice    : std_logic_vector(1 downto 0);
  signal OTHERchoice   : std_logic_vector(1 downto 0);
  signal RD_WRchoice   : std_logic_vector(1 downto 0);
  signal AC_PCchoice   : std_logic_vector(1 downto 0);
  signal LASTactive    : std_logic_vector(1 downto 0);
  signal LASTpc        : std_logic_vector(1 downto 0);
  signal IDLEdone      : std_logic;
  signal MRSdone       : std_logic;
  signal WRITEdone     : std_logic;
  signal READdone      : std_logic;
  signal STOPdone      : std_logic;
  signal PRECHARGEdone : std_logic;
  signal ACTIVEdone    : std_logic;
  signal NOPdone       : std_logic;
  signal NOT_RWdone    : std_logic;
  signal RD_WRdone     : std_logic;
  signal RW_STdone     : std_logic;
  signal AC_STdone     : std_logic;
  signal RW_ST_PCdone  : std_logic;
  -- 4 Port Version
--signal RW_ACCESS_IP  : std_logic_vector(1 downto 0);
  signal do_write      : std_logic;
  signal do_read       : std_logic;
  signal do_same_pc    : std_logic;
  signal do_same_pc_rg : std_logic;
  signal do_stop       : std_logic;
  signal do_other_pc   : std_logic;
  signal do_active     : std_logic;
  signal do_idle       : std_logic;
  signal ACTIVE_A      : std_logic;
  signal ACTIVE_B      : std_logic;
  signal ACTIVE_C      : std_logic;
  signal ACTIVE_D      : std_logic;
  signal ACTIVE_X      : std_logic;
  signal WR_RDY_A      : std_logic;
  signal WR_RDY_B      : std_logic;
  signal WR_RDY_C      : std_logic;
  signal WR_RDY_D      : std_logic;
  -- 4 Port Version
--signal RD_RDY_A      : std_logic;
--signal RD_RDY_B      : std_logic;
--signal RD_RDY_C      : std_logic;
--signal RD_RDY_D      : std_logic;
  signal RD_RDY_AB     : std_logic;
  signal RD_RDY_CD     : std_logic;
  signal DO_SAME_PC_A  : std_logic;
  signal DO_SAME_PC_B  : std_logic;
  signal DO_SAME_PC_C  : std_logic;
  signal DO_SAME_PC_D  : std_logic;
  signal PC_RDY_A      : std_logic;
  signal PC_RDY_B      : std_logic;
  signal PC_RDY_C      : std_logic;
  signal PC_RDY_D      : std_logic;
  signal AC_RDY_A      : std_logic;
  signal AC_RDY_B      : std_logic;
  signal AC_RDY_C      : std_logic;
  signal AC_RDY_D      : std_logic;
  signal S0            : std_logic;
  signal S1            : std_logic;
  signal S2            : std_logic;
  signal S3            : std_logic;
  signal S4            : std_logic;
  signal S5            : std_logic;
  signal S6            : std_logic;
  signal S7            : std_logic;
  signal FN_X          : std_logic;
  signal FN_Y          : std_logic;
  signal FN_Z          : std_logic;
  signal FPC_ACTIVE    : std_logic;
  signal PAGE          : std_logic;
--signal PAGE0_ACTIVE  : std_logic;
--signal PAGE1_ACTIVE  : std_logic;
  signal COMMAND       : std_logic_vector(5 downto 1);
  signal WAITcount     : std_logic_vector(2 downto 0);
  signal WAITce        : std_logic;
  signal refresh_wait  : std_logic;
  -- SDRAM IO signals
  signal DATA_IN       : std_logic_vector(31 downto 0);
  signal DATA          : std_logic_vector(31 downto 0);
  signal ADDR          : std_logic_vector(12 downto 0);
  signal DIN_CE        : std_logic;
  signal DIN_CEd       : std_logic;
  signal DOUT_CE       : std_logic;
  signal ADDR_CE       : std_logic;
  signal AD10_CE       : std_logic;
  signal DRIVE         : std_logic;
  signal DQMc          : std_logic;
  -- Address Logic signals
  signal BANK_A_rd_okc : std_logic;
  signal BANK_B_rd_okc : std_logic;
  signal BANK_C_rd_okc : std_logic;
  signal BANK_D_rd_okc : std_logic;
  signal BANK_A_wr_okc : std_logic;
  signal BANK_B_wr_okc : std_logic;
  signal BANK_C_wr_okc : std_logic;
  signal BANK_D_wr_okc : std_logic;
  signal BANK_A_rd_ok  : std_logic;
  signal BANK_B_rd_ok  : std_logic;
  signal BANK_C_rd_ok  : std_logic;
  signal BANK_D_rd_ok  : std_logic;
  signal BANK_A_wr_ok  : std_logic;
  signal BANK_B_wr_ok  : std_logic;
  signal BANK_C_wr_ok  : std_logic;
  signal BANK_D_wr_ok  : std_logic;
  -- 1 Port Version
  signal BANK_X_choice : std_logic_vector(13 downto 0);
  -- 4 Port Version
--signal BANK_A_choice : std_logic_vector(13 downto 0);
--signal BANK_B_choice : std_logic_vector(13 downto 0);
--signal BANK_C_choice : std_logic_vector(13 downto 0);
--signal BANK_D_choice : std_logic_vector(13 downto 0);
  signal BANK_A_addr   : std_logic_vector(13 downto 0);
  signal BANK_B_addr   : std_logic_vector(13 downto 0);
  signal BANK_C_addr   : std_logic_vector(13 downto 0);
  signal BANK_D_addr   : std_logic_vector(13 downto 0);
  signal ROW_ADDR      : std_logic_vector(12 downto 0);
  signal COL_ADDR      : std_logic_vector(9 downto 0);
  signal RD_COL_ADDR   : std_logic_vector(9 downto 0);
  signal WR_COL_ADDR   : std_logic_vector(9 downto 0);
  -- Read in Progress and Write in progress signals
  signal WRITE_TO_A    : std_logic;
  signal WRITE_TO_B    : std_logic;
  signal WRITE_TO_C    : std_logic;
  signal WRITE_TO_D    : std_logic;
  signal READ_FROM_A   : std_logic;
  signal READ_FROM_B   : std_logic;
  signal READ_FROM_C   : std_logic;
  signal READ_FROM_D   : std_logic;
  signal ALLOW_NEW_RD_A : std_logic;
  signal ALLOW_NEW_RD_B : std_logic;
  signal ALLOW_NEW_RD_C : std_logic;
  signal ALLOW_NEW_RD_D : std_logic;
  signal ALLOW_NEW_RD  : std_logic;
  signal ALLOW_NEW_WR  : std_logic;
  signal RIP_RG        : std_logic;
  signal BANK_X_rip    : std_logic;
  signal BANK_X_wip    : std_logic;
  signal BANK_X_rip_rg : std_logic;
  signal BANK_X_wip_rg : std_logic;
  signal BANK_A_rip    : std_logic;
  signal BANK_B_rip    : std_logic;
  signal BANK_C_rip    : std_logic;
  signal BANK_D_rip    : std_logic;
  signal BANK_A_wip    : std_logic;
  signal BANK_B_wip    : std_logic;
  signal BANK_C_wip    : std_logic;
  signal BANK_D_wip    : std_logic;
  signal BANK_A_rip_rg : std_logic;
  signal BANK_B_rip_rg : std_logic;
  signal BANK_C_rip_rg : std_logic;
  signal BANK_D_rip_rg : std_logic;
  signal BANK_A_wip_rg : std_logic;
  signal BANK_B_wip_rg : std_logic;
  signal BANK_C_wip_rg : std_logic;
  signal BANK_D_wip_rg : std_logic;
  signal WRITEend      : std_logic;
  signal READend       : std_logic;
  signal READend_dl1   : std_logic;
  signal READend_dl2   : std_logic;
  signal READend_dl3   : std_logic;
  signal RDdone_dl1    : std_logic;
  signal RDdone_dl2    : std_logic;
  signal RDdone_dl3    : std_logic;
  -- 1 Port Version
  signal STOP_RD_X_DL  : std_logic;
  signal STOP_WR_X_DL  : std_logic;
  signal NEXT_X_order  : std_logic;
  signal LAST_X_order  : std_logic;
  -- 4 Port Version
--signal STOP_RD_A_DL  : std_logic;
--signal STOP_RD_B_DL  : std_logic;
--signal STOP_RD_C_DL  : std_logic;
--signal STOP_RD_D_DL  : std_logic;
--signal STOP_WR_A_DL  : std_logic;
--signal STOP_WR_B_DL  : std_logic;
--signal STOP_WR_C_DL  : std_logic;
--signal STOP_WR_D_DL  : std_logic;
--signal RDaccess_dl1  : std_logic_vector(1 downto 0);
--signal RDaccess_dl2  : std_logic_vector(1 downto 0);
--signal RDaccess_dl3  : std_logic_vector(1 downto 0);
--signal NEXT_A_order  : std_logic;
--signal NEXT_B_order  : std_logic;
--signal NEXT_C_order  : std_logic;
--signal NEXT_D_order  : std_logic;
--signal LAST_A_order  : std_logic;
--signal LAST_B_order  : std_logic;
--signal LAST_C_order  : std_logic;
--signal LAST_D_order  : std_logic;
--signal P_A_RD_R_ADDR : std_logic_vector(13 downto 0);
--signal P_B_RD_R_ADDR : std_logic_vector(13 downto 0);
--signal P_C_RD_R_ADDR : std_logic_vector(13 downto 0);
--signal P_D_RD_R_ADDR : std_logic_vector(13 downto 0);
--signal P_A_WR_R_ADDR : std_logic_vector(13 downto 0);
--signal P_B_WR_R_ADDR : std_logic_vector(13 downto 0);
--signal P_C_WR_R_ADDR : std_logic_vector(13 downto 0);
--signal P_D_WR_R_ADDR : std_logic_vector(13 downto 0);
--signal P_A_RD_C_ADDR : std_logic_vector(9 downto 0);
--signal P_B_RD_C_ADDR : std_logic_vector(9 downto 0);
--signal P_C_RD_C_ADDR : std_logic_vector(9 downto 0);
--signal P_D_RD_C_ADDR : std_logic_vector(9 downto 0);
--signal P_A_WR_C_ADDR : std_logic_vector(9 downto 0);
--signal P_B_WR_C_ADDR : std_logic_vector(9 downto 0);
--signal P_C_WR_C_ADDR : std_logic_vector(9 downto 0);
--signal P_D_WR_C_ADDR : std_logic_vector(9 downto 0);
--signal P_A_RD_TC_A   : std_logic;
--signal P_B_RD_TC_A   : std_logic;
--signal P_C_RD_TC_A   : std_logic;
--signal P_D_RD_TC_A   : std_logic;
--signal P_A_RD_TC_B   : std_logic;
--signal P_B_RD_TC_B   : std_logic;
--signal P_C_RD_TC_B   : std_logic;
--signal P_D_RD_TC_B   : std_logic;
--signal P_A_RD_TC_C   : std_logic;
--signal P_B_RD_TC_C   : std_logic;
--signal P_C_RD_TC_C   : std_logic;
--signal P_D_RD_TC_C   : std_logic;
--signal P_A_WR_TC_A   : std_logic;
--signal P_B_WR_TC_A   : std_logic;
--signal P_C_WR_TC_A   : std_logic;
--signal P_D_WR_TC_A   : std_logic;
--signal P_A_WR_TC_B   : std_logic;
--signal P_B_WR_TC_B   : std_logic;
--signal P_C_WR_TC_B   : std_logic;
--signal P_D_WR_TC_B   : std_logic;
--signal P_A_WR_TC_C   : std_logic;
--signal P_B_WR_TC_C   : std_logic;
--signal P_C_WR_TC_C   : std_logic;
--signal P_D_WR_TC_C   : std_logic;
  -- Open, Close, Read and Write signals
  signal OPEN_A        : std_logic;
  signal OPEN_B        : std_logic;
  signal OPEN_C        : std_logic;
  signal OPEN_D        : std_logic;
--signal RD_OPEN_OK    : std_logic;
--signal WR_OPEN_OK    : std_logic;
  signal DVALID_RD_X   : std_logic;
  signal DVALID_WR_X   : std_logic;
  signal DVALID_RD_A   : std_logic;
  signal DVALID_RD_B   : std_logic;
  signal DVALID_RD_C   : std_logic;
  signal DVALID_RD_D   : std_logic;
  signal DVALID_WR_A   : std_logic;
  signal DVALID_WR_B   : std_logic;
  signal DVALID_WR_C   : std_logic;
  signal DVALID_WR_D   : std_logic;
  signal CLOSE_A       : std_logic;
  signal CLOSE_B       : std_logic;
  signal CLOSE_C       : std_logic;
  signal CLOSE_D       : std_logic;
  signal START_RD_A    : std_logic;
  signal START_RD_B    : std_logic;
  signal START_RD_C    : std_logic;
  signal START_RD_D    : std_logic;
  signal START_WR_A    : std_logic;
  signal START_WR_B    : std_logic;
  signal START_WR_C    : std_logic;
  signal START_WR_D    : std_logic;
  -- 1 Port Version
  signal STOP_RD_X     : std_logic;
  signal STOP_WR_X     : std_logic;
  signal STOPCOND_RD_X : std_logic;
  signal STOPCOND_WR_X : std_logic;
  signal RD_R_ADDR     : std_logic_vector(13 downto 0);
  signal WR_R_ADDR     : std_logic_vector(13 downto 0);
  signal RD_C_ADDR     : std_logic_vector(9 downto 0);
  signal WR_C_ADDR     : std_logic_vector(9 downto 0);
  signal RD_TC_A       : std_logic;
  signal RD_TC_B       : std_logic;
  signal RD_TC_C       : std_logic;
  signal WR_TC_A       : std_logic;
  signal WR_TC_B       : std_logic;
  signal WR_TC_C       : std_logic;
  -- 4 Port Version
--signal STOP_RD_A     : std_logic;
--signal STOP_RD_B     : std_logic;
--signal STOP_RD_C     : std_logic;
--signal STOP_RD_D     : std_logic;
--signal STOP_WR_A     : std_logic;
--signal STOP_WR_B     : std_logic;
--signal STOP_WR_C     : std_logic;
--signal STOP_WR_D     : std_logic;
--signal STOPCOND_RD_A : std_logic;
--signal STOPCOND_RD_B : std_logic;
--signal STOPCOND_RD_C : std_logic;
--signal STOPCOND_RD_D : std_logic;
--signal STOPCOND_WR_A : std_logic;
--signal STOPCOND_WR_B : std_logic;
--signal STOPCOND_WR_C : std_logic;
--signal STOPCOND_WR_D : std_logic;
--signal DMUX_SEL      : std_logic_vector(1 downto 0);

begin

  GND <= '0';

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  -- Power On Control State-Machine

  CLKFX_LOCKEDd <= CLKFX_LOCKED after 1 ns;
  SDCLK_LOCKEDd <= SDCLK_LOCKED after 1 ns;

--iLOCK0 : FDC port map (LOCK_A, D=>LOCK_C,       C=>OSC, CLR=>GND);

  iLOCK0 : FDC port map (LOCK_A, D=>CLKFX_LOCKEDd, C=>OSC, CLR=>GND);
  iLOCK1 : FDC port map (LOCK_B, D=>SDCLK_LOCKEDd, C=>OSC, CLR=>GND);

--iLOCK2 : FDC port map (LOCK_C, D=>CLKFX_LOCKED, C=>xSDCLK, CLR=>GND);

  POWER_ON_smc(0) <= ((NOT POWER_ON_sm(1)) AND (NOT POWER_ON_sm(0)) AND LOCK_A) OR
                     (     POWER_ON_sm(2)  AND (NOT POWER_ON_sm(1))) OR
                     (     POWER_ON_sm(1)  AND      POWER_ON_sm(0) );

  POWER_ON_smc(1) <= ((NOT POWER_ON_sm(1)) AND      POWER_ON_sm(0) ) OR
                     ((NOT POWER_ON_sm(2)) AND      POWER_ON_sm(1) ) OR
                     (     POWER_ON_sm(2)  AND      POWER_ON_sm(1)  AND (NOT LOCK_B)) OR
                     (     POWER_ON_sm(2)  AND (NOT POWER_ON_sm(1))) OR
                     (     POWER_ON_sm(1)  AND      POWER_ON_sm(0) );

  POWER_ON_smc(2) <= ((NOT POWER_ON_sm(2)) AND      POWER_ON_sm(1)  AND (NOT LOCK_B)) OR
                     (     POWER_ON_sm(2));

  iPOFD0 : FDC port map (POWER_ON_sm(0), D=>POWER_ON_smc(0), C=>OSC, CLR=>GND);
  iPOFD1 : FDC port map (POWER_ON_sm(1), D=>POWER_ON_smc(1), C=>OSC, CLR=>GND);
  iPOFD2 : FDC port map (POWER_ON_sm(2), D=>POWER_ON_smc(2), C=>OSC, CLR=>GND);

  R1 <= (NOT POWER_ON_sm(2)) AND (NOT POWER_ON_sm(1)) AND POWER_ON_sm(0);

  iRST0 : FDC port map (R2, D=>R1, C=>OSC, CLR=>GND);
  iRST1 : FDC port map (R3, D=>R2, C=>OSC, CLR=>GND);
  iRST2 : FDC port map (R4, D=>R3, C=>OSC, CLR=>GND);

  RST_DLL <= R2 OR R3 OR R4;

--RST_DLL   <= (NOT POWER_ON_sm(2)) AND (NOT POWER_ON_sm(1)) AND      POWER_ON_sm(0);
  RST_SDRAM <= (NOT POWER_ON_sm(2)) OR  (NOT POWER_ON_sm(1)) OR  (NOT POWER_ON_sm(0));

  -- DCM Instantiation
  iDFS_DCM : DCM
    generic map (
      -- Required for DLL operation
      DLL_FREQUENCY_MODE    => "LOW",
      DUTY_CYCLE_CORRECTION => TRUE,
      -- Required for DFS operation
      DFS_FREQUENCY_MODE    => "LOW",
      CLKFX_DIVIDE          => dfs_dcm_d,
      CLKFX_MULTIPLY        => dfs_dcm_m,
      CLK_FEEDBACK          => "NONE",
      -- Required
      STARTUP_WAIT          => FALSE )
    port map (
      CLKIN    => OSC,
      CLKFB    => GND,
      DSSEN    => GND,
      PSINCDEC => GND,
      PSEN     => GND,
      PSCLK    => GND,
      RST      => GND,
      CLKFX    => CLKFX_133MHZ,
      CLKFX180 => open,
      LOCKED   => CLKFX_LOCKED );

  iDLL_DCM : DCM
    generic map (
      -- Required for DLL operation
      DLL_FREQUENCY_MODE    => "LOW",
      DUTY_CYCLE_CORRECTION => TRUE,
      -- Required for DFS operation
      DFS_FREQUENCY_MODE    => "LOW",
      CLK_FEEDBACK          => "1X",
      -- Required
      STARTUP_WAIT          => FALSE )
    port map (
      CLKIN    => xSDCLK,
      CLKFB    => SDCLK,
      DSSEN    => GND,
      PSINCDEC => GND,
      PSEN     => GND,
      PSCLK    => GND,
      RST      => RST_DLL,
      CLK0     => SDCLK_DCM,
      LOCKED   => SDCLK_LOCKED );

  iSDCLK :  BUFG port map (I=>SDCLK_DCM, O=>SDCLK);

  X_SDCLK <= SDCLK;

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  -- SDRAM Refresh
  REFRESHtc <= "10000010000";

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      REFRESHcount <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if (REFRESHcount = REFRESHtc) then
        REFRESHcount <= "00000000000";
      else
        REFRESHcount <= REFRESHcount + 1;
      end if;
    end if;
  end process;

--STARTUPtc <= "11010";
  STARTUPtc <= "00001";

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      STARTUPcount <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if REFRESHen='1' and STARTUPen='0' then
        STARTUPcount <= STARTUPcount + 1;
      end if;
    end if;
  end process;

  STARTUPen  <= '1' when STARTUPcount=STARTUPtc else '0';

  REFRESHen  <= '1' when REFRESHcount=REFRESHtc else '0';

  REFRESHce  <= ((REFRESHen XOR REFRESHdone) AND (NOT REFRESHhold)) OR REFRESHset;

  REFRESHdir <= REFRESHen;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      REFRESHto_do <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if REFRESHce='1' then
        if REFRESHset='1' then
          REFRESHto_do <= (others=>'1');
        else
          if REFRESHdir='1' then
            REFRESHto_do <= REFRESHto_do + 1;
          else
            REFRESHto_do <= REFRESHto_do - 1;
          end if;
        end if;
      end if;
    end if;
  end process;

  do_refresh  <= REFRESHto_do(0) OR REFRESHto_do(1);

  -- Startup Control State Machine
  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      state1 <= power_on;
    elsif rising_edge(SDCLK) then
      case state1 is
        when power_on     => if    STARTUPen='1'            then state1 <= do_precharge;
                             else                                state1 <= power_on;
                             end if;
        when do_precharge => if    PRECHARGEdone='1'        then state1 <= set_refresh;
                             else                                state1 <= do_precharge;
                             end if;
        when set_refresh  =>                                     state1 <= do_mr_set;
        when do_mr_set    => if    MRSdone='1'              then state1 <= idle;
                             else                                state1 <= do_mr_set;
                             end if;
        when idle         =>                                     state1 <= idle;
      end case;
    end if;
  end process;

  process(state1)
  begin
    case state1 is
      when power_on     => do_first_pc <= '0'; do_mrs <= '0'; REFRESHhold <= '1'; REFRESHset <= '0';
                           proceed <= '0';
      when do_precharge => do_first_pc <= '1'; do_mrs <= '0'; REFRESHhold <= '1'; REFRESHset <= '0';
                           proceed <= '0';
      when set_refresh  => do_first_pc <= '0'; do_mrs <= '0'; REFRESHhold <= '0'; REFRESHset <= '1';
                           proceed <= '0';
      when do_mr_set    => do_first_pc <= '0'; do_mrs <= '1'; REFRESHhold <= '0'; REFRESHset <= '0';
                           proceed <= '0';
      when idle         => do_first_pc <= '0'; do_mrs <= '0'; REFRESHhold <= '0'; REFRESHset <= '0';
                           proceed <= '1';
    end case;
  end process;

  ---------------------------------------------------------------------------------------
  ---------------------------------------------------------------------------------------

  -- Banks A, B, C and D Timing Control

  -- Bank A Control State-Machine
  BANK_A_smc(0) <= (((NOT BANK_A_sm(2)) AND (NOT BANK_A_sm(0)) AND ACTIVE_A) OR
                    (     BANK_A_sm(2)  AND (NOT BANK_A_sm(0))) OR
                    (     BANK_A_sm(2)  AND      BANK_A_sm(1) )) after 1 ns;

  BANK_A_smc(1) <= ((BANK_A_sm(2) AND BANK_A_sm(0)) OR
                    (BANK_A_sm(2) AND BANK_A_sm(1))) after 1 ns;

  BANK_A_smc(2) <= (((NOT BANK_A_sm(1)) AND BANK_A_sm(0)) OR
                    (     BANK_A_sm(2)  AND 
                    (NOT (PRECHARGEdone AND (NOT PCchoice(1)) AND (NOT PCchoice(0)))))) after 1 ns;

  iBAFD0 : FDC port map (BANK_A_sm(0), D=>BANK_A_smc(0), C=>SDCLK, CLR=>RST_SDRAM);
  iBAFD1 : FDC port map (BANK_A_sm(1), D=>BANK_A_smc(1), C=>SDCLK, CLR=>RST_SDRAM);
  iBAFD2 : FDC port map (BANK_A_sm(2), D=>BANK_A_smc(2), C=>SDCLK, CLR=>RST_SDRAM);

  BANK_A_active <= ((NOT BANK_A_sm(2)) AND (NOT BANK_A_sm(1)) AND (NOT BANK_A_sm(0)));
  BANK_A_idle   <= ((NOT BANK_A_sm(2)) AND (NOT BANK_A_sm(1)) AND (NOT BANK_A_sm(0))) OR
                   ((NOT BANK_A_sm(2)) AND      BANK_A_sm(1)  AND      BANK_A_sm(0) );
  BANK_A_pc     <= (     BANK_A_sm(2)  AND      BANK_A_sm(1)  AND      BANK_A_sm(0) );

  BANK_A_rw <= BANK_A_sm(2) AND (NOT (PRECHARGEdone AND (NOT PCchoice(1)) AND (NOT PCchoice(0))));

  -- Bank B Control State-Machine
  BANK_B_smc(0) <= (((NOT BANK_B_sm(2)) AND (NOT BANK_B_sm(0)) AND ACTIVE_B) OR
                    (     BANK_B_sm(2)  AND (NOT BANK_B_sm(0))) OR
                    (     BANK_B_sm(2)  AND      BANK_B_sm(1) )) after 1 ns;

  BANK_B_smc(1) <= ((BANK_B_sm(2) AND BANK_B_sm(0)) OR
                    (BANK_B_sm(2) AND BANK_B_sm(1))) after 1 ns;

  BANK_B_smc(2) <= (((NOT BANK_B_sm(1)) AND BANK_B_sm(0)) OR
                    (     BANK_B_sm(2)  AND 
                    (NOT (PRECHARGEdone AND (NOT PCchoice(1)) AND PCchoice(0))))) after 1 ns;

  iBBFD0 : FDC port map (BANK_B_sm(0), D=>BANK_B_smc(0), C=>SDCLK, CLR=>RST_SDRAM);
  iBBFD1 : FDC port map (BANK_B_sm(1), D=>BANK_B_smc(1), C=>SDCLK, CLR=>RST_SDRAM);
  iBBFD2 : FDC port map (BANK_B_sm(2), D=>BANK_B_smc(2), C=>SDCLK, CLR=>RST_SDRAM);

  BANK_B_active <= ((NOT BANK_B_sm(2)) AND (NOT BANK_B_sm(1)) AND (NOT BANK_B_sm(0)));
  BANK_B_idle   <= ((NOT BANK_B_sm(2)) AND (NOT BANK_B_sm(1)) AND (NOT BANK_B_sm(0))) OR
                   ((NOT BANK_B_sm(2)) AND      BANK_B_sm(1)  AND      BANK_B_sm(0) );
  BANK_B_pc     <= (     BANK_B_sm(2)  AND      BANK_B_sm(1)  AND      BANK_B_sm(0) );

  BANK_B_rw <= BANK_B_sm(2) AND (NOT (PRECHARGEdone AND (NOT PCchoice(1)) AND PCchoice(0)));

  -- Bank C Control State-Machine
  BANK_C_smc(0) <= (((NOT BANK_C_sm(2)) AND (NOT BANK_C_sm(0)) AND ACTIVE_C) OR
                    (     BANK_C_sm(2)  AND (NOT BANK_C_sm(0))) OR
                    (     BANK_C_sm(2)  AND      BANK_C_sm(1) )) after 1 ns;

  BANK_C_smc(1) <= ((BANK_C_sm(2) AND BANK_C_sm(0)) OR
                    (BANK_C_sm(2) AND BANK_C_sm(1))) after 1 ns;

  BANK_C_smc(2) <= (((NOT BANK_C_sm(1)) AND BANK_C_sm(0)) OR
                    (     BANK_C_sm(2)  AND 
                    (NOT (PRECHARGEdone AND PCchoice(1) AND (NOT PCchoice(0)))))) after 1 ns;

  iBCFD0 : FDC port map (BANK_C_sm(0), D=>BANK_C_smc(0), C=>SDCLK, CLR=>RST_SDRAM);
  iBCFD1 : FDC port map (BANK_C_sm(1), D=>BANK_C_smc(1), C=>SDCLK, CLR=>RST_SDRAM);
  iBCFD2 : FDC port map (BANK_C_sm(2), D=>BANK_C_smc(2), C=>SDCLK, CLR=>RST_SDRAM);

  BANK_C_active <= ((NOT BANK_C_sm(2)) AND (NOT BANK_C_sm(1)) AND (NOT BANK_C_sm(0)));
  BANK_C_idle   <= ((NOT BANK_C_sm(2)) AND (NOT BANK_C_sm(1)) AND (NOT BANK_C_sm(0))) OR
                   ((NOT BANK_C_sm(2)) AND      BANK_C_sm(1)  AND      BANK_C_sm(0) );
  BANK_C_pc     <= (     BANK_C_sm(2)  AND      BANK_C_sm(1)  AND      BANK_C_sm(0) );

  BANK_C_rw <= BANK_C_sm(2) AND (NOT (PRECHARGEdone AND PCchoice(1) AND (NOT PCchoice(0))));

  -- Bank D Control State-Machine
  BANK_D_smc(0) <= (((NOT BANK_D_sm(2)) AND (NOT BANK_D_sm(0)) AND ACTIVE_D) OR
                    (     BANK_D_sm(2)  AND (NOT BANK_D_sm(0))) OR
                    (     BANK_D_sm(2)  AND      BANK_D_sm(1) )) after 1 ns;

  BANK_D_smc(1) <= ((BANK_D_sm(2) AND BANK_D_sm(0)) OR
                    (BANK_D_sm(2) AND BANK_D_sm(1))) after 1 ns;

  BANK_D_smc(2) <= (((NOT BANK_D_sm(1)) AND BANK_D_sm(0)) OR
                    (     BANK_D_sm(2)  AND 
                    (NOT (PRECHARGEdone AND PCchoice(1) AND PCchoice(0))))) after 1 ns;

  iBDFD0 : FDC port map (BANK_D_sm(0), D=>BANK_D_smc(0), C=>SDCLK, CLR=>RST_SDRAM);
  iBDFD1 : FDC port map (BANK_D_sm(1), D=>BANK_D_smc(1), C=>SDCLK, CLR=>RST_SDRAM);
  iBDFD2 : FDC port map (BANK_D_sm(2), D=>BANK_D_smc(2), C=>SDCLK, CLR=>RST_SDRAM);

  BANK_D_active <= ((NOT BANK_D_sm(2)) AND (NOT BANK_D_sm(1)) AND (NOT BANK_D_sm(0)));
  BANK_D_idle   <= ((NOT BANK_D_sm(2)) AND (NOT BANK_D_sm(1)) AND (NOT BANK_D_sm(0))) OR
                   ((NOT BANK_D_sm(2)) AND      BANK_D_sm(1)  AND      BANK_D_sm(0) );
  BANK_D_pc     <= (     BANK_D_sm(2)  AND      BANK_D_sm(1)  AND      BANK_D_sm(0) );

  BANK_D_rw <= BANK_D_sm(2) AND (NOT (PRECHARGEdone AND PCchoice(1) AND PCchoice(0)));

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

--when idle      => COMMAND(4 downto 1) <= "1111";
--when nop       => COMMAND(4 downto 1) <= "0111";
--when stop      => COMMAND(4 downto 1) <= "0110";
--when read      => COMMAND(4 downto 1) <= "0101";
--when write     => COMMAND(4 downto 1) <= "0100";
--when active    => COMMAND(4 downto 1) <= "0011";
--when precharge => COMMAND(4 downto 1) <= "0010";
--when refresh   => COMMAND(4 downto 1) <= "0001";
--when nop_wait  => COMMAND(4 downto 1) <= "1111";
--when mrs       => COMMAND(4 downto 1) <= "0000";

  -- The current read or write access ends when:
  --   A new read command is issued,
  --   A new write command is issued,
  --   A precharge command is issued (to the 'in-progress' bank!),
  --   A stop command is issued.

  -- Main State Machine and Control Logic
  AC_PCchoice(0) <= ((NOT S2) AND (NOT S1) AND ACTIVEchoice(0)) OR
                    ((    S2   OR      S1) AND PCchoice(0));
  AC_PCchoice(1) <= ((NOT S2) AND (NOT S1) AND ACTIVEchoice(1)) OR
                    ((    S2   OR      S1) AND PCchoice(1));

  -- 1 Port Version
  RD_WRchoice(0) <= ((NOT S4) AND READ_BANK(0)) OR
                    ((    S4) AND WRITE_BANK(0));
  RD_WRchoice(1) <= ((NOT S4) AND READ_BANK(1)) OR
                    ((    S4) AND WRITE_BANK(1));

  -- Monolithic SDRAM version
  PAGE <= WR_R_ADDR(0) when S4='1' else RD_R_ADDR(0);

  -- DDP SDRAM version
--PAGE_SELECT <= S4 OR (BANK_X_wip_rg AND (NOT S3));

--PAGE <= ((NOT PAGE_SELECT) AND RD_R_ADDR(13)) OR
--        ((    PAGE_SELECT) AND WR_R_ADDR(13));

  BANK_SELECT(0) <= (((NOT S4) AND (NOT S3) AND AC_PCchoice(0)) OR
                     ((    S4   OR      S3) AND RD_WRchoice(0))) after 1 ns;
  BANK_SELECT(1) <= (((NOT S4) AND (NOT S3) AND AC_PCchoice(1)) OR
                     ((    S4   OR      S3) AND RD_WRchoice(1))) after 1 ns;

  -- 4 Port Version
--RD_WRchoice(0) <= ((NOT S4) AND READchoice(0)) OR
--                  ((    S4) AND WRITEchoice(0));
--RD_WRchoice(1) <= ((NOT S4) AND READchoice(1)) OR
--                  ((    S4) AND WRITEchoice(1));

--PAGE_SELECT(0) <= ((NOT S4) AND (NOT S3) AND RW_ACCESS_IP(0)) OR
--                  ((    S4   OR      S3) AND RD_WRchoice(0));
--PAGE_SELECT(1) <= ((NOT S4) AND (NOT S3) AND RW_ACCESS_IP(1)) OR
--                  ((    S4   OR      S3) AND RD_WRchoice(1));

--process (PAGE_SELECT, BANK_A_addr(13), BANK_B_addr(13), BANK_C_addr(13), BANK_D_addr(13))
--begin
--  case PAGE_SELECT is
--    when "00"   => PAGE <= BANK_A_addr(13);
--    when "01"   => PAGE <= BANK_B_addr(13);
--    when "10"   => PAGE <= BANK_C_addr(13);
--    when "11"   => PAGE <= BANK_D_addr(13);
--    when others => PAGE <= '0';
--  end case;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    RW_ACCESS_IP <= (others=>'0');
--  elsif rising_edge(SDCLK) then
--    if RD_WRdone='1' then
--      RW_ACCESS_IP <= BANK_SELECT;
--    end if;
--  end if;
--end process;

  -- Decision logic for performing a WRITE command
  do_write <= WR_RDY_A OR WR_RDY_B OR WR_RDY_C OR WR_RDY_D;

  WR_RDY_A <= START_WR_A AND BANK_A_rw AND ALLOW_NEW_WR;
  WR_RDY_B <= START_WR_B AND BANK_B_rw AND ALLOW_NEW_WR;
  WR_RDY_C <= START_WR_C AND BANK_C_rw AND ALLOW_NEW_WR;
  WR_RDY_D <= START_WR_D AND BANK_D_rw AND ALLOW_NEW_WR;

  -- 4 Port Version
--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    WRITEchoice <= (others=>'0');
--  elsif rising_edge(SDCLK) then
--    WRITEchoice <= (others=>'0');
--    if (LASTwrite(1)='0' and LASTwrite(0)='0' and WR_RDY_B='1') or
--       (LASTwrite(1)='1' and LASTwrite(0)='0' and WR_RDY_D='1') or
--       (LASTwrite(0)='1' and WR_RDY_A='0'     and WR_RDY_B='1' and WR_RDY_C='0') or
--       (LASTwrite(1)='0' and WR_RDY_C='0'     and WR_RDY_D='1') or
--       (LASTwrite(1)='1' and WR_RDY_A='0'     and WR_RDY_B='1') or
--       (WR_RDY_A='0'     and WR_RDY_C='0'     and WR_RDY_D='1') then
--      WRITEchoice(0) <= '1';
--    end if;
--    if (LASTwrite(1)='0' and LASTwrite(0)='1' and WR_RDY_C='1') or
--       (LASTwrite(1)='0' and LASTwrite(0)='1' and WR_RDY_D='1') or
--       (LASTwrite(1)='1' and LASTwrite(0)='0' and WR_RDY_D='1') or
--       (LASTwrite(1)='1' and WR_RDY_A='0'     and WR_RDY_B='0' and WR_RDY_C='1') or
--       (LASTwrite(1)='1' and WR_RDY_A='0'     and WR_RDY_B='0' and WR_RDY_D='1') or
--       (LASTwrite(1)='0' and WR_RDY_B='0'     and WR_RDY_C='1') or
--       (LASTwrite(1)='0' and WR_RDY_B='0'     and WR_RDY_D='1') then
--      WRITEchoice(1) <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    LASTwrite <= (others=>'1');
--  elsif rising_edge(SDCLK) then
--    if S4='1' then
--      LASTwrite <= WRITEchoice;
--    end if;
--  end if;
--end process;

  -- Decision logic for performing a READ command
  do_read  <= (RD_RDY_AB OR RD_RDY_CD) AND ALLOW_NEW_RD;

  RD_RDY_AB <= (START_RD_A AND BANK_A_rw) OR (START_RD_B AND BANK_B_rw);
  RD_RDY_CD <= (START_RD_C AND BANK_C_rw) OR (START_RD_D AND BANK_D_rw);

  -- 4 Port Version
--RD_RDY_A <= START_RD_A AND BANK_A_rw;
--RD_RDY_B <= START_RD_B AND BANK_B_rw;
--RD_RDY_C <= START_RD_C AND BANK_C_rw;
--RD_RDY_D <= START_RD_D AND BANK_D_rw;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    READchoice <= (others=>'0');
--  elsif rising_edge(SDCLK) then
--    READchoice <= (others=>'0');
--    if (LASTread(1)='0' and LASTread(0)='0' and RD_RDY_B='1') or
--       (LASTread(1)='1' and LASTread(0)='0' and RD_RDY_D='1') or
--       (LASTread(0)='1' and RD_RDY_A='0'    and RD_RDY_B='1' and RD_RDY_C='0') or
--       (LASTread(1)='0' and RD_RDY_C='0'    and RD_RDY_D='1') or
--       (LASTread(1)='1' and RD_RDY_A='0'    and RD_RDY_B='1') or
--       (RD_RDY_A='0'    and RD_RDY_C='0'    and RD_RDY_D='1') then
--      READchoice(0) <= '1';
--    end if;
--    if (LASTread(1)='0' and LASTread(0)='1' and RD_RDY_C='1') or
--       (LASTread(1)='0' and LASTread(0)='1' and RD_RDY_D='1') or
--       (LASTread(1)='1' and LASTread(0)='0' and RD_RDY_D='1') or
--       (LASTread(1)='1' and RD_RDY_A='0'    and RD_RDY_B='0' and RD_RDY_C='1') or
--       (LASTread(1)='1' and RD_RDY_A='0'    and RD_RDY_B='0' and RD_RDY_D='1') or
--       (LASTread(1)='0' and RD_RDY_B='0'    and RD_RDY_C='1') or
--       (LASTread(1)='0' and RD_RDY_B='0'    and RD_RDY_D='1') then
--      READchoice(1) <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    LASTread <= (others=>'1');
--  elsif rising_edge(SDCLK) then
--    if READdone='1' then
--      LASTread <= READchoice;
--    end if;
--  end if;
--end process;

  -- Decision logic for performing a PRECHARGE command
  DO_SAME_PC_A <= CLOSE_A AND BANK_A_pc AND (BANK_A_wip OR BANK_A_rip);
  DO_SAME_PC_B <= CLOSE_B AND BANK_B_pc AND (BANK_B_wip OR BANK_B_rip);
  DO_SAME_PC_C <= CLOSE_C AND BANK_C_pc AND (BANK_C_wip OR BANK_C_rip);
  DO_SAME_PC_D <= CLOSE_D AND BANK_D_pc AND (BANK_D_wip OR BANK_D_rip);

  do_same_pc <= DO_SAME_PC_A OR DO_SAME_PC_B OR DO_SAME_PC_C OR DO_SAME_PC_D;

  PC_RDY_A <= CLOSE_A AND BANK_A_pc;
  PC_RDY_B <= CLOSE_B AND BANK_B_pc;
  PC_RDY_C <= CLOSE_C AND BANK_C_pc;
  PC_RDY_D <= CLOSE_D AND BANK_D_pc;

  do_other_pc <= PC_RDY_A OR PC_RDY_B OR PC_RDY_C OR PC_RDY_D;

  PCchoice <= SAMEchoice when do_same_pc_rg='1' else OTHERchoice;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      SAMEchoice <= (others=>'0');
    elsif rising_edge(SDCLK) then
      SAMEchoice <= (others=>'0');
      if BANK_B_wip='1' or BANK_B_rip='1' or BANK_D_wip='1' or BANK_D_rip='1' then
        SAMEchoice(0) <= '1';
      end if;
      if BANK_C_wip='1' or BANK_C_rip='1' or BANK_D_wip='1' or BANK_D_rip='1' then
        SAMEchoice(1) <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      OTHERchoice <= (others=>'0');
    elsif rising_edge(SDCLK) then
      OTHERchoice <= (others=>'0');
      if (LASTpc(1)='0' and LASTpc(0)='0' and PC_RDY_B='1') or
         (LASTpc(1)='1' and LASTpc(0)='0' and PC_RDY_D='1') or
         (LASTpc(0)='1' and PC_RDY_A='0'  and PC_RDY_B='1' and PC_RDY_C='0') or
         (LASTpc(1)='0' and PC_RDY_C='0'  and PC_RDY_D='1') or
         (LASTpc(1)='1' and PC_RDY_A='0'  and PC_RDY_B='1') or
         (PC_RDY_A='0'  and PC_RDY_C='0'  and PC_RDY_D='1') then
        OTHERchoice(0) <= '1';
      end if;
      if (LASTpc(1)='0' and LASTpc(0)='1' and PC_RDY_C='1') or
         (LASTpc(1)='0' and LASTpc(0)='1' and PC_RDY_D='1') or
         (LASTpc(1)='1' and LASTpc(0)='0' and PC_RDY_D='1') or
         (LASTpc(1)='1' and PC_RDY_A='0'  and PC_RDY_B='0' and PC_RDY_C='1') or
         (LASTpc(1)='1' and PC_RDY_A='0'  and PC_RDY_B='0' and PC_RDY_D='1') or
         (LASTpc(1)='0' and PC_RDY_B='0'  and PC_RDY_C='1') or
         (LASTpc(1)='0' and PC_RDY_B='0'  and PC_RDY_D='1') then
        OTHERchoice(1) <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      LASTpc <= (others=>'1');
    elsif rising_edge(SDCLK) then
      if PRECHARGEdone='1' then
        LASTpc <= PCchoice;
      end if;
    end if;
  end process;

  -- Decision logic for performing a STOP command
  -- 1 Port Version
  do_stop <= STOP_WR_X OR STOP_RD_X;

  -- 4 Port Version
--do_stop <= STOP_WR_A OR STOP_WR_B OR STOP_WR_C OR STOP_WR_D OR
--           STOP_RD_A OR STOP_RD_B OR STOP_RD_C OR STOP_RD_D;

  -- Decision logic for performing a ROW ACTIVE command
  AC_RDY_A <= (OPEN_A AND BANK_A_active AND (NOT do_refresh)) OR FPC_ACTIVE;
  AC_RDY_B <= (OPEN_B AND BANK_B_active AND (NOT do_refresh));
  AC_RDY_C <= (OPEN_C AND BANK_C_active AND (NOT do_refresh));
  AC_RDY_D <= (OPEN_D AND BANK_D_active AND (NOT do_refresh));

  do_active <= AC_RDY_A OR AC_RDY_B OR AC_RDY_C OR AC_RDY_D;

  FPC_ACTIVE <= do_first_pc AND (NOT (S5 AND S1));

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      ACTIVEchoice <= (others=>'0');
    elsif rising_edge(SDCLK) then
      ACTIVEchoice <= (others=>'0');
      if (LASTactive(1)='0' and LASTactive(0)='0' and AC_RDY_B='1') or
         (LASTactive(1)='1' and LASTactive(0)='0' and AC_RDY_D='1') or
         (LASTactive(0)='1' and AC_RDY_A='0'      and AC_RDY_B='1' and AC_RDY_C='0') or
         (LASTactive(1)='0' and AC_RDY_C='0'      and AC_RDY_D='1') or
         (LASTactive(1)='1' and AC_RDY_A='0'      and AC_RDY_B='1') or
         (AC_RDY_A='0'      and AC_RDY_C='0'      and AC_RDY_D='1') then
        ACTIVEchoice(0) <= '1';
      end if;
      if (LASTactive(1)='0' and LASTactive(0)='1' and AC_RDY_C='1') or
         (LASTactive(1)='0' and LASTactive(0)='1' and AC_RDY_D='1') or
         (LASTactive(1)='1' and LASTactive(0)='0' and AC_RDY_D='1') or
         (LASTactive(1)='1' and AC_RDY_A='0'      and AC_RDY_B='0' and AC_RDY_C='1') or
         (LASTactive(1)='1' and AC_RDY_A='0'      and AC_RDY_B='0' and AC_RDY_D='1') or
         (LASTactive(1)='0' and AC_RDY_B='0'      and AC_RDY_C='1') or
         (LASTactive(1)='0' and AC_RDY_B='0'      and AC_RDY_D='1') then
        ACTIVEchoice(1) <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      LASTactive <= (others=>'1');
    elsif rising_edge(SDCLK) then
      if (NOT_RWdone='1' and ACTIVEdone='1') then
        LASTactive <= ACTIVEchoice;
      end if;
    end if;
  end process;

  -- Decision logic for returning to the IDLE state
  do_idle <= BANK_A_idle AND BANK_B_idle AND BANK_C_idle AND BANK_D_idle;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      WAITcount <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if WAITce='1' then
        if (WAITcount = "110") then
          WAITcount <= "000";
        else
          WAITcount <= WAITcount + 1;
        end if;
      end if;
    end if;
  end process;

  refresh_wait <= '0' when WAITcount = "110" else '1';

  REFRESHdone <= ((    S7) AND (    S6) AND (NOT S5));
  WAITce      <= ((    S7) AND (NOT S6) AND (NOT S5));
  MRSdone     <= ((NOT S7) AND (    S6) AND (NOT S5));
  IDLEdone    <= ((NOT S7) AND (NOT S6) AND (NOT S5));

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      S7 <= '0';
    elsif rising_edge(SDCLK) then
      S7 <= '0';
      if (IDLEdone='1' and do_first_pc='0' and do_refresh='1') or
         (S7='1' and S6='1') or
         (S7='1' and S6='0' and refresh_wait='1') then
        S7 <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      S6 <= '0';
    elsif rising_edge(SDCLK) then
      S6 <= '0';
      if IDLEdone='1' and do_first_pc='0' and (do_refresh='1' or do_mrs='1') then
        S6 <= '1';
      end if;
    end if;
  end process;

  FN_X <= (NOT do_idle) OR (NOT (NOT_RWdone AND NOPdone));

  FN_Y <= IDLEdone AND (do_first_pc OR ((NOT do_refresh) AND (NOT do_mrs)));

  FN_Z <= IDLEdone AND (NOT do_first_pc) AND (NOT do_refresh) AND (NOT do_mrs);

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      S5 <= '0';
    elsif rising_edge(SDCLK) then
      S5 <= '0';
      if (S5='1' and do_active='1') or
         (S5='1' and FN_X='1') or
         (do_active='1' and FN_Y='1') then
        S5 <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      S4 <= '0';
    elsif rising_edge(SDCLK) then
      S4 <= '0';
      if (S5='1' and (S4='1' or S3='0') and do_write='1') then
        S4 <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      S3 <= '0';
    elsif rising_edge(SDCLK) then
      S3 <= '0';
      if (S5='1' and (S4='1' or S3='0') and do_read='1') then
        S3 <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      S2 <= '0';
    elsif rising_edge(SDCLK) then
      S2 <= '0';
      if (S5='1' and (NOT_RWdone='0' or STOPdone='0') and do_stop='1') then
        S2 <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      S1 <= '0';
    elsif rising_edge(SDCLK) then
      S1 <= '0';
      if (S5='1' and RW_ST_PCdone='0' and do_other_pc='1') or FPC_ACTIVE='1' then
        S1 <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      S0 <= '0';
    elsif rising_edge(SDCLK) then
      S0 <= '0';
      if (S5='1' and AC_STdone='0' and do_active='1') or
         (do_active='1' and FN_Z='1') then
        S0 <= '1';
      end if;
    end if;
  end process;

  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      do_same_pc_rg <= do_same_pc;
    end if;
  end process;

  -- Monolithic SDRAM version
  COMMAND(5) <= '0';
  COMMAND(4) <= ((NOT S6) AND (NOT S5)) after 1 ns;

  -- DDP SDRAM version
--COMMAND(5) <= (((NOT S6) AND (NOT S5)) OR (RW_STdone AND (NOT PAGE))) after 1 ns;
--COMMAND(4) <= (((NOT S6) AND (NOT S5)) OR (RW_STdone AND (    PAGE))) after 1 ns;

  COMMAND(3) <= (((NOT S6) AND (NOT S5)) OR
                 (S4) OR
                 (S3) OR
                 (S2 AND (NOT do_same_pc_rg)) OR
                 (S5 AND (NOT S2) AND (NOT S1) AND (NOT S0))) after 1 ns;

  COMMAND(2) <= (((NOT S6) AND (NOT S5)) OR NOT_RWdone) after 1 ns;

  COMMAND(1) <= (((NOT S6) AND (NOT S5)) OR
                 (S7) OR
                 ((NOT S4) AND S3) OR
                 (S5 AND (NOT S4) AND (NOT S2) AND (NOT S1))) after 1 ns;

  WRITEdone     <= (S5 AND (    S4)             );
  READdone      <= (S5 AND (NOT S4) AND (    S3));
  RD_WRdone     <= (S5 AND (    S4   OR      S3));
  NOT_RWdone    <= (S5 AND (NOT S4) AND (NOT S3));

  RW_STdone     <= ((    S4) OR  (    S3)  OR (S2 AND (NOT do_same_pc_rg)));

  RW_ST_PCdone  <= (S4 OR S3 OR S2 OR S1);

  STOPdone      <= ((    S2) AND (NOT do_same_pc_rg));
  ACTIVEdone    <= ((NOT S2) AND (NOT S1) AND (    S0));
  NOPdone       <= ((NOT S2) AND (NOT S1) AND (NOT S0));

  PRECHARGEdone <= (NOT_RWdone AND (    S2) AND (    do_same_pc_rg)) OR
                   (NOT_RWdone AND (NOT S2) AND (    S1));

  AC_STdone     <= (NOT_RWdone AND ACTIVEdone) OR
                   (NOT_RWdone AND S2 AND (NOT do_same_pc_rg));

  ACTIVE_A <= NOT_RWdone AND ACTIVEdone AND (NOT ACTIVEchoice(1)) AND (NOT ACTIVEchoice(0));
  ACTIVE_B <= NOT_RWdone AND ACTIVEdone AND (NOT ACTIVEchoice(1)) AND (    ACTIVEchoice(0));
  ACTIVE_C <= NOT_RWdone AND ACTIVEdone AND (    ACTIVEchoice(1)) AND (NOT ACTIVEchoice(0));
  ACTIVE_D <= NOT_RWdone AND ACTIVEdone AND (    ACTIVEchoice(1)) AND (    ACTIVEchoice(0));

  ACTIVE_X <= NOT_RWdone AND ACTIVEdone;

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  --SDRAM Inputs and Outputs

  -- Instanciate these flip-flops to ensure they will always be OFDs
  iOFD0 : FD port map (SDRAM_WE,    D=>COMMAND(1), C=>SDCLK);  -- IOB
  iOFD1 : FD port map (SDRAM_CAS,   D=>COMMAND(2), C=>SDCLK);  -- IOB
  iOFD2 : FD port map (SDRAM_RAS,   D=>COMMAND(3), C=>SDCLK);  -- IOB
  iOFD3 : FD port map (SDRAM_CS(0), D=>COMMAND(4), C=>SDCLK);  -- IOB

  -- Monolithic SDRAM version
  SDRAM_CS(1) <= '0';

  -- DDP SDRAM version
--iOFD4 : FD port map (SDRAM_CS(1), D=>COMMAND(5), C=>SDCLK);  -- IOB

  iOFD5 : FD port map (SDRAM_DQM,   D=>DQMc,           C=>SDCLK);  -- IOB
  iOFD6 : FD port map (SDRAM_BA(0), D=>BANK_SELECT(0), C=>SDCLK);  -- IOB
  iOFD7 : FD port map (SDRAM_BA(1), D=>BANK_SELECT(1), C=>SDCLK);  -- IOB

  DQMc <= (NOT proceed) after 1 ns;

  -- ADDR_CE must be high for ACTIVE, READ, WRITE, and MRS commands
  ADDR_CE <= (((NOT COMMAND(3)) AND (    COMMAND(2)) AND (    COMMAND(1))) OR
              ((    COMMAND(3)) AND (NOT COMMAND(2))                     ) OR
              (                     (NOT COMMAND(2)) AND (NOT COMMAND(1)))) after 1 ns;

  -- AD10_CE must be high for ACTIVE, READ, WRITE, MRS, and PRECHARGE commands
  AD10_CE <= (((NOT COMMAND(3)) AND (    COMMAND(2))                     ) OR
              ((    COMMAND(3)) AND (NOT COMMAND(2))                     ) OR
              (                     (NOT COMMAND(2)) AND (NOT COMMAND(1)))) after 1 ns;

  iAD0  : FDCE port map (SDRAM_A(0),  D=>ADDR(0),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD1  : FDCE port map (SDRAM_A(1),  D=>ADDR(1),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD2  : FDCE port map (SDRAM_A(2),  D=>ADDR(2),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD3  : FDCE port map (SDRAM_A(3),  D=>ADDR(3),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD4  : FDCE port map (SDRAM_A(4),  D=>ADDR(4),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD5  : FDCE port map (SDRAM_A(5),  D=>ADDR(5),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD6  : FDCE port map (SDRAM_A(6),  D=>ADDR(6),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD7  : FDCE port map (SDRAM_A(7),  D=>ADDR(7),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD8  : FDCE port map (SDRAM_A(8),  D=>ADDR(8),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD9  : FDCE port map (SDRAM_A(9),  D=>ADDR(9),  C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD10 : FDCE port map (SDRAM_A(10), D=>ADDR(10), C=>SDCLK, CE=>AD10_CE, CLR=>GND);  -- IOB
  iAD11 : FDCE port map (SDRAM_A(11), D=>ADDR(11), C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB
  iAD12 : FDCE port map (SDRAM_A(12), D=>ADDR(12), C=>SDCLK, CE=>ADDR_CE, CLR=>GND);  -- IOB

  DIN_CEd <= DIN_CE after 1 ns;

  iDI0  : FDCE port map (DATA_IN(0),  D=>SD_DIN(0),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI1  : FDCE port map (DATA_IN(1),  D=>SD_DIN(1),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI2  : FDCE port map (DATA_IN(2),  D=>SD_DIN(2),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI3  : FDCE port map (DATA_IN(3),  D=>SD_DIN(3),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI4  : FDCE port map (DATA_IN(4),  D=>SD_DIN(4),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI5  : FDCE port map (DATA_IN(5),  D=>SD_DIN(5),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI6  : FDCE port map (DATA_IN(6),  D=>SD_DIN(6),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI7  : FDCE port map (DATA_IN(7),  D=>SD_DIN(7),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI8  : FDCE port map (DATA_IN(8),  D=>SD_DIN(8),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI9  : FDCE port map (DATA_IN(9),  D=>SD_DIN(9),  C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI10 : FDCE port map (DATA_IN(10), D=>SD_DIN(10), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI11 : FDCE port map (DATA_IN(11), D=>SD_DIN(11), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI12 : FDCE port map (DATA_IN(12), D=>SD_DIN(12), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI13 : FDCE port map (DATA_IN(13), D=>SD_DIN(13), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI14 : FDCE port map (DATA_IN(14), D=>SD_DIN(14), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI15 : FDCE port map (DATA_IN(15), D=>SD_DIN(15), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI16 : FDCE port map (DATA_IN(16), D=>SD_DIN(16), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI17 : FDCE port map (DATA_IN(17), D=>SD_DIN(17), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI18 : FDCE port map (DATA_IN(18), D=>SD_DIN(18), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI19 : FDCE port map (DATA_IN(19), D=>SD_DIN(19), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI20 : FDCE port map (DATA_IN(20), D=>SD_DIN(20), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI21 : FDCE port map (DATA_IN(21), D=>SD_DIN(21), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI22 : FDCE port map (DATA_IN(22), D=>SD_DIN(22), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI23 : FDCE port map (DATA_IN(23), D=>SD_DIN(23), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI24 : FDCE port map (DATA_IN(24), D=>SD_DIN(24), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI25 : FDCE port map (DATA_IN(25), D=>SD_DIN(25), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI26 : FDCE port map (DATA_IN(26), D=>SD_DIN(26), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI27 : FDCE port map (DATA_IN(27), D=>SD_DIN(27), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI28 : FDCE port map (DATA_IN(28), D=>SD_DIN(28), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI29 : FDCE port map (DATA_IN(29), D=>SD_DIN(29), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI30 : FDCE port map (DATA_IN(30), D=>SD_DIN(30), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB
  iDI31 : FDCE port map (DATA_IN(31), D=>SD_DIN(31), C=>SDCLK, CE=>DIN_CEd, CLR=>GND);  -- IOB

  iDO0  : FDCE port map (SD_DOUT(0),  D=>DATA(0),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO1  : FDCE port map (SD_DOUT(1),  D=>DATA(1),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO2  : FDCE port map (SD_DOUT(2),  D=>DATA(2),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO3  : FDCE port map (SD_DOUT(3),  D=>DATA(3),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO4  : FDCE port map (SD_DOUT(4),  D=>DATA(4),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO5  : FDCE port map (SD_DOUT(5),  D=>DATA(5),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO6  : FDCE port map (SD_DOUT(6),  D=>DATA(6),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO7  : FDCE port map (SD_DOUT(7),  D=>DATA(7),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO8  : FDCE port map (SD_DOUT(8),  D=>DATA(8),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO9  : FDCE port map (SD_DOUT(9),  D=>DATA(9),  C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO10 : FDCE port map (SD_DOUT(10), D=>DATA(10), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO11 : FDCE port map (SD_DOUT(11), D=>DATA(11), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO12 : FDCE port map (SD_DOUT(12), D=>DATA(12), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO13 : FDCE port map (SD_DOUT(13), D=>DATA(13), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO14 : FDCE port map (SD_DOUT(14), D=>DATA(14), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO15 : FDCE port map (SD_DOUT(15), D=>DATA(15), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO16 : FDCE port map (SD_DOUT(16), D=>DATA(16), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO17 : FDCE port map (SD_DOUT(17), D=>DATA(17), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO18 : FDCE port map (SD_DOUT(18), D=>DATA(18), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO19 : FDCE port map (SD_DOUT(19), D=>DATA(19), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO20 : FDCE port map (SD_DOUT(20), D=>DATA(20), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO21 : FDCE port map (SD_DOUT(21), D=>DATA(21), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO22 : FDCE port map (SD_DOUT(22), D=>DATA(22), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO23 : FDCE port map (SD_DOUT(23), D=>DATA(23), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO24 : FDCE port map (SD_DOUT(24), D=>DATA(24), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO25 : FDCE port map (SD_DOUT(25), D=>DATA(25), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO26 : FDCE port map (SD_DOUT(26), D=>DATA(26), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO27 : FDCE port map (SD_DOUT(27), D=>DATA(27), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO28 : FDCE port map (SD_DOUT(28), D=>DATA(28), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO29 : FDCE port map (SD_DOUT(29), D=>DATA(29), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO30 : FDCE port map (SD_DOUT(30), D=>DATA(30), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB
  iDO31 : FDCE port map (SD_DOUT(31), D=>DATA(31), C=>SDCLK, CE=>DOUT_CE, CLR=>GND);  -- IOB

  DOUT_CE <= '1';

  SD_OEN  <= (NOT DRIVE);

  -- Monolithic SDRAM version
  SDRAM_CKE <= "01";

  -- DDP SDRAM version
--SDRAM_CKE <= "11";

  -- A10/AP must be low for bank specific precharge. Set high for all banks precharge.
  -- A10/AP must be set low for r/w without auto-precharge. Set high for auto-precharge.
  -- Therefore A10/AP must be LOW during READ command, WRITE command and PRECHARGE command.
  -- During READ, WRITE or PRECHARGE command, ADDR_SEL="00", therefore ADDR(10)='0';

  -- Full Page burst length
--MODEdata <= "0000000110111";

  -- First Precharge value (set A10/AP to 1 to precharge all banks)
--F_PCdata <= "0010000110111";

  ADDR(0)  <= ((((NOT RD_WRdone) AND ROW_ADDR(0)) OR (RD_WRdone AND COL_ADDR(0))) OR  (NOT proceed)) after 1 ns;
  ADDR(1)  <= ((((NOT RD_WRdone) AND ROW_ADDR(1)) OR (RD_WRdone AND COL_ADDR(1))) OR  (NOT proceed)) after 1 ns;
  ADDR(2)  <= ((((NOT RD_WRdone) AND ROW_ADDR(2)) OR (RD_WRdone AND COL_ADDR(2))) OR  (NOT proceed)) after 1 ns;
  ADDR(3)  <= ((((NOT RD_WRdone) AND ROW_ADDR(3)) OR (RD_WRdone AND COL_ADDR(3))) AND (    proceed)) after 1 ns;
  ADDR(4)  <= ((((NOT RD_WRdone) AND ROW_ADDR(4)) OR (RD_WRdone AND COL_ADDR(4))) OR  (NOT proceed)) after 1 ns;
  ADDR(5)  <= ((((NOT RD_WRdone) AND ROW_ADDR(5)) OR (RD_WRdone AND COL_ADDR(5))) OR  (NOT proceed)) after 1 ns;
  ADDR(6)  <= ((((NOT RD_WRdone) AND ROW_ADDR(6)) OR (RD_WRdone AND COL_ADDR(6))) AND (    proceed)) after 1 ns;
  ADDR(7)  <= ((((NOT RD_WRdone) AND ROW_ADDR(7)) OR (RD_WRdone AND COL_ADDR(7))) AND (    proceed)) after 1 ns;
  ADDR(8)  <= ((((NOT RD_WRdone) AND ROW_ADDR(8)) OR (RD_WRdone AND COL_ADDR(8))) AND (    proceed)) after 1 ns;
  ADDR(9)  <= ((((NOT RD_WRdone) AND ROW_ADDR(9)) OR (RD_WRdone AND COL_ADDR(9))) AND (    proceed)) after 1 ns;

  ADDR(10) <= ((((NOT_RWdone AND ACTIVEdone) AND ROW_ADDR(10)) AND proceed) OR do_first_pc) after 1 ns;

  -- Monolithic SDRAM version
  ADDR(11) <= ((((NOT RD_WRdone) AND ROW_ADDR(11)) OR (RD_WRdone AND PAGE)) AND (    proceed)) after 1 ns;

  -- DDP SDRAM version
--ADDR(11) <= (((NOT RD_WRdone) AND ROW_ADDR(11)) AND proceed) after 1 ns;

  ADDR(12) <= (((NOT RD_WRdone) AND ROW_ADDR(12)) AND proceed) after 1 ns;

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  BANK_A_rd_okc <= '1' when RD_R_ADDR=BANK_A_addr else '0';
  BANK_B_rd_okc <= '1' when RD_R_ADDR=BANK_B_addr else '0';
  BANK_C_rd_okc <= '1' when RD_R_ADDR=BANK_C_addr else '0';
  BANK_D_rd_okc <= '1' when RD_R_ADDR=BANK_D_addr else '0';

  BANK_A_wr_okc <= '1' when WR_R_ADDR=BANK_A_addr else '0';
  BANK_B_wr_okc <= '1' when WR_R_ADDR=BANK_B_addr else '0';
  BANK_C_wr_okc <= '1' when WR_R_ADDR=BANK_C_addr else '0';
  BANK_D_wr_okc <= '1' when WR_R_ADDR=BANK_D_addr else '0';

  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      BANK_A_rd_ok <= BANK_A_rd_okc;
      BANK_B_rd_ok <= BANK_B_rd_okc;
      BANK_C_rd_ok <= BANK_C_rd_okc;
      BANK_D_rd_ok <= BANK_D_rd_okc;
      BANK_A_wr_ok <= BANK_A_wr_okc;
      BANK_B_wr_ok <= BANK_B_wr_okc;
      BANK_C_wr_ok <= BANK_C_wr_okc;
      BANK_D_wr_ok <= BANK_D_wr_okc;
    end if;
  end process;

  -- 1 Port Version
  BANK_X_choice <= RD_R_ADDR when NEXT_X_order='0' else WR_R_ADDR;

  -- DDP SDRAM version
--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    PAGE0_ACTIVE <= '0';
--    PAGE1_ACTIVE <= '0';
--  elsif rising_edge(SDCLK) then
--    PAGE0_ACTIVE <= '0';
--    PAGE1_ACTIVE <= '0';
--    if (ACTIVE_X='1' and BANK_X_choice(13)='0') or (PAGE0_ACTIVE='1' and do_idle='0') then
--      PAGE0_ACTIVE <= '1';
--    end if;
--    if (ACTIVE_X='1' and BANK_X_choice(13)='1') or (PAGE1_ACTIVE='1' and do_idle='0') then
--      PAGE1_ACTIVE <= '1';
--    end if;
--  end if;
--end process;

  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      if ACTIVE_A='1' then
        BANK_A_addr <= BANK_X_choice;
      end if;
      if ACTIVE_B='1' then
        BANK_B_addr <= BANK_X_choice;
      end if;
      if ACTIVE_C='1' then
        BANK_C_addr <= BANK_X_choice;
      end if;
      if ACTIVE_D='1' then
        BANK_D_addr <= BANK_X_choice;
      end if;
    end if;
  end process;

  -- 4 Port Version
--BANK_A_choice <= P_A_RD_R_ADDR when NEXT_A_order='0' else P_A_WR_R_ADDR;
--BANK_B_choice <= P_B_RD_R_ADDR when NEXT_B_order='0' else P_B_WR_R_ADDR;
--BANK_C_choice <= P_C_RD_R_ADDR when NEXT_C_order='0' else P_C_WR_R_ADDR;
--BANK_D_choice <= P_D_RD_R_ADDR when NEXT_D_order='0' else P_D_WR_R_ADDR;

--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    if ACTIVE_A='1' then
--      BANK_A_addr <= BANK_A_choice;
--    end if;
--    if ACTIVE_B='1' then
--      BANK_B_addr <= BANK_B_choice;
--    end if;
--    if ACTIVE_C='1' then
--      BANK_C_addr <= BANK_C_choice;
--    end if;
--    if ACTIVE_D='1' then
--      BANK_D_addr <= BANK_D_choice;
--    end if;
--  end if;
--end process;

  -- 1 Port Version
  -- ROW_ADDR must be valid during clock cycle where ACTIVEdone is asserted (set high)
  -- Monolithic SDRAM version
  ROW_ADDR <= BANK_X_choice(13 downto 1);

  -- DDP SDRAM version
--ROW_ADDR <= BANK_X_choice(12 downto 0);

  RD_COL_ADDR <= RD_C_ADDR;
  WR_COL_ADDR <= WR_C_ADDR;

  -- 4 Port Version
--process (ACTIVEchoice, BANK_A_choice(12 downto 0), BANK_B_choice(12 downto 0),
--                       BANK_C_choice(12 downto 0), BANK_D_choice(12 downto 0))
--begin
--  case ACTIVEchoice is
--    when "00"   => ROW_ADDR <= BANK_A_choice(12 downto 0);
--    when "01"   => ROW_ADDR <= BANK_B_choice(12 downto 0);
--    when "10"   => ROW_ADDR <= BANK_C_choice(12 downto 0);
--    when "11"   => ROW_ADDR <= BANK_D_choice(12 downto 0);
--    when others => ROW_ADDR <= (others=>'0');
--  end case;
--end process;

  -- RD_COL_ADDR must be valid during clock cycle where READdone is asserted (set high)
--process (READchoice, P_A_RD_C_ADDR, P_B_RD_C_ADDR, P_C_RD_C_ADDR, P_D_RD_C_ADDR)
--begin
--  case READchoice is
--    when "00"   => RD_COL_ADDR <= P_A_RD_C_ADDR;
--    when "01"   => RD_COL_ADDR <= P_B_RD_C_ADDR;
--    when "10"   => RD_COL_ADDR <= P_C_RD_C_ADDR;
--    when "11"   => RD_COL_ADDR <= P_D_RD_C_ADDR;
--    when others => RD_COL_ADDR <= (others=>'0');
--  end case;
--end process;

  -- WR_COL_ADDR must be valid during clock cycle where WRITEdone is asserted (set high)
--process (WRITEchoice, P_A_WR_C_ADDR, P_B_WR_C_ADDR, P_C_WR_C_ADDR, P_D_WR_C_ADDR)
--begin
--  case WRITEchoice is
--    when "00"   => WR_COL_ADDR <= P_A_WR_C_ADDR;
--    when "01"   => WR_COL_ADDR <= P_B_WR_C_ADDR;
--    when "10"   => WR_COL_ADDR <= P_C_WR_C_ADDR;
--    when "11"   => WR_COL_ADDR <= P_D_WR_C_ADDR;
--    when others => WR_COL_ADDR <= (others=>'0');
--  end case;
--end process;

  COL_ADDR <= WR_COL_ADDR when S4='1' else RD_COL_ADDR;

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  -- 1 Port Version
  WRITE_TO_A <= S4 AND (NOT WRITE_BANK(1)) AND (NOT WRITE_BANK(0));
  WRITE_TO_B <= S4 AND (NOT WRITE_BANK(1)) AND (    WRITE_BANK(0));
  WRITE_TO_C <= S4 AND (    WRITE_BANK(1)) AND (NOT WRITE_BANK(0));
  WRITE_TO_D <= S4 AND (    WRITE_BANK(1)) AND (    WRITE_BANK(0));

  READ_FROM_A <= (NOT S4) AND S3 AND (NOT READ_BANK(1)) AND (NOT READ_BANK(0));
  READ_FROM_B <= (NOT S4) AND S3 AND (NOT READ_BANK(1)) AND (    READ_BANK(0));
  READ_FROM_C <= (NOT S4) AND S3 AND (    READ_BANK(1)) AND (NOT READ_BANK(0));
  READ_FROM_D <= (NOT S4) AND S3 AND (    READ_BANK(1)) AND (    READ_BANK(0));

  -- 4 Port Version
--WRITE_TO_A <= S4 AND (NOT WRITEchoice(1)) AND (NOT WRITEchoice(0));
--WRITE_TO_B <= S4 AND (NOT WRITEchoice(1)) AND (    WRITEchoice(0));
--WRITE_TO_C <= S4 AND (    WRITEchoice(1)) AND (NOT WRITEchoice(0));
--WRITE_TO_D <= S4 AND (    WRITEchoice(1)) AND (    WRITEchoice(0));

--READ_FROM_A <= (NOT S4) AND S3 AND (NOT READchoice(1)) AND (NOT READchoice(0));
--READ_FROM_B <= (NOT S4) AND S3 AND (NOT READchoice(1)) AND (    READchoice(0));
--READ_FROM_C <= (NOT S4) AND S3 AND (    READchoice(1)) AND (NOT READchoice(0));
--READ_FROM_D <= (NOT S4) AND S3 AND (    READchoice(1)) AND (    READchoice(0));

  -- 1 Port Version
  BANK_A_wip <= WRITE_TO_A OR (BANK_A_wip_rg AND (NOT STOP_WR_X_DL));
  BANK_B_wip <= WRITE_TO_B OR (BANK_B_wip_rg AND (NOT STOP_WR_X_DL));
  BANK_C_wip <= WRITE_TO_C OR (BANK_C_wip_rg AND (NOT STOP_WR_X_DL));
  BANK_D_wip <= WRITE_TO_D OR (BANK_D_wip_rg AND (NOT STOP_WR_X_DL));

  BANK_A_rip <= READ_FROM_A OR (BANK_A_rip_rg AND (NOT STOP_RD_X_DL));
  BANK_B_rip <= READ_FROM_B OR (BANK_B_rip_rg AND (NOT STOP_RD_X_DL));
  BANK_C_rip <= READ_FROM_C OR (BANK_C_rip_rg AND (NOT STOP_RD_X_DL));
  BANK_D_rip <= READ_FROM_D OR (BANK_D_rip_rg AND (NOT STOP_RD_X_DL));

  -- 4 Port Version
--BANK_A_wip <= WRITE_TO_A OR (BANK_A_wip_rg AND (NOT STOP_WR_A_DL));
--BANK_B_wip <= WRITE_TO_B OR (BANK_B_wip_rg AND (NOT STOP_WR_B_DL));
--BANK_C_wip <= WRITE_TO_C OR (BANK_C_wip_rg AND (NOT STOP_WR_C_DL));
--BANK_D_wip <= WRITE_TO_D OR (BANK_D_wip_rg AND (NOT STOP_WR_D_DL));

--BANK_A_rip <= READ_FROM_A OR (BANK_A_rip_rg AND (NOT STOP_RD_A_DL));
--BANK_B_rip <= READ_FROM_B OR (BANK_B_rip_rg AND (NOT STOP_RD_B_DL));
--BANK_C_rip <= READ_FROM_C OR (BANK_C_rip_rg AND (NOT STOP_RD_C_DL));
--BANK_D_rip <= READ_FROM_D OR (BANK_D_rip_rg AND (NOT STOP_RD_D_DL));

  BANK_X_wip <= (S4) OR (BANK_X_wip_rg AND (NOT WRITEend));

  BANK_X_rip <= ((NOT S4) AND S3) OR (BANK_X_rip_rg AND (NOT READend));

  -- Change access when:
  --  i) current access is running out of data
  -- ii) burst limit has been reached

  -- 1 Port Version
  ALLOW_NEW_RD_A <= (BANK_A_rip AND (NOT STOPCOND_RD_X)) OR
                    (BANK_A_wip AND (NOT STOPCOND_WR_X));

  ALLOW_NEW_RD_B <= (BANK_B_rip AND (NOT STOPCOND_RD_X)) OR
                    (BANK_B_wip AND (NOT STOPCOND_WR_X));

  ALLOW_NEW_RD_C <= (BANK_C_rip AND (NOT STOPCOND_RD_X)) OR
                    (BANK_C_wip AND (NOT STOPCOND_WR_X));

  ALLOW_NEW_RD_D <= (BANK_D_rip AND (NOT STOPCOND_RD_X)) OR
                    (BANK_D_wip AND (NOT STOPCOND_WR_X));

  ALLOW_NEW_RD <= NOT (ALLOW_NEW_RD_A OR ALLOW_NEW_RD_B OR ALLOW_NEW_RD_C OR ALLOW_NEW_RD_D);

  ALLOW_NEW_WR <= NOT ((BANK_A_wip AND (NOT STOPCOND_WR_X)) OR
                       (BANK_B_wip AND (NOT STOPCOND_WR_X)) OR
                       (BANK_C_wip AND (NOT STOPCOND_WR_X)) OR
                       (BANK_D_wip AND (NOT STOPCOND_WR_X)) OR
                       (READdone                          ) OR
                       (RIP_RG                            ));

  -- 4 Port Version
--ALLOW_NEW_RD_A <= (BANK_A_rip AND (NOT STOPCOND_RD_A)) OR
--                  (BANK_A_wip AND (NOT STOPCOND_WR_A));

--ALLOW_NEW_RD_B <= (BANK_B_rip AND (NOT STOPCOND_RD_B)) OR
--                  (BANK_B_wip AND (NOT STOPCOND_WR_B));

--ALLOW_NEW_RD_C <= (BANK_C_rip AND (NOT STOPCOND_RD_C)) OR
--                  (BANK_C_wip AND (NOT STOPCOND_WR_C));

--ALLOW_NEW_RD_D <= (BANK_D_rip AND (NOT STOPCOND_RD_D)) OR
--                  (BANK_D_wip AND (NOT STOPCOND_WR_D));

--ALLOW_NEW_WR <= NOT ((BANK_A_wip AND (NOT STOPCOND_WR_A)) OR
--                     (BANK_B_wip AND (NOT STOPCOND_WR_B)) OR
--                     (BANK_C_wip AND (NOT STOPCOND_WR_C)) OR
--                     (BANK_D_wip AND (NOT STOPCOND_WR_D)) OR
--                     (READdone                          ) OR
--                     (RIP_RG                            ));

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      RIP_RG <= '0';
    elsif rising_edge(SDCLK) then
      RIP_RG <= '0';
      if ((S4='0' and S3='1') or (RIP_RG='1' and READend_dl2='0') or
          BANK_A_rip_rg='1' or BANK_B_rip_rg='1' or BANK_C_rip_rg='1' or BANK_D_rip_rg='1') then
        RIP_RG <= '1';
      end if;
    end if;
  end process;

  -- 1 Port Version
  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_A_wip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_A_wip_rg <= '0';
      if WRITE_TO_A='1' or (BANK_A_wip_rg='1' and STOP_WR_X_DL='0')  then
        BANK_A_wip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_A_rip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_A_rip_rg <= '0';
      if READ_FROM_A='1' or (BANK_A_rip_rg='1' and STOP_RD_X_DL='0') then
        BANK_A_rip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_B_wip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_B_wip_rg <= '0';
      if WRITE_TO_B='1' or (BANK_B_wip_rg='1' and STOP_WR_X_DL='0')  then
        BANK_B_wip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_B_rip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_B_rip_rg <= '0';
      if READ_FROM_B='1' or (BANK_B_rip_rg='1' and STOP_RD_X_DL='0')  then
        BANK_B_rip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_C_wip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_C_wip_rg <= '0';
      if WRITE_TO_C='1' or (BANK_C_wip_rg='1' and STOP_WR_X_DL='0')  then
        BANK_C_wip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_C_rip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_C_rip_rg <= '0';
      if READ_FROM_C='1' or (BANK_C_rip_rg='1' and STOP_RD_X_DL='0')  then
        BANK_C_rip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_D_wip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_D_wip_rg <= '0';
      if WRITE_TO_D='1' or (BANK_D_wip_rg='1' and STOP_WR_X_DL='0')  then
        BANK_D_wip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_D_rip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_D_rip_rg <= '0';
      if READ_FROM_D='1' or (BANK_D_rip_rg='1' and STOP_RD_X_DL='0')  then
        BANK_D_rip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_X_wip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_X_wip_rg <= '0';
      if S4='1' or (BANK_X_wip_rg='1' and WRITEend='0') then
        BANK_X_wip_rg <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      BANK_X_rip_rg <= '0';
    elsif rising_edge(SDCLK) then
      BANK_X_rip_rg <= '0';
      if (S4='0' and S3='1') or (BANK_X_rip_rg='1' and READend='0')  then
        BANK_X_rip_rg <= '1';
      end if;
    end if;
  end process;

  -- 4 Port Version
--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    BANK_A_wip_rg <= '0';
--  elsif rising_edge(SDCLK) then
--    BANK_A_wip_rg <= '0';
--    if WRITE_TO_A='1' or (BANK_A_wip_rg='1' and STOP_WR_A_DL='0')  then
--      BANK_A_wip_rg <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    BANK_A_rip_rg <= '0';
--  elsif rising_edge(SDCLK) then
--    BANK_A_rip_rg <= '0';
--    if READ_FROM_A='1' or (BANK_A_rip_rg='1' and STOP_RD_A_DL='0') then
--      BANK_A_rip_rg <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    BANK_B_wip_rg <= '0';
--  elsif rising_edge(SDCLK) then
--    BANK_B_wip_rg <= '0';
--    if WRITE_TO_B='1' or (BANK_B_wip_rg='1' and STOP_WR_B_DL='0')  then
--      BANK_B_wip_rg <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    BANK_B_rip_rg <= '0';
--  elsif rising_edge(SDCLK) then
--    BANK_B_rip_rg <= '0';
--    if READ_FROM_B='1' or (BANK_B_rip_rg='1' and STOP_RD_B_DL='0')  then
--      BANK_B_rip_rg <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    BANK_C_wip_rg <= '0';
--  elsif rising_edge(SDCLK) then
--    BANK_C_wip_rg <= '0';
--    if WRITE_TO_C='1' or (BANK_C_wip_rg='1' and STOP_WR_C_DL='0')  then
--      BANK_C_wip_rg <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    BANK_C_rip_rg <= '0';
--  elsif rising_edge(SDCLK) then
--    BANK_C_rip_rg <= '0';
--    if READ_FROM_C='1' or (BANK_C_rip_rg='1' and STOP_RD_C_DL='0')  then
--      BANK_C_rip_rg <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    BANK_D_wip_rg <= '0';
--  elsif rising_edge(SDCLK) then
--    BANK_D_wip_rg <= '0';
--    if WRITE_TO_D='1' or (BANK_D_wip_rg='1' and STOP_WR_D_DL='0')  then
--      BANK_D_wip_rg <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    BANK_D_rip_rg <= '0';
--  elsif rising_edge(SDCLK) then
--    BANK_D_rip_rg <= '0';
--    if READ_FROM_D='1' or (BANK_D_rip_rg='1' and STOP_RD_D_DL='0')  then
--      BANK_D_rip_rg <= '1';
--    end if;
--  end if;
--end process;

  -- 1 Port Version
  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      STOP_RD_X_DL <= STOP_RD_X;
    end if;
  end process;

  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      STOP_WR_X_DL <= STOP_WR_X;
    end if;
  end process;

  READend  <= STOP_RD_X_DL;
  WRITEend <= STOP_WR_X_DL;

  -- 4 Port Version
--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    STOP_RD_A_DL <= STOP_RD_A;
--    STOP_RD_B_DL <= STOP_RD_B;
--    STOP_RD_C_DL <= STOP_RD_C;
--    STOP_RD_D_DL <= STOP_RD_D;
--  end if;
--end process;

--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    STOP_WR_A_DL <= STOP_WR_A;
--    STOP_WR_B_DL <= STOP_WR_B;
--    STOP_WR_C_DL <= STOP_WR_C;
--    STOP_WR_D_DL <= STOP_WR_D;
--  end if;
--end process;

--READend  <= STOP_RD_A_DL OR STOP_RD_B_DL OR STOP_RD_C_DL OR STOP_RD_D_DL;
--WRITEend <= STOP_WR_A_DL OR STOP_WR_B_DL OR STOP_WR_C_DL OR STOP_WR_D_DL;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      DRIVE <= '0';
    elsif rising_edge(SDCLK) then
      DRIVE <= '0';
      if WRITEdone='1' or (DRIVE='1' and WRITEend='0') then
        DRIVE <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      RDdone_dl1 <= '0';
      RDdone_dl2 <= '0';
      RDdone_dl3 <= '0';
    elsif rising_edge(SDCLK) then
      RDdone_dl1 <= READdone;
      RDdone_dl2 <= RDdone_dl1;
      RDdone_dl3 <= RDdone_dl2;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      READend_dl1 <= '0';
      READend_dl2 <= '0';
      READend_dl3 <= '0';
    elsif rising_edge(SDCLK) then
      READend_dl1 <= READend;
      READend_dl2 <= READend_dl1;
      READend_dl3 <= READend_dl2;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      DIN_CE <= '0';
    elsif rising_edge(SDCLK) then
      DIN_CE <= '0';
      if RDdone_dl3='1' or (DIN_CE='1' and READend_dl3='0') then
        DIN_CE <= '1';
      end if;
    end if;
  end process;

  -- 1 Port Version
  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      SDRAM_RD_DVAL <= '0';
      if DIN_CE='1' then
        SDRAM_RD_DVAL <= '1';
      end if;
    end if;
  end process;

  -- 4 Port Version
--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    RDaccess_dl1 <= RW_ACCESS_IP;
--    RDaccess_dl2 <= RDaccess_dl1;
--    RDaccess_dl3 <= RDaccess_dl2;
--  end if;
--end process;

--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    SDRAM_RD_DVAL_A <= '0';
--    SDRAM_RD_DVAL_B <= '0';
--    SDRAM_RD_DVAL_C <= '0';
--    SDRAM_RD_DVAL_D <= '0';
--    if DIN_CE='1' and RDaccess_dl3="00" then
--      SDRAM_RD_DVAL_A <= '1';
--    end if;
--    if DIN_CE='1' and RDaccess_dl3="01" then
--      SDRAM_RD_DVAL_B <= '1';
--    end if;
--    if DIN_CE='1' and RDaccess_dl3="10" then
--      SDRAM_RD_DVAL_C <= '1';
--    end if;
--    if DIN_CE='1' and RDaccess_dl3="11" then
--      SDRAM_RD_DVAL_D <= '1';
--    end if;
--  end if;
--end process;

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  -- 1 Port Version
  -- X_order=1 for write priority, 0 for read priority
  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      NEXT_X_order <= '0';
      if (DVALID_WR_X='1' and LAST_X_order='0') or
         (DVALID_RD_X='0' and LAST_X_order='1') then
        NEXT_X_order <= '1';
      end if;
    end if;
  end process;

  process(RST_SDRAM, SDCLK)
  begin
    if RST_SDRAM='1' then
      LAST_X_order <= '0';
    elsif rising_edge(SDCLK) then
      if ACTIVE_X='1' then
        LAST_X_order <= NEXT_X_order;
      end if;
    end if;
  end process;

  -- 4 Port Version
---- A_order=1 for write priority, 0 for read priority
--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    NEXT_A_order <= '0';
--    if (DVALID_WR_A='1' and LAST_A_order='0') or
--       (DVALID_RD_A='0' and LAST_A_order='1') then
--      NEXT_A_order <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    LAST_A_order <= '0';
--  elsif rising_edge(SDCLK) then
--    if ACTIVE_A='1' then
--      LAST_A_order <= NEXT_A_order;
--    end if;
--  end if;
--end process;

---- B_order=1 for write priority, 0 for read priority
--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    NEXT_B_order <= '0';
--    if (DVALID_WR_B='1' and LAST_B_order='0') or
--       (DVALID_RD_B='0' and LAST_B_order='1') then
--      NEXT_B_order <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    LAST_B_order <= '0';
--  elsif rising_edge(SDCLK) then
--    if ACTIVE_B='1' then
--      LAST_B_order <= NEXT_B_order;
--    end if;
--  end if;
--end process;

---- C_order=1 for write priority, 0 for read priority
--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    NEXT_C_order <= '0';
--    if (DVALID_WR_C='1' and LAST_C_order='0') or
--       (DVALID_RD_C='0' and LAST_C_order='1') then
--      NEXT_C_order <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    LAST_C_order <= '0';
--  elsif rising_edge(SDCLK) then
--    if ACTIVE_C='1' then
--      LAST_C_order <= NEXT_C_order;
--    end if;
--  end if;
--end process;

---- D_order=1 for write priority, 0 for read priority
--process(SDCLK)
--begin
--  if rising_edge(SDCLK) then
--    NEXT_D_order <= '0';
--    if (DVALID_WR_D='1' and LAST_D_order='0') or
--       (DVALID_RD_D='0' and LAST_D_order='1') then
--      NEXT_D_order <= '1';
--    end if;
--  end if;
--end process;

--process(RST_SDRAM, SDCLK)
--begin
--  if RST_SDRAM='1' then
--    LAST_D_order <= '0';
--  elsif rising_edge(SDCLK) then
--    if ACTIVE_D='1' then
--      LAST_D_order <= NEXT_D_order;
--    end if;
--  end if;
--end process;

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  -- 1 Port Version
  iRD : HE_SD_CNTR
    port map (
      SDCLK    => SDCLK,
      RESET    => RST_SDRAM,
      COUNT_EN => BANK_X_rip,
      ADDR_EN  => SDRAM_RD_AEN,
      ADDR_IN  => SDRAM_RD_ADDR,
      ROW_ADDR => RD_R_ADDR,
      COL_ADDR => RD_C_ADDR,
      BANK     => READ_BANK,
      TC_FN_A  => RD_TC_A,
      TC_FN_B  => RD_TC_B,
      TC_FN_C  => RD_TC_C );

  iWR : HE_SD_CNTR
    port map (
      SDCLK    => SDCLK,
      RESET    => RST_SDRAM,
      COUNT_EN => BANK_X_wip,
      ADDR_EN  => SDRAM_WR_AEN,
      ADDR_IN  => SDRAM_WR_ADDR,
      ROW_ADDR => WR_R_ADDR,
      COL_ADDR => WR_C_ADDR,
      BANK     => WRITE_BANK,
      TC_FN_A  => WR_TC_A,
      TC_FN_B  => WR_TC_B,
      TC_FN_C  => WR_TC_C );

  -- 4 Port Version
--iP_A_RD : HE_SD_CNTR
--  port map (
--    SDCLK    => SDCLK,
--    RESET    => RST_SDRAM,
--    COUNT_EN => BANK_A_rip,
--    ADDR_EN  => SDRAM_RD_AEN,
--    ADDR_IN  => SDRAM_RD_ADDR(23 downto 0),
--    ROW_ADDR => P_A_RD_R_ADDR,
--    COL_ADDR => P_A_RD_C_ADDR,
--    TC_FN_A  => P_A_RD_TC_A,
--    TC_FN_B  => P_A_RD_TC_B,
--    TC_FN_C  => P_A_RD_TC_C );

--iP_B_RD : HE_SD_CNTR
--  port map (
--    SDCLK    => SDCLK,
--    RESET    => RST_SDRAM,
--    COUNT_EN => BANK_B_rip,
--    ADDR_EN  => SDRAM_RD_AEN,
--    ADDR_IN  => SDRAM_RD_ADDR(23 downto 0),
--    ROW_ADDR => P_B_RD_R_ADDR,
--    COL_ADDR => P_B_RD_C_ADDR,
--    TC_FN_A  => P_B_RD_TC_A,
--    TC_FN_B  => P_B_RD_TC_B,
--    TC_FN_C  => P_B_RD_TC_C );

--iP_C_RD : HE_SD_CNTR
--  port map (
--    SDCLK    => SDCLK,
--    RESET    => RST_SDRAM,
--    COUNT_EN => BANK_C_rip,
--    ADDR_EN  => SDRAM_RD_AEN,
--    ADDR_IN  => SDRAM_RD_ADDR(23 downto 0),
--    ROW_ADDR => P_C_RD_R_ADDR,
--    COL_ADDR => P_C_RD_C_ADDR,
--    TC_FN_A  => P_C_RD_TC_A,
--    TC_FN_B  => P_C_RD_TC_B,
--    TC_FN_C  => P_C_RD_TC_C );

--iP_D_RD : HE_SD_CNTR
--  port map (
--    SDCLK    => SDCLK,
--    RESET    => RST_SDRAM,
--    COUNT_EN => BANK_D_rip,
--    ADDR_EN  => SDRAM_RD_AEN,
--    ADDR_IN  => SDRAM_RD_ADDR(23 downto 0),
--    ROW_ADDR => P_D_RD_R_ADDR,
--    COL_ADDR => P_D_RD_C_ADDR,
--    TC_FN_A  => P_D_RD_TC_A,
--    TC_FN_B  => P_D_RD_TC_B,
--    TC_FN_C  => P_D_RD_TC_C );

--iP_A_WR : HE_SD_CNTR
--  port map (
--    SDCLK    => SDCLK,
--    RESET    => RST_SDRAM,
--    COUNT_EN => BANK_A_wip,
--    ADDR_EN  => SDRAM_WR_AEN,
--    ADDR_IN  => SDRAM_WR_ADDR(23 downto 0),
--    ROW_ADDR => P_A_WR_R_ADDR,
--    COL_ADDR => P_A_WR_C_ADDR,
--    TC_FN_A  => P_A_WR_TC_A,
--    TC_FN_B  => P_A_WR_TC_B,
--    TC_FN_C  => P_A_WR_TC_C );

--iP_B_WR : HE_SD_CNTR
--  port map (
--    SDCLK    => SDCLK,
--    RESET    => RST_SDRAM,
--    COUNT_EN => BANK_B_wip,
--    ADDR_EN  => SDRAM_WR_AEN,
--    ADDR_IN  => SDRAM_WR_ADDR(23 downto 0),
--    ROW_ADDR => P_B_WR_R_ADDR,
--    COL_ADDR => P_B_WR_C_ADDR,
--    TC_FN_A  => P_B_WR_TC_A,
--    TC_FN_B  => P_B_WR_TC_B,
--    TC_FN_C  => P_B_WR_TC_C );

--iP_C_WR : HE_SD_CNTR
--  port map (
--    SDCLK    => SDCLK,
--    RESET    => RST_SDRAM,
--    COUNT_EN => BANK_C_wip,
--    ADDR_EN  => SDRAM_WR_AEN,
--    ADDR_IN  => SDRAM_WR_ADDR(23 downto 0),
--    ROW_ADDR => P_C_WR_R_ADDR,
--    COL_ADDR => P_C_WR_C_ADDR,
--    TC_FN_A  => P_C_WR_TC_A,
--    TC_FN_B  => P_C_WR_TC_B,
--    TC_FN_C  => P_C_WR_TC_C );

--iP_D_WR : HE_SD_CNTR
--  port map (
--    SDCLK    => SDCLK,
--    RESET    => RST_SDRAM,
--    COUNT_EN => BANK_D_wip,
--    ADDR_EN  => SDRAM_WR_AEN,
--    ADDR_IN  => SDRAM_WR_ADDR(23 downto 0),
--    ROW_ADDR => P_D_WR_R_ADDR,
--    COL_ADDR => P_D_WR_C_ADDR,
--    TC_FN_A  => P_D_WR_TC_A,
--    TC_FN_B  => P_D_WR_TC_B,
--    TC_FN_C  => P_D_WR_TC_C );

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  -- Logic for opening a ROW

  -- DDP SDRAM version
--RD_OPEN_OK <= ((NOT PAGE1_ACTIVE) AND (NOT PAGE0_ACTIVE )                        ) OR
--              ((NOT PAGE1_ACTIVE) AND (    PAGE0_ACTIVE ) AND (NOT RD_R_ADDR(13))) OR
--              ((    PAGE1_ACTIVE) AND (NOT PAGE0_ACTIVE ) AND (    RD_R_ADDR(13)));

  -- DDP SDRAM version
--WR_OPEN_OK <= ((NOT PAGE1_ACTIVE) AND (NOT PAGE0_ACTIVE )                        ) OR
--              ((NOT PAGE1_ACTIVE) AND (    PAGE0_ACTIVE ) AND (NOT WR_R_ADDR(13))) OR
--              ((    PAGE1_ACTIVE) AND (NOT PAGE0_ACTIVE ) AND (    WR_R_ADDR(13)));

  -- Monolithic SDRAM version
  OPEN_A <= DVALID_RD_A OR DVALID_WR_A;
  OPEN_B <= DVALID_RD_B OR DVALID_WR_B;
  OPEN_C <= DVALID_RD_C OR DVALID_WR_C;
  OPEN_D <= DVALID_RD_D OR DVALID_WR_D;

  -- DDP SDRAM version
--OPEN_A <= (DVALID_RD_A AND RD_OPEN_OK) OR (DVALID_WR_A AND WR_OPEN_OK);
--OPEN_B <= (DVALID_RD_B AND RD_OPEN_OK) OR (DVALID_WR_B AND WR_OPEN_OK);
--OPEN_C <= (DVALID_RD_C AND RD_OPEN_OK) OR (DVALID_WR_C AND WR_OPEN_OK);
--OPEN_D <= (DVALID_RD_D AND RD_OPEN_OK) OR (DVALID_WR_D AND WR_OPEN_OK);

  -- 1 Port Version
  DVALID_RD_X <= SDRAM_RD_READY;
  DVALID_WR_X <= SDRAM_WR_READY;

  DVALID_RD_A <= SDRAM_RD_READY AND (NOT READ_BANK(1)) AND (NOT READ_BANK(0));
  DVALID_RD_B <= SDRAM_RD_READY AND (NOT READ_BANK(1)) AND (    READ_BANK(0));
  DVALID_RD_C <= SDRAM_RD_READY AND (    READ_BANK(1)) AND (NOT READ_BANK(0));
  DVALID_RD_D <= SDRAM_RD_READY AND (    READ_BANK(1)) AND (    READ_BANK(0));

  DVALID_WR_A <= SDRAM_WR_READY AND (NOT WRITE_BANK(1)) AND (NOT WRITE_BANK(0));
  DVALID_WR_B <= SDRAM_WR_READY AND (NOT WRITE_BANK(1)) AND (    WRITE_BANK(0));
  DVALID_WR_C <= SDRAM_WR_READY AND (    WRITE_BANK(1)) AND (NOT WRITE_BANK(0));
  DVALID_WR_D <= SDRAM_WR_READY AND (    WRITE_BANK(1)) AND (    WRITE_BANK(0));

  -- 4 Port Version
--DVALID_RD_A <= SDRAM_RD_READY_A;
--DVALID_RD_B <= SDRAM_RD_READY_B;
--DVALID_RD_C <= SDRAM_RD_READY_C;
--DVALID_RD_D <= SDRAM_RD_READY_D;

--DVALID_WR_A <= SDRAM_WR_READY_A;
--DVALID_WR_B <= SDRAM_WR_READY_B;
--DVALID_WR_C <= SDRAM_WR_READY_C;
--DVALID_WR_D <= SDRAM_WR_READY_D;

  -- Logic for closing a ROW

  -- 1 Port Version
  CLOSE_A <= ((NOT DVALID_RD_A) OR (NOT BANK_A_rd_ok)) AND ((NOT DVALID_WR_A) OR (NOT BANK_A_wr_ok));
  CLOSE_B <= ((NOT DVALID_RD_B) OR (NOT BANK_B_rd_ok)) AND ((NOT DVALID_WR_B) OR (NOT BANK_B_wr_ok));
  CLOSE_C <= ((NOT DVALID_RD_C) OR (NOT BANK_C_rd_ok)) AND ((NOT DVALID_WR_C) OR (NOT BANK_C_wr_ok));
  CLOSE_D <= ((NOT DVALID_RD_D) OR (NOT BANK_D_rd_ok)) AND ((NOT DVALID_WR_D) OR (NOT BANK_D_wr_ok));

  -- 4 Port Version
--CLOSE_A <= ((NOT DVALID_RD_A) OR (NOT BANK_A_rd_ok)) AND ((NOT DVALID_WR_A) OR (NOT BANK_A_wr_ok));
--CLOSE_B <= ((NOT DVALID_RD_B) OR (NOT BANK_B_rd_ok)) AND ((NOT DVALID_WR_B) OR (NOT BANK_B_wr_ok));
--CLOSE_C <= ((NOT DVALID_RD_C) OR (NOT BANK_C_rd_ok)) AND ((NOT DVALID_WR_C) OR (NOT BANK_C_wr_ok));
--CLOSE_D <= ((NOT DVALID_RD_D) OR (NOT BANK_D_rd_ok)) AND ((NOT DVALID_WR_D) OR (NOT BANK_D_wr_ok));
   -- !! If doing write can't do PRECHARGE immediately. First do stop, then do precharge.
  --    Therefore, STOP_WR_n must first be set to 1. Then following clock cycle CLOSE_n set to 1.

  -- Logic for starting a read or write

  -- Must include BANK_n_rip_rg and BANK_n_wip_rg so that we don't start a new read or write
  -- from the same bank, immediately following a STOP (while the end of row counter logic is
  -- adjusting)

  -- 1 Port Version
  START_RD_A <= DVALID_RD_X AND BANK_A_rd_ok AND (NOT ((NOT S4) AND S3)) AND (NOT BANK_X_rip_rg);
  START_RD_B <= DVALID_RD_X AND BANK_B_rd_ok AND (NOT ((NOT S4) AND S3)) AND (NOT BANK_X_rip_rg);
  START_RD_C <= DVALID_RD_X AND BANK_C_rd_ok AND (NOT ((NOT S4) AND S3)) AND (NOT BANK_X_rip_rg);
  START_RD_D <= DVALID_RD_X AND BANK_D_rd_ok AND (NOT ((NOT S4) AND S3)) AND (NOT BANK_X_rip_rg);

  START_WR_A <= DVALID_WR_X AND BANK_A_wr_ok AND (NOT (     S4        )) AND (NOT BANK_X_wip_rg);
  START_WR_B <= DVALID_WR_X AND BANK_B_wr_ok AND (NOT (     S4        )) AND (NOT BANK_X_wip_rg);
  START_WR_C <= DVALID_WR_X AND BANK_C_wr_ok AND (NOT (     S4        )) AND (NOT BANK_X_wip_rg);
  START_WR_D <= DVALID_WR_X AND BANK_D_wr_ok AND (NOT (     S4        )) AND (NOT BANK_X_wip_rg);

  -- 4 Port Version
--START_RD_A <= DVALID_RD_A AND BANK_A_rd_ok AND (NOT READ_FROM_A) AND (NOT BANK_A_rip_rg);
--START_RD_B <= DVALID_RD_B AND BANK_B_rd_ok AND (NOT READ_FROM_B) AND (NOT BANK_B_rip_rg);
--START_RD_C <= DVALID_RD_C AND BANK_C_rd_ok AND (NOT READ_FROM_C) AND (NOT BANK_C_rip_rg);
--START_RD_D <= DVALID_RD_D AND BANK_D_rd_ok AND (NOT READ_FROM_D) AND (NOT BANK_D_rip_rg);

--START_WR_A <= DVALID_WR_A AND BANK_A_wr_ok AND (NOT WRITE_TO_A) AND (NOT BANK_A_wip_rg);
--START_WR_B <= DVALID_WR_B AND BANK_B_wr_ok AND (NOT WRITE_TO_B) AND (NOT BANK_B_wip_rg);
--START_WR_C <= DVALID_WR_C AND BANK_C_wr_ok AND (NOT WRITE_TO_C) AND (NOT BANK_C_wip_rg);
--START_WR_D <= DVALID_WR_D AND BANK_D_wr_ok AND (NOT WRITE_TO_D) AND (NOT BANK_D_wip_rg);

  -- Logic for stopping a read or write

  -- 1 Port Version
  STOP_RD_X <= ((((NOT S4) AND S3) OR (BANK_X_rip_rg AND (NOT  READend))) AND STOPCOND_RD_X);
  STOP_WR_X <= (((     S4        ) OR (BANK_X_wip_rg AND (NOT WRITEend))) AND STOPCOND_WR_X);

  STOPCOND_RD_X <= (NOT SDRAM_RD_BURST) OR (RD_TC_A AND RD_TC_B AND RD_TC_C);
  STOPCOND_WR_X <= (NOT SDRAM_WR_BURST) OR (WR_TC_A AND WR_TC_B AND WR_TC_C);

  -- 4 Port Version
--STOP_RD_A <= ((READ_FROM_A OR (BANK_A_rip_rg AND (NOT STOP_RD_A_DL))) AND STOPCOND_RD_A);
--STOP_RD_B <= ((READ_FROM_B OR (BANK_B_rip_rg AND (NOT STOP_RD_B_DL))) AND STOPCOND_RD_B);
--STOP_RD_C <= ((READ_FROM_C OR (BANK_C_rip_rg AND (NOT STOP_RD_C_DL))) AND STOPCOND_RD_C);
--STOP_RD_D <= ((READ_FROM_D OR (BANK_D_rip_rg AND (NOT STOP_RD_D_DL))) AND STOPCOND_RD_D);

--STOP_WR_A <= ((WRITE_TO_A OR (BANK_A_wip_rg AND (NOT STOP_WR_A_DL))) AND STOPCOND_WR_A);
--STOP_WR_B <= ((WRITE_TO_B OR (BANK_B_wip_rg AND (NOT STOP_WR_B_DL))) AND STOPCOND_WR_B);
--STOP_WR_C <= ((WRITE_TO_C OR (BANK_C_wip_rg AND (NOT STOP_WR_C_DL))) AND STOPCOND_WR_C);
--STOP_WR_D <= ((WRITE_TO_D OR (BANK_D_wip_rg AND (NOT STOP_WR_D_DL))) AND STOPCOND_WR_D);

--STOPCOND_RD_A <= (NOT SDRAM_RD_BURST_A) OR (P_A_RD_TC_A AND P_A_RD_TC_B AND P_A_RD_TC_C);
--STOPCOND_RD_B <= (NOT SDRAM_RD_BURST_B) OR (P_B_RD_TC_A AND P_B_RD_TC_B AND P_B_RD_TC_C);
--STOPCOND_RD_C <= (NOT SDRAM_RD_BURST_C) OR (P_C_RD_TC_A AND P_C_RD_TC_B AND P_C_RD_TC_C);
--STOPCOND_RD_D <= (NOT SDRAM_RD_BURST_D) OR (P_D_RD_TC_A AND P_D_RD_TC_B AND P_D_RD_TC_C);

--STOPCOND_WR_A <= (NOT SDRAM_WR_BURST_A) OR (P_A_WR_TC_A AND P_A_WR_TC_B AND P_A_WR_TC_C);
--STOPCOND_WR_B <= (NOT SDRAM_WR_BURST_B) OR (P_B_WR_TC_A AND P_B_WR_TC_B AND P_B_WR_TC_C);
--STOPCOND_WR_C <= (NOT SDRAM_WR_BURST_C) OR (P_C_WR_TC_A AND P_C_WR_TC_B AND P_C_WR_TC_C);
--STOPCOND_WR_D <= (NOT SDRAM_WR_BURST_D) OR (P_D_WR_TC_A AND P_D_WR_TC_B AND P_D_WR_TC_C);

  --------------------------------------------------------------------------------------------------
  --------------------------------------------------------------------------------------------------

  -- 1 Port Version
  SDRAM_WR_DONE <= BANK_X_wip;
  SDRAM_RD_DONE <= BANK_X_rip;

  SDRAM_RD_DATA <= DATA_IN;
  DATA          <= SDRAM_WR_DATA after 1 ns;

  -- 4 Port Version
--SDRAM_WR_DONE_A <= BANK_A_wip;
--SDRAM_WR_DONE_B <= BANK_B_wip;
--SDRAM_WR_DONE_C <= BANK_C_wip;
--SDRAM_WR_DONE_D <= BANK_D_wip;

--SDRAM_RD_DONE_A <= BANK_A_rip;
--SDRAM_RD_DONE_B <= BANK_B_rip;
--SDRAM_RD_DONE_C <= BANK_C_rip;
--SDRAM_RD_DONE_D <= BANK_D_rip;

--SDRAM_RD_DATA_A <= DATA_IN;
--SDRAM_RD_DATA_B <= DATA_IN;
--SDRAM_RD_DATA_C <= DATA_IN;
--SDRAM_RD_DATA_D <= DATA_IN;

--DMUX_SEL(0) <= BANK_B_wip OR BANK_D_wip;
--DMUX_SEL(1) <= BANK_C_wip OR BANK_D_wip;

--process (DMUX_SEL, SDRAM_WR_DATA_A, SDRAM_WR_DATA_B, SDRAM_WR_DATA_C, SDRAM_WR_DATA_D)
--begin
--  case DMUX_SEL is
--    when "00"   => DATA <= SDRAM_WR_DATA_A;
--    when "01"   => DATA <= SDRAM_WR_DATA_B;
--    when "10"   => DATA <= SDRAM_WR_DATA_C;
--    when "11"   => DATA <= SDRAM_WR_DATA_D;
--    when others => DATA <= (others=>'0');
--  end case;
--end process;

end RTL;


-- ==============================================================
--
--  Module      : HE_SD_CNTR
--  Date        : 24/03/2003
--  Author      : R. Williams - HUNT ENGINEERING
--  Description : Row, Column and Bank Address Counters
--
-- ==============================================================


library IEEE;
  use IEEE.std_logic_1164.all;
  use IEEE.std_logic_arith.all;
  use IEEE.std_logic_unsigned.all;

-- synopsys translate_off
library UNISIM;
  use UNISIM.vcomponents.all;
-- synopsys translate_on

entity HE_SD_CNTR is
  port (
    SDCLK    : in  std_logic;
    RESET    : in  std_logic;
    COUNT_EN : in  std_logic;
    ADDR_EN  : in  std_logic;
    ADDR_IN  : in  std_logic_vector(25 downto 0);
    ROW_ADDR : out std_logic_vector(13 downto 0);
    COL_ADDR : out std_logic_vector(9 downto 0);
    BANK     : out std_logic_vector(1 downto 0);
    TC_FN_A  : out std_logic;
    TC_FN_B  : out std_logic;
    TC_FN_C  : out std_logic
  );
end HE_SD_CNTR;


architecture RTL of HE_SD_CNTR is

  signal R_ADDR   : std_logic_vector(13 downto 0);
  signal C_ADDR   : std_logic_vector(9 downto 0);
  signal B_ADDR   : std_logic_vector(1 downto 0);

  signal ROW_CE   : std_logic;
  signal COL_CE   : std_logic;
--signal PAGE_CE  : std_logic;
  signal BANK_CE  : std_logic;

  signal C_TC_TOP : std_logic;
  signal R_TC_TOP : std_logic;

begin

  ROW_ADDR <= R_ADDR;
  COL_ADDR <= C_ADDR;
  BANK     <= B_ADDR;

  BANK_CE <= ROW_CE   AND R_TC_TOP AND R_ADDR(1) AND R_ADDR(0);
  ROW_CE  <= COUNT_EN AND C_TC_TOP AND C_ADDR(1) AND C_ADDR(0);
  COL_CE  <= COUNT_EN;

  -- DDP SDRAM version
--process(RESET, SDCLK)
--begin
--  if RESET='1' then
--    PAGE_CE <= '0';
--  elsif rising_edge(SDCLK) then
--    PAGE_CE <= '0';
--    if ROW_CE='1' and R_ADDR(12 downto 0)="1111111111111" then
--      PAGE_CE <= ROW_CE;
--    end if;
--  end if;
--end process;

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      B_ADDR <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if BANK_CE='1' or ADDR_EN='1' then
        if ADDR_EN='1' then
          B_ADDR <= ADDR_IN(25 downto 24);
        else
          B_ADDR <= B_ADDR + 1;
        end if;
      end if;
    end if;
  end process;

  process(RESET, SDCLK)
  begin
    if RESET='1' then
  -- DDP SDRAM version
--    R_ADDR(12 downto 0) <= (others=>'0');
      R_ADDR <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if ROW_CE='1' or ADDR_EN='1' then
        if ADDR_EN='1' then
  -- DDP SDRAM version
--        R_ADDR(12 downto 0) <= ADDR_IN(22 downto 10);
          R_ADDR <= ADDR_IN(23 downto 10);
        else
  -- DDP SDRAM version
--        R_ADDR(12 downto 0) <= R_ADDR(12 downto 0) + 1;
          R_ADDR <= R_ADDR + 1;
        end if;
      end if;
    end if;
  end process;

  -- DDP SDRAM version
--process(RESET, SDCLK)
--begin
--  if RESET='1' then
--    R_ADDR(13) <= '0';
--  elsif rising_edge(SDCLK) then
--    if PAGE_CE='1' or ADDR_EN='1' then
--      if ADDR_EN='1' then
--        R_ADDR(13) <= ADDR_IN(23);
--      else
--        R_ADDR(13) <= NOT R_ADDR(13);
--      end if;
--    end if;
--  end if;
--end process;

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      C_ADDR <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if COL_CE='1' or ADDR_EN='1' then
        if ADDR_EN='1' then
          C_ADDR <= ADDR_IN(9 downto 0);
        else 
          C_ADDR <= C_ADDR + 1;
        end if;
      end if;
    end if;
  end process;

  TC_FN_A <= C_ADDR(0);
  TC_FN_B <= C_ADDR(1);
  TC_FN_C <= C_TC_TOP;

  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      C_TC_TOP <= '0';
      if C_ADDR(9 downto 2)="11111111" then
        C_TC_TOP <= '1';
      end if;
    end if;
  end process;

  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      R_TC_TOP <= '0';
      if R_ADDR(13 downto 2)="111111111111" then
        R_TC_TOP <= '1';
      end if;
    end if;
  end process;

end RTL;
