-- ==============================================================
-- FILE : TOP.VHD - may be modified by the ADVANCED USER
-- ==============================================================
--
-- This file is the top level of your project when using your
-- HERON-FPGA7 module.
--
-- This file could be modified by ADVANCED users.
--
-- ==============================================================
--
--  Module      : TOP
--  Date        : 10/09/2003
--  Author      : R. Williams - HUNT ENGINEERING
--  Description : TOP level
--
-- ==============================================================
--
--  Ver     Modified By      Date      Changes
--  ---     -----------      ----      -------
--  1.0     R. Williams    10-09-03    First Written
--  1.2     R. Williams    08-06-06    Removed HEPC8 support.
--                                     Minor fix to HE_RD_6F logic.
--  1.3     R. Williams    16-10-07    Added address enable to
--                                     user message interface to
--                                     improve response to bursts.
--
-- ==============================================================


library ieee;
  use ieee.std_logic_1164.ALL;
  use work.CONFIG.all;

-- synopsys translate_off
library UNISIM;
  use UNISIM.vcomponents.all;
-- synopsys translate_on

entity TOP is
  port (
    -- PADS - HE_WCLK
    doclk         : in  std_logic;                       -- FIFO_OUT clock (use with DLL)
    f0            : out std_logic;                       -- FIFO_OUT clock loop driver
    -- PADS - HE_WR_6F
    dof_ff        : in  std_logic_vector(5 downto 0);    -- FIFO_OUT full flag
    dof_af        : in  std_logic_vector(5 downto 0);    -- FIFO_OUT almost full flag
    dof_wen       : out std_logic_vector(5 downto 0);    -- FIFO_OUT write enable
    do            : out std_logic_vector(31 downto 0);   -- FIFO_OUT data
    -- PADS - HE_RCLK
    diclk         : in  std_logic;                       -- FIFO_IN clock (use with DLL)
    f1            : out std_logic;                       -- FIFO_IN clock loop driver
    -- PADS - HE_RD_6F
    dif_ef        : in  std_logic_vector(5 downto 0);    -- FIFO_IN empty flag
    dif_ae        : in  std_logic_vector(5 downto 0);    -- FIFO_IN almost empty flag
    dif_ren       : out std_logic_vector(5 downto 0);    -- FIFO_IN read enable
    dif_oen       : out std_logic_vector(5 downto 0);    -- FIFO_IN output enable
    di            : in  std_logic_vector(31 downto 0);   -- FIFO_IN data
    -- PADS - HE_SDRAM
    sdram_dq      : inout std_logic_vector(31 downto 0); -- SDRAM Data
    sdram_a       : out std_logic_vector(12 downto 0);   -- SDRAM Address
    sdram_ba      : out std_logic_vector(1 downto 0);    -- SDRAM Bank Address
    sdram_cs      : out std_logic_vector(1 downto 0);    -- SDRAM Chip Select
    sdram_cke     : out std_logic_vector(1 downto 0);    -- SDRAM Clock Enable
    sdram_clk_src : out std_logic_vector(1 downto 0);    -- SDRAM Clock Source
    sdram_dqm     : out std_logic;                       -- SDRAM Data mask
    sdram_ras     : out std_logic;                       -- SDRAM Row address strobe
    sdram_cas     : out std_logic;                       -- SDRAM Col address strobe
    sdram_we      : out std_logic;                       -- SDRAM Write Enable
    sdram_clk     : in  std_logic;                       -- SDRAM Clock input
    -- PADS - HSB
    vpd           : inout std_logic_vector(7 downto 0);  -- HSB data
    vinit         : inout std_logic;                     -- HSB address strobe
    vcs           : inout std_logic;                     -- HSB data strobe
    vwrite        : inout std_logic;                     -- HSB read(high)/write(low)
    vbusy         : inout std_logic;                     -- HSB ready
    -- PADS - Connector Digital I/O
    dio_a         : inout std_logic_vector(14 downto 0); -- Connector A IO
    dio_b         : inout std_logic_vector(14 downto 0); -- Connector B IO
    dio_c         : inout std_logic_vector(14 downto 0); -- Connector C IO
    -- CONTROL signals
    config        : in  std_logic;                       -- config from carrier card
    reset_pin     : in  std_logic;                       -- reset from carrier card
    udpres        : out std_logic;                       -- reset the carrier (high if NOT used)
    -- OSC/CLOCKS
    osc0_pin      : in  std_logic;                       -- user oscillator from socketed Xtal Osc
    osc3_pin      : in  std_logic;                       -- user oscillator from surface mount Xtal Osc
    -- MISC I/O
    led           : out std_logic_vector(4 downto 0);    -- LEDs
    cid           : in  std_logic_vector(3 downto 0);    -- Carrier ID
    mid           : in  std_logic_vector(3 downto 0);    -- Module ID
    umi           : inout std_logic_vector(3 downto 0);  -- Uncommitted UMI interconnects
    addr_flagsel  : in  std_logic;                       -- Carrier Addr/Flagsel signal
    booten        : out std_logic;                       -- Carrier Boot Enable signal
    -- SERIAL I/O
    r1out_a       : in  std_logic;                       -- input serial data (connected to MAX3160)
    r2out_a       : in  std_logic;                       -- input serial data (connected to MAX3160)
    t1in_a        : out std_logic;                       -- output serial data (connected to MAX3160)
    t2in_a        : out std_logic;                       -- output serial data (connected to MAX3160)
    rs485_232_a   : out std_logic;                       -- control MAX3160
    hdplx_a       : out std_logic;                       -- control MAX3160
    fast_a        : out std_logic                        -- control MAX3160
  );
end TOP;


architecture RTL of TOP is

  component IBUF
    port(I : in  std_logic;
         O : out std_logic);
  end component;

  component IBUFG
    port(I : in  std_logic;
         O : out std_logic);
  end component;

  component IOBUF
    port(T  : in  std_logic;
         I  : in  std_logic;
         O  : out std_logic;
         IO : inout std_logic);
  end component;

  component IOBUF_F_8
    port(T  : in  std_logic;
         I  : in  std_logic;
         O  : out std_logic;
         IO : inout std_logic);
  end component;

  component OBUF_F_24
    port(I : in  std_logic;
         O : out std_logic);
  end component;

  component OBUF_F_8
    port(I : in  std_logic;
         O : out std_logic);
  end component;

  component IBUFDS
    port(
      I  : in  std_logic;
      IB : in  std_logic;
      O  : out std_logic );
  end component;

  component OBUFDS
    port(
      I  : in  std_logic;
      O  : out std_logic;
      OB : out std_logic );
  end component;

  component FD
    port(
      Q   : out std_logic;
      D   : in  std_logic;
      C   : in  std_logic );
  end component;

  component HE_RWCLK
    generic ( HFreq : boolean := FALSE );
    port (
      -- PADS
      DCLK : in  std_logic;
      FCLK : out std_logic;
      -- I/O module
      RST  : in  std_logic;
      CLK  : in  std_logic;
      GCLK : out std_logic
    );
  end component;

  component HE_RD_6F
    port (
      -- PADS
      DIF_EF          : in  std_logic_vector(5 downto 0);
      DIF_AE          : in  std_logic_vector(5 downto 0);
      DIF_REN         : out std_logic_vector(5 downto 0);
      DIF_OEN         : out std_logic_vector(5 downto 0);
      DI              : in  std_logic_vector(31 downto 0);
      -- I/O module
      RST             : in  std_logic;
      CLK             : in  std_logic;
      INFIFO_READ_REQ : in  std_logic_vector(5 downto 0);
      INFIFO_DVALID   : out std_logic_vector(5 downto 0);
      INFIFO_SINGLE   : out std_logic_vector(5 downto 0);
      INFIFO_BURST    : out std_logic_vector(5 downto 0);
      INFIFO0_D       : out std_logic_vector(31 downto 0);
      INFIFO1_D       : out std_logic_vector(31 downto 0);
      INFIFO2_D       : out std_logic_vector(31 downto 0);
      INFIFO3_D       : out std_logic_vector(31 downto 0);
      INFIFO4_D       : out std_logic_vector(31 downto 0);
      INFIFO5_D       : out std_logic_vector(31 downto 0)
    );
  end component;

  component HE_WR_6F
    port (
      -- PADS
      DOF_FF        : in  std_logic_vector(5 downto 0);
      DOF_AF        : in  std_logic_vector(5 downto 0);
      DOF_WEN       : out std_logic_vector(5 downto 0);
      DO            : out std_logic_vector(31 downto 0);
      -- I/O module
      RST           : in  std_logic;
      CLK           : in  std_logic;
      OUTFIFO_WRITE : in  std_logic_vector(5 downto 0);
      OUTFIFO_D     : in  std_logic_vector(31 downto 0);
      OUTFIFO_READY : out std_logic_vector(5 downto 0)
    );
  end component;

  component HE_SDRAM
    port (
      -- PADS
      SDRAM_A        : out std_logic_vector(12 downto 0);
      SDRAM_BA       : out std_logic_vector(1 downto 0);
      SDRAM_CS       : out std_logic_vector(1 downto 0);
      SDRAM_CKE      : out std_logic_vector(1 downto 0);
      SDRAM_DQM      : out std_logic;
      SDRAM_RAS      : out std_logic;
      SDRAM_CAS      : out std_logic;
      SDRAM_WE       : out std_logic;
      -- I/O module
      OSC            : in  std_logic;
      xSDCLK         : in  std_logic;
      CLKFX_133MHZ   : out std_logic;
      SD_OEN         : out std_logic;
      SD_DIN         : in  std_logic_vector(31 downto 0);
      SD_DOUT        : out std_logic_vector(31 downto 0);
      -- SDRAM Clock
      X_SDCLK        : out  std_logic;
      -- SDRAM Write Port
      SDRAM_WR_DATA  : in  std_logic_vector(31 downto 0);
      SDRAM_WR_ADDR  : in  std_logic_vector(25 downto 0);
      SDRAM_WR_AEN   : in  std_logic;
      SDRAM_WR_READY : in  std_logic;
      SDRAM_WR_BURST : in  std_logic;
      SDRAM_WR_DONE  : out std_logic;
      -- SDRAM Read Port
      SDRAM_RD_DATA  : out std_logic_vector(31 downto 0);
      SDRAM_RD_ADDR  : in  std_logic_vector(25 downto 0);
      SDRAM_RD_AEN   : in  std_logic;
      SDRAM_RD_READY : in  std_logic;
      SDRAM_RD_BURST : in  std_logic;
      SDRAM_RD_DONE  : out std_logic;
      SDRAM_RD_DVAL  : out std_logic
    );
  end component;

  component HE_USER
    port (
      -- PADS - HE_V/S_CTL
      VPD       : inout std_logic_vector(7 downto 0);
      VINIT     : inout std_logic;
      VCS       : inout std_logic;
      VWRITE    : inout std_logic;
      VBUSY     : inout std_logic;
      -- I/O module
      RST       : in  std_logic;
      CLK       : in  std_logic;
      READY     : in  std_logic;
      SEND_MSG  : in  std_logic;
      CE        : in  std_logic;
      DOUT      : in  std_logic_vector(7 downto 0);
      CID       : in  std_logic_vector(3 downto 0);
      MID       : in  std_logic_vector(2 downto 0);
      SEND_ID   : in  std_logic;
      LAST_BYTE : in  std_logic;
      DIN       : out std_logic_vector(7 downto 0);
      ADDR      : out std_logic_vector(7 downto 0);
      AEN       : out std_logic;
      WEN       : out std_logic;
      REN       : out std_logic;
      DONE      : out std_logic;
      COUNT     : out std_logic;
      CLEAR     : out std_logic
    );
  end component;

  component USER_AP
    port (
      RESET           : in  std_logic;
      CONFIG          : in  std_logic;
  -- CLOCK SIGNALS
      OSC0            : in  std_logic;
      OSC3            : in  std_logic;
  -- SERIAL IO
      R1OUT_A         : in  std_logic;
      R2OUT_A         : in  std_logic;
      T1IN_A          : out std_logic;
      T2IN_A          : out std_logic;
      RS485_232_A     : out std_logic;
      HDPLX_A         : out std_logic;
      FAST_A          : out std_logic;
  -- GENERAL IO
      UDPRES          : out std_logic;
      LED             : out std_logic_vector(4 downto 0);
      ADDR_FLAGSEL    : in  std_logic;
      BOOTEN          : out std_logic;
      CID             : in  std_logic_vector(3 downto 0);
      MID             : in  std_logic_vector(3 downto 0);
  -- UMIs
      UMI_IN          : in  std_logic_vector(3 downto 0);
      UMI_OUT         : out std_logic_vector(3 downto 0);
      UMI_EN          : out std_logic_vector(3 downto 0);
  -- READ FIFO CLK
      FCLK_RD         : in  std_logic;
      SRC_FCLK_RD     : out std_logic;
  -- WRITE FIFO CLK
      FCLK_WR         : in  std_logic;
      SRC_FCLK_WR     : out std_logic;
  -- GLOBAL FIFO CLK
      FCLK_G          : in  std_logic;
      SRC_FCLK_G      : out std_logic;
  -- FIFO IN / HE_RD_6F interface
      INFIFO_READ_REQ : out std_logic_vector(5 downto 0);
      INFIFO_DVALID   : in  std_logic_vector(5 downto 0);
      INFIFO_SINGLE   : in  std_logic_vector(5 downto 0);
      INFIFO_BURST    : in  std_logic_vector(5 downto 0);
      INFIFO0_D       : in  std_logic_vector(31 downto 0);
      INFIFO1_D       : in  std_logic_vector(31 downto 0);
      INFIFO2_D       : in  std_logic_vector(31 downto 0);
      INFIFO3_D       : in  std_logic_vector(31 downto 0);
      INFIFO4_D       : in  std_logic_vector(31 downto 0);
      INFIFO5_D       : in  std_logic_vector(31 downto 0);
  -- FIFO OUT / HE_WR_6F interface
      OUTFIFO_READY   : in  std_logic_vector(5 downto 0);
      OUTFIFO_WRITE   : out std_logic_vector(5 downto 0);
      OUTFIFO_D       : out std_logic_vector(31 downto 0);
  -- SERIAL BUS / HE_USER interface
      MSG_CLK         : out std_logic;
      MSG_DIN         : in  std_logic_vector(7 downto 0);
      MSG_ADDR        : in  std_logic_vector(7 downto 0);
      MSG_AEN         : in  std_logic;
      MSG_WEN         : in  std_logic;
      MSG_REN         : in  std_logic;
      MSG_DONE        : in  std_logic;
      MSG_COUNT       : in  std_logic;
      MSG_CLEAR       : in  std_logic;
      MSG_READY       : out std_logic;
      MSG_SEND_MSG    : out std_logic;
      MSG_CE          : out std_logic;
      MSG_DOUT        : out std_logic_vector(7 downto 0);
      MSG_SEND_ID     : out std_logic;
      MSG_LAST_BYTE   : out std_logic;
  -- CONNECTOR DIGITAL IO
      CONN_A_IN       : in  std_logic_vector(14 downto 0);
      CONN_A_OUT      : out std_logic_vector(14 downto 0);
      CONN_A_EN       : out std_logic_vector(14 downto 0);
      CONN_B_IN       : in  std_logic_vector(14 downto 0);
      CONN_B_OUT      : out std_logic_vector(14 downto 0);
      CONN_B_EN       : out std_logic_vector(14 downto 0);
      CONN_C_IN       : in  std_logic_vector(14 downto 0);
      CONN_C_OUT      : out std_logic_vector(14 downto 0);
      CONN_C_EN       : out std_logic_vector(14 downto 0);
  -- SDRAM Clock
      SDCLK           : in  std_logic;
  -- SDRAM Write Port
      SDRAM_WR_DATA   : out std_logic_vector(31 downto 0);
      SDRAM_WR_ADDR   : out std_logic_vector(25 downto 0);
      SDRAM_WR_AEN    : out std_logic;
      SDRAM_WR_READY  : out std_logic;
      SDRAM_WR_BURST  : out std_logic;
      SDRAM_WR_DONE   : in  std_logic;
  -- SDRAM Read Port
      SDRAM_RD_DATA   : in  std_logic_vector(31 downto 0);
      SDRAM_RD_ADDR   : out std_logic_vector(25 downto 0);
      SDRAM_RD_AEN    : out std_logic;
      SDRAM_RD_READY  : out std_logic;
      SDRAM_RD_BURST  : out std_logic;
      SDRAM_RD_DONE   : in  std_logic;
      SDRAM_RD_DVAL   : in  std_logic
    );
  end component;

  signal RESET           : std_logic;

  signal OSC0            : std_logic;
  signal OSC3            : std_logic;

  signal FCLK_RD         : std_logic;
  signal SRC_FCLK_RD     : std_logic;
  signal FCLK_WR         : std_logic;
  signal SRC_FCLK_WR     : std_logic;
  signal FCLK_G          : std_logic;
  signal xFCLK_G         : std_logic;
  signal SRC_FCLK_G      : std_logic;
  signal xDICLK          : std_logic;
  signal xF1             : std_logic;
  signal xDOCLK          : std_logic;
  signal xF0             : std_logic;

  signal xDIF_REN        : std_logic_vector(5 downto 0);
  signal xDIF_OEN        : std_logic_vector(5 downto 0);
  signal INFIFO_READ_REQ : std_logic_vector(5 downto 0);
  signal INFIFO_DVALID   : std_logic_vector(5 downto 0);
  signal INFIFO_SINGLE   : std_logic_vector(5 downto 0);
  signal INFIFO_BURST    : std_logic_vector(5 downto 0);
  signal INFIFO0_D       : std_logic_vector(31 downto 0);
  signal INFIFO1_D       : std_logic_vector(31 downto 0);
  signal INFIFO2_D       : std_logic_vector(31 downto 0);
  signal INFIFO3_D       : std_logic_vector(31 downto 0);
  signal INFIFO4_D       : std_logic_vector(31 downto 0);
  signal INFIFO5_D       : std_logic_vector(31 downto 0);

  signal xDOF_WEN        : std_logic_vector(5 downto 0);
  signal OUTFIFO_READY   : std_logic_vector(5 downto 0);
  signal OUTFIFO_WRITE   : std_logic_vector(5 downto 0);
  signal OUTFIFO_D       : std_logic_vector(31 downto 0);

  signal xSDCLK          : std_logic;
  signal CLKFX_133MHZ    : std_logic;
  signal sd_oen          : std_logic;
  signal sd_din          : std_logic_vector(31 downto 0);
  signal sd_dout         : std_logic_vector(31 downto 0);
  signal xSDRAM_A        : std_logic_vector(12 downto 0);
  signal xSDRAM_DQM      : std_logic;
  signal xSDRAM_WE       : std_logic;
  signal xSDRAM_CAS      : std_logic;
  signal xSDRAM_RAS      : std_logic;
  signal xSDRAM_CS       : std_logic_vector(1 downto 0);

  signal SDCLK           : std_logic;
  signal SDRAM_WR_DATA   : std_logic_vector(31 downto 0);
  signal SDRAM_WR_ADDR   : std_logic_vector(25 downto 0);
  signal SDRAM_WR_AEN    : std_logic;
  signal SDRAM_WR_READY  : std_logic;
  signal SDRAM_WR_BURST  : std_logic;
  signal SDRAM_WR_DONE   : std_logic;
  signal SDRAM_RD_DATA   : std_logic_vector(31 downto 0);
  signal SDRAM_RD_ADDR   : std_logic_vector(25 downto 0);
  signal SDRAM_RD_AEN    : std_logic;
  signal SDRAM_RD_READY  : std_logic;
  signal SDRAM_RD_BURST  : std_logic;
  signal SDRAM_RD_DONE   : std_logic;
  signal SDRAM_RD_DVAL   : std_logic;

  signal MSG_CLK         : std_logic;
  signal MSG_READY       : std_logic;
  signal MSG_SEND_MSG    : std_logic;
  signal MSG_CE          : std_logic;
  signal MSG_DOUT        : std_logic_vector(7 downto 0);
  signal MSG_SEND_ID     : std_logic;
  signal MSG_LAST_BYTE   : std_logic;
  signal MSG_DIN         : std_logic_vector(7 downto 0);
  signal MSG_ADDR        : std_logic_vector(7 downto 0);
  signal MSG_AEN         : std_logic;
  signal MSG_WEN         : std_logic;
  signal MSG_REN         : std_logic;
  signal MSG_DONE        : std_logic;
  signal MSG_COUNT       : std_logic;
  signal MSG_CLEAR       : std_logic;

  signal UMI_IN          : std_logic_vector(3 downto 0);
  signal UMI_OUT         : std_logic_vector(3 downto 0);
  signal UMI_EN          : std_logic_vector(3 downto 0);

  signal CONN_A_IN       : std_logic_vector(14 downto 0);
  signal CONN_A_OUT      : std_logic_vector(14 downto 0);
  signal CONN_A_EN       : std_logic_vector(14 downto 0);

  signal CONN_B_IN       : std_logic_vector(14 downto 0);
  signal CONN_B_OUT      : std_logic_vector(14 downto 0);
  signal CONN_B_EN       : std_logic_vector(14 downto 0);

  signal CONN_C_IN       : std_logic_vector(14 downto 0);
  signal CONN_C_OUT      : std_logic_vector(14 downto 0);
  signal CONN_C_EN       : std_logic_vector(14 downto 0);

begin

  -- RESET active low
  RESET <= NOT reset_pin;

  -- UMIs
  iUMI0 : iobuf port map (T=>UMI_EN(0), I=>UMI_OUT(0), O=>UMI_IN(0), IO=>umi(0));
  iUMI1 : iobuf port map (T=>UMI_EN(1), I=>UMI_OUT(1), O=>UMI_IN(1), IO=>umi(1));
  iUMI2 : iobuf port map (T=>UMI_EN(2), I=>UMI_OUT(2), O=>UMI_IN(2), IO=>umi(2));
  iUMI3 : iobuf port map (T=>UMI_EN(3), I=>UMI_OUT(3), O=>UMI_IN(3), IO=>umi(3));

  -- Oscillator inputs
  iOSC0 : IBUF port map (I=>osc0_pin, O=>OSC0);
  iOSC3 : IBUF port map (I=>osc3_pin, O=>OSC3);

  -------------------------------------------------------------
  -- FIFO clocks

  iDICLK : IBUFG port map (I=>diclk, O=>xDICLK);
  oF1 : OBUF_F_24 port map (I=>xF1, O=>f1);

  iDOCLK : IBUFG port map (I=>doclk, O=>xDOCLK);
  oF0 : OBUF_F_24 port map (I=>xF0, O=>f0);

  -- same clock for both FIFO interfaces
  -- ===================================
  gFCKG : if (FCLK_G_DOMAIN) generate

    -- read clock
    -- ==========
    iGCLK : HE_RWCLK
      generic map (HFreq => HIGH_FCLK_G)
      port map (
        DCLK => xDICLK,
        FCLK => xF1,
        --
        RST  => RESET,
        CLK  => SRC_FCLK_G,
        GCLK => xFCLK_G );

    -- write clock
    -- ===========
    xF0 <= SRC_FCLK_G;

    -- USER_AP clocks
    FCLK_RD <= xFCLK_G;
    FCLK_WR <= xFCLK_G;
    FCLK_G  <= xFCLK_G;

  end generate gFCKG;

  -- separate clocks for RD/WR FIFOs
  -- ===============================
  gNFCKG : if (NOT FCLK_G_DOMAIN) generate

    -- read clock
    -- ==========
    iRDCLK : HE_RWCLK
      generic map (HFreq => HIGH_FCLK_RD)
      port map (
        DCLK => xDICLK,
        FCLK => xF1,
        --
        RST  => RESET,
        CLK  => SRC_FCLK_RD,
        GCLK => FCLK_RD );

    -- write clock
    -- ===========
    iWRCLK : HE_RWCLK
      generic map (HFreq => HIGH_FCLK_WR)
      port map (
        DCLK => xDOCLK,
        FCLK => xF0,
        --
        RST  => RESET,
        CLK  => SRC_FCLK_WR,
        GCLK => FCLK_WR );

    -- unused clock
    FCLK_G  <= '0';

  end generate gNFCKG;

  -------------------------------------------------------------
  -- FIFO input

  iRD : HE_RD_6F
    port map (
      DIF_EF          => dif_ef,
      DIF_AE          => dif_ae,
      DIF_REN         => xDIF_REN,
      DIF_OEN         => xDIF_OEN,
      DI              => di,
      --
      RST             => RESET,
      CLK             => FCLK_RD,
      INFIFO_READ_REQ => INFIFO_READ_REQ,
      INFIFO_DVALID   => INFIFO_DVALID,
      INFIFO_SINGLE   => INFIFO_SINGLE,
      INFIFO_BURST    => INFIFO_BURST,
      INFIFO0_D       => INFIFO0_D,
      INFIFO1_D       => INFIFO1_D,
      INFIFO2_D       => INFIFO2_D,
      INFIFO3_D       => INFIFO3_D,
      INFIFO4_D       => INFIFO4_D,
      INFIFO5_D       => INFIFO5_D );

  gREN : for I in 0 to 5 generate
    oREN : OBUF_F_8 port map (I=>xDIF_REN(I), O=>dif_ren(I));
    oOEN : OBUF_F_8 port map (I=>xDIF_OEN(I), O=>dif_oen(I));
  end generate gREN;

  -------------------------------------------------------------
  -- FIFO output

  iWR : HE_WR_6F
    port map (
      DOF_FF        => dof_ff,
      DOF_AF        => dof_af,
      DOF_WEN       => xDOF_WEN,
      DO            => do,
      --
      RST           => RESET,
      CLK           => FCLK_WR,
      OUTFIFO_WRITE => OUTFIFO_WRITE,
      OUTFIFO_D     => OUTFIFO_D,
      OUTFIFO_READY => OUTFIFO_READY );

  gWEN : for I in 0 to 5 generate
    oWEN : OBUF_F_8 port map (I=>xDOF_WEN(I), O=>dof_wen(I));
  end generate gWEN;

  -------------------------------------------------------------
  -- SDRAM clock

  iSCLKI : IBUFG port map (I=>sdram_clk, O=>xSDCLK);

  oSCLK0 : OBUF_F_24 port map (I=>CLKFX_133MHZ, O=>sdram_clk_src(0));
  oSCLK1 : OBUF_F_24 port map (I=>CLKFX_133MHZ, O=>sdram_clk_src(1));

  -------------------------------------------------------------
  -- SDRAM

  iSDRAM : HE_SDRAM
    port map (
      SDRAM_A        => xSDRAM_A,
      SDRAM_BA       => sdram_ba,
      SDRAM_CS       => xSDRAM_CS,
      SDRAM_CKE      => sdram_cke,
      SDRAM_DQM      => xSDRAM_DQM,
      SDRAM_RAS      => xSDRAM_RAS,
      SDRAM_CAS      => xSDRAM_CAS,
      SDRAM_WE       => xSDRAM_WE,
      --
      OSC            => OSC3,
      xSDCLK         => xSDCLK,
      CLKFX_133MHZ   => CLKFX_133MHZ,
      SD_OEN         => sd_oen,
      SD_DIN         => sd_din,
      SD_DOUT        => sd_dout,
      --
      X_SDCLK        => SDCLK,
      SDRAM_WR_DATA  => SDRAM_WR_DATA,
      SDRAM_WR_ADDR  => SDRAM_WR_ADDR,
      SDRAM_WR_AEN   => SDRAM_WR_AEN,
      SDRAM_WR_READY => SDRAM_WR_READY,
      SDRAM_WR_BURST => SDRAM_WR_BURST,
      SDRAM_WR_DONE  => SDRAM_WR_DONE,
      SDRAM_RD_DATA  => SDRAM_RD_DATA,
      SDRAM_RD_ADDR  => SDRAM_RD_ADDR,
      SDRAM_RD_AEN   => SDRAM_RD_AEN,
      SDRAM_RD_READY => SDRAM_RD_READY,
      SDRAM_RD_BURST => SDRAM_RD_BURST,
      SDRAM_RD_DONE  => SDRAM_RD_DONE,
      SDRAM_RD_DVAL  => SDRAM_RD_DVAL );

  oSD_C0 : OBUF_F_8 port map (I=>xSDRAM_WE,    O=>sdram_we);
  oSD_C1 : OBUF_F_8 port map (I=>xSDRAM_CAS,   O=>sdram_cas);
  oSD_C2 : OBUF_F_8 port map (I=>xSDRAM_RAS,   O=>sdram_ras);
  oSD_C3 : OBUF_F_8 port map (I=>xSDRAM_CS(0), O=>sdram_cs(0));
  oSD_C4 : OBUF_F_8 port map (I=>xSDRAM_CS(1), O=>sdram_cs(1));
  oSD_C5 : OBUF_F_8 port map (I=>xSDRAM_DQM,   O=>sdram_dqm);

  gSDADDR : for I in 0 to 12 generate
    iSD_A : OBUF_F_8 port map (I=>xSDRAM_A(I), O=>sdram_a(I));
  end generate gSDADDR;

  gSDDATA : for I in 0 to 31 generate
    iSD_D : IOBUF_F_8 port map (T=>sd_oen, I=>sd_dout(I), O=>sd_din(I), IO=>sdram_dq(I));
  end generate gSDDATA;

  -------------------------------------------------------------
  -- Serial Bus Interface

  iUSR : HE_USER
    port map (
      VPD       => vpd,
      VINIT     => vinit,
      VCS       => vcs,
      VWRITE    => vwrite,
      VBUSY     => vbusy,
      --
      RST       => RESET,
      CLK       => MSG_CLK,
      READY     => MSG_READY,
      SEND_MSG  => MSG_SEND_MSG,
      CE        => MSG_CE,
      DOUT      => MSG_DOUT,
      CID       => cid,
      MID       => mid(2 downto 0),
      SEND_ID   => MSG_SEND_ID,
      LAST_BYTE => MSG_LAST_BYTE,
      DIN       => MSG_DIN,
      ADDR      => MSG_ADDR,
      AEN       => MSG_AEN,
      WEN       => MSG_WEN,
      REN       => MSG_REN,
      DONE      => MSG_DONE,
      COUNT     => MSG_COUNT,
      CLEAR     => MSG_CLEAR );

  -------------------------------------------------------------
  -- Connector Digital IO

  iDIOA0  : iobuf port map (T=>CONN_A_EN(0),  I=>CONN_A_OUT(0),  O=>CONN_A_IN(0),  IO=>dio_a(0) );
  iDIOA1  : iobuf port map (T=>CONN_A_EN(1),  I=>CONN_A_OUT(1),  O=>CONN_A_IN(1),  IO=>dio_a(1) );
  iDIOA2  : iobuf port map (T=>CONN_A_EN(2),  I=>CONN_A_OUT(2),  O=>CONN_A_IN(2),  IO=>dio_a(2) );
  iDIOA3  : iobuf port map (T=>CONN_A_EN(3),  I=>CONN_A_OUT(3),  O=>CONN_A_IN(3),  IO=>dio_a(3) );
  iDIOA4  : iobuf port map (T=>CONN_A_EN(4),  I=>CONN_A_OUT(4),  O=>CONN_A_IN(4),  IO=>dio_a(4) );
  iDIOA5  : iobuf port map (T=>CONN_A_EN(5),  I=>CONN_A_OUT(5),  O=>CONN_A_IN(5),  IO=>dio_a(5) );
  iDIOA6  : iobuf port map (T=>CONN_A_EN(6),  I=>CONN_A_OUT(6),  O=>CONN_A_IN(6),  IO=>dio_a(6) );
  iDIOA7  : iobuf port map (T=>CONN_A_EN(7),  I=>CONN_A_OUT(7),  O=>CONN_A_IN(7),  IO=>dio_a(7) );
  iDIOA8  : iobuf port map (T=>CONN_A_EN(8),  I=>CONN_A_OUT(8),  O=>CONN_A_IN(8),  IO=>dio_a(8) );
  iDIOA9  : iobuf port map (T=>CONN_A_EN(9),  I=>CONN_A_OUT(9),  O=>CONN_A_IN(9),  IO=>dio_a(9) );
  iDIOA10 : iobuf port map (T=>CONN_A_EN(10), I=>CONN_A_OUT(10), O=>CONN_A_IN(10), IO=>dio_a(10));
  iDIOA11 : iobuf port map (T=>CONN_A_EN(11), I=>CONN_A_OUT(11), O=>CONN_A_IN(11), IO=>dio_a(11));
  iDIOA12 : iobuf port map (T=>CONN_A_EN(12), I=>CONN_A_OUT(12), O=>CONN_A_IN(12), IO=>dio_a(12));
  iDIOA13 : iobuf port map (T=>CONN_A_EN(13), I=>CONN_A_OUT(13), O=>CONN_A_IN(13), IO=>dio_a(13));
  iDIOA14 : iobuf port map (T=>CONN_A_EN(14), I=>CONN_A_OUT(14), O=>CONN_A_IN(14), IO=>dio_a(14));

  iDIOB0  : iobuf port map (T=>CONN_B_EN(0),  I=>CONN_B_OUT(0),  O=>CONN_B_IN(0),  IO=>dio_b(0) );
  iDIOB1  : iobuf port map (T=>CONN_B_EN(1),  I=>CONN_B_OUT(1),  O=>CONN_B_IN(1),  IO=>dio_b(1) );
  iDIOB2  : iobuf port map (T=>CONN_B_EN(2),  I=>CONN_B_OUT(2),  O=>CONN_B_IN(2),  IO=>dio_b(2) );
  iDIOB3  : iobuf port map (T=>CONN_B_EN(3),  I=>CONN_B_OUT(3),  O=>CONN_B_IN(3),  IO=>dio_b(3) );
  iDIOB4  : iobuf port map (T=>CONN_B_EN(4),  I=>CONN_B_OUT(4),  O=>CONN_B_IN(4),  IO=>dio_b(4) );
  iDIOB5  : iobuf port map (T=>CONN_B_EN(5),  I=>CONN_B_OUT(5),  O=>CONN_B_IN(5),  IO=>dio_b(5) );
  iDIOB6  : iobuf port map (T=>CONN_B_EN(6),  I=>CONN_B_OUT(6),  O=>CONN_B_IN(6),  IO=>dio_b(6) );
  iDIOB7  : iobuf port map (T=>CONN_B_EN(7),  I=>CONN_B_OUT(7),  O=>CONN_B_IN(7),  IO=>dio_b(7) );
  iDIOB8  : iobuf port map (T=>CONN_B_EN(8),  I=>CONN_B_OUT(8),  O=>CONN_B_IN(8),  IO=>dio_b(8) );
  iDIOB9  : iobuf port map (T=>CONN_B_EN(9),  I=>CONN_B_OUT(9),  O=>CONN_B_IN(9),  IO=>dio_b(9) );
  iDIOB10 : iobuf port map (T=>CONN_B_EN(10), I=>CONN_B_OUT(10), O=>CONN_B_IN(10), IO=>dio_b(10));
  iDIOB11 : iobuf port map (T=>CONN_B_EN(11), I=>CONN_B_OUT(11), O=>CONN_B_IN(11), IO=>dio_b(11));
  iDIOB12 : iobuf port map (T=>CONN_B_EN(12), I=>CONN_B_OUT(12), O=>CONN_B_IN(12), IO=>dio_b(12));
  iDIOB13 : iobuf port map (T=>CONN_B_EN(13), I=>CONN_B_OUT(13), O=>CONN_B_IN(13), IO=>dio_b(13));
  iDIOB14 : iobuf port map (T=>CONN_B_EN(14), I=>CONN_B_OUT(14), O=>CONN_B_IN(14), IO=>dio_b(14));

  iDIOC0  : iobuf port map (T=>CONN_C_EN(0),  I=>CONN_C_OUT(0),  O=>CONN_C_IN(0),  IO=>dio_c(0) );
  iDIOC1  : iobuf port map (T=>CONN_C_EN(1),  I=>CONN_C_OUT(1),  O=>CONN_C_IN(1),  IO=>dio_c(1) );
  iDIOC2  : iobuf port map (T=>CONN_C_EN(2),  I=>CONN_C_OUT(2),  O=>CONN_C_IN(2),  IO=>dio_c(2) );
  iDIOC3  : iobuf port map (T=>CONN_C_EN(3),  I=>CONN_C_OUT(3),  O=>CONN_C_IN(3),  IO=>dio_c(3) );
  iDIOC4  : iobuf port map (T=>CONN_C_EN(4),  I=>CONN_C_OUT(4),  O=>CONN_C_IN(4),  IO=>dio_c(4) );
  iDIOC5  : iobuf port map (T=>CONN_C_EN(5),  I=>CONN_C_OUT(5),  O=>CONN_C_IN(5),  IO=>dio_c(5) );
  iDIOC6  : iobuf port map (T=>CONN_C_EN(6),  I=>CONN_C_OUT(6),  O=>CONN_C_IN(6),  IO=>dio_c(6) );
  iDIOC7  : iobuf port map (T=>CONN_C_EN(7),  I=>CONN_C_OUT(7),  O=>CONN_C_IN(7),  IO=>dio_c(7) );
  iDIOC8  : iobuf port map (T=>CONN_C_EN(8),  I=>CONN_C_OUT(8),  O=>CONN_C_IN(8),  IO=>dio_c(8) );
  iDIOC9  : iobuf port map (T=>CONN_C_EN(9),  I=>CONN_C_OUT(9),  O=>CONN_C_IN(9),  IO=>dio_c(9) );
  iDIOC10 : iobuf port map (T=>CONN_C_EN(10), I=>CONN_C_OUT(10), O=>CONN_C_IN(10), IO=>dio_c(10));
  iDIOC11 : iobuf port map (T=>CONN_C_EN(11), I=>CONN_C_OUT(11), O=>CONN_C_IN(11), IO=>dio_c(11));
  iDIOC12 : iobuf port map (T=>CONN_C_EN(12), I=>CONN_C_OUT(12), O=>CONN_C_IN(12), IO=>dio_c(12));
  iDIOC13 : iobuf port map (T=>CONN_C_EN(13), I=>CONN_C_OUT(13), O=>CONN_C_IN(13), IO=>dio_c(13));
  iDIOC14 : iobuf port map (T=>CONN_C_EN(14), I=>CONN_C_OUT(14), O=>CONN_C_IN(14), IO=>dio_c(14));

  -------------------------------------------------------------
  -- USER_AP

  iUSER_AP : USER_AP
    port map (
      RESET           => RESET,
      CONFIG          => config,
      --
      OSC0            => OSC0,
      OSC3            => OSC3,
      --
      UDPRES          => udpres,
      LED             => led,
      --
      ADDR_FLAGSEL    => addr_flagsel,
      BOOTEN          => booten,
      --
      CID             => cid,
      MID             => mid,
      UMI_IN          => UMI_IN,
      UMI_OUT         => UMI_OUT,
      UMI_EN          => UMI_EN,
      --
      R1OUT_A         => r1out_a,
      R2OUT_A         => r2out_a,
      T1IN_A          => t1in_a,
      T2IN_A          => t2in_a,
      RS485_232_A     => rs485_232_a,
      HDPLX_A         => hdplx_a,
      FAST_A          => fast_a,
      --
      FCLK_RD         => FCLK_RD,
      SRC_FCLK_RD     => SRC_FCLK_RD,
      --
      FCLK_WR         => FCLK_WR,
      SRC_FCLK_WR     => SRC_FCLK_WR,
      --
      FCLK_G          => FCLK_G,
      SRC_FCLK_G      => SRC_FCLK_G,
      --
      INFIFO_READ_REQ => INFIFO_READ_REQ,
      INFIFO_DVALID   => INFIFO_DVALID,
      INFIFO_SINGLE   => INFIFO_SINGLE,
      INFIFO_BURST    => INFIFO_BURST,
      INFIFO0_D       => INFIFO0_D,
      INFIFO1_D       => INFIFO1_D,
      INFIFO2_D       => INFIFO2_D,
      INFIFO3_D       => INFIFO3_D,
      INFIFO4_D       => INFIFO4_D,
      INFIFO5_D       => INFIFO5_D,
      --
      OUTFIFO_READY   => OUTFIFO_READY,
      OUTFIFO_WRITE   => OUTFIFO_WRITE,
      OUTFIFO_D       => OUTFIFO_D,
      --
      MSG_CLK         => MSG_CLK,
      MSG_DIN         => MSG_DIN,
      MSG_ADDR        => MSG_ADDR,
      MSG_AEN         => MSG_AEN,
      MSG_WEN         => MSG_WEN,
      MSG_REN         => MSG_REN,
      MSG_DONE        => MSG_DONE,
      MSG_COUNT       => MSG_COUNT,
      MSG_CLEAR       => MSG_CLEAR,
      MSG_READY       => MSG_READY,
      MSG_SEND_MSG    => MSG_SEND_MSG,
      MSG_CE          => MSG_CE,
      MSG_DOUT        => MSG_DOUT,
      MSG_SEND_ID     => MSG_SEND_ID,
      MSG_LAST_BYTE   => MSG_LAST_BYTE,
      --
      CONN_A_IN       => CONN_A_IN,
      CONN_A_OUT      => CONN_A_OUT,
      CONN_A_EN       => CONN_A_EN,
      CONN_B_IN       => CONN_B_IN,
      CONN_B_OUT      => CONN_B_OUT,
      CONN_B_EN       => CONN_B_EN,
      CONN_C_IN       => CONN_C_IN,
      CONN_C_OUT      => CONN_C_OUT,
      CONN_C_EN       => CONN_C_EN,
      --
      SDCLK           => SDCLK,
      SDRAM_WR_DATA   => SDRAM_WR_DATA,
      SDRAM_WR_ADDR   => SDRAM_WR_ADDR,
      SDRAM_WR_AEN    => SDRAM_WR_AEN,
      SDRAM_WR_READY  => SDRAM_WR_READY,
      SDRAM_WR_BURST  => SDRAM_WR_BURST,
      SDRAM_WR_DONE   => SDRAM_WR_DONE,
      SDRAM_RD_DATA   => SDRAM_RD_DATA,
      SDRAM_RD_ADDR   => SDRAM_RD_ADDR,
      SDRAM_RD_AEN    => SDRAM_RD_AEN,
      SDRAM_RD_READY  => SDRAM_RD_READY,
      SDRAM_RD_BURST  => SDRAM_RD_BURST,
      SDRAM_RD_DONE   => SDRAM_RD_DONE,
      SDRAM_RD_DVAL   => SDRAM_RD_DVAL );

end RTL;
