-- ==============================================================
-- FILE : USER_AP2.VHD  -  MUST be modified by the USER
-- ==============================================================
--
-- This file is part of Example 2, for the HERON-FPGA7 module.
--
-- This file must be modified by the user.
--
-- ==============================================================
--
--  Date        : 10/09/2003
--  Author      : R. Williams - HUNT ENGINEERING
--  Description : User code section
--
-- ==============================================================
--
--  Ver     Modified By      Date      Changes
--  ---     -----------      ----      -------
--  1.0     R. Williams    10-09-03    First Written
--  1.2     R. Williams    08-06-06    Removed HEPC8 support.
--                                     Minor fix to HE_RD_6F logic.
--
-- ==============================================================



-- ==============================================================
--   Package : CONFIG
-- ==============================================================
--
-- This package allow users to set up the different clock domains
-- of the application (FIFO interfaces) in the top level.
-- Please refer to the HUNT documentation for more details.
--
-- ==============================================================

package CONFIG is

  -- FIFO CLOCK
  -- ==========

--   -----------------------------------------------------------------
--  | FCLK_G_DOMAIN |  HIGH_FCLK_G  |  HIGH_FCLK_RD  |  HIGH_FCLK_WR  |
--   -----------------------------------------------------------------
--  |     True      | True / False  |     n.a.       |     n.a.       |  single FIFO clock
--   -----------------------------------------------------------------
--  |     False     |     n.a.      |  True / False  |  True / False  |  two FIFO clocks
--   -----------------------------------------------------------------

  -- To use a single common FIFO clock, set FCLK_G_DOMAIN = TRUE
  constant FCLK_G_DOMAIN : boolean := TRUE;

  -- When read/write FIFO clock frequency > 60 MHz, set HIGH_FCLK_G = TRUE
  -- This constant is used only when FCLK_G_DOMAIN = TRUE
  constant HIGH_FCLK_G   : boolean := TRUE;

  -- When read FIFO clock frequency > 60 MHz, set HIGH_FCLK_RD = TRUE
  -- This constant is used only when FCLK_G_DOMAIN = FALSE
  constant HIGH_FCLK_RD  : boolean := FALSE;

  -- When write FIFO clock frequency > 60 MHz, set HIGH_FCLK_WR = TRUE
  -- This constant is used only when FCLK_G_DOMAIN = FALSE
  constant HIGH_FCLK_WR  : boolean := FALSE;

end CONFIG;


-- ==============================================================
--   Module : USER_AP
-- ==============================================================

library IEEE;
  use IEEE.std_logic_1164.ALL;
  use IEEE.numeric_std.all;
  use IEEE.std_logic_unsigned.all;
  use WORK.CONFIG.all;


--pragma translate_off
library sim_lib;
  use sim_lib.all;

library UNISIM;
  use UNISIM.VCOMPONENTS.all;
--pragma translate_on

entity USER_AP is
  port (
    RESET           : in  std_logic;                     -- asynchronous reset active high
    CONFIG          : in  std_logic;                     -- System wide Config signal
-- CLOCK IO
    OSC0            : in  std_logic;                     -- Xtal OSC0
    OSC3            : in  std_logic;                     -- Xtal OSC3 factory default fitted with 100Mhz
-- SERIAL IO
    R1OUT_A         : in  std_logic;
    R2OUT_A         : in  std_logic;
    T1IN_A          : out std_logic;
    T2IN_A          : out std_logic;
    RS485_232_A     : out std_logic;
    HDPLX_A         : out std_logic;
    FAST_A          : out std_logic;
-- MISCELLANEOUS IO
    UDPRES          : out std_logic;                     -- reset the carrier (high if NOT used)
    LED             : out std_logic_vector(4 downto 0);  -- LEDs
    ADDR_FLAGSEL    : in  std_logic;
    BOOTEN          : out std_logic;
    CID             : in  std_logic_vector(3 downto 0);
    MID             : in  std_logic_vector(3 downto 0);
-- UMIs
    UMI_IN          : in  std_logic_vector(3 downto 0);
    UMI_OUT         : out std_logic_vector(3 downto 0);
    UMI_EN          : out std_logic_vector(3 downto 0);
-- READ FIFO CLK
    FCLK_RD         : in  std_logic;                     -- FIFO_IN clock to be used in this module (only when FCLK_G_DOMAIN=FALSE)
    SRC_FCLK_RD     : out std_logic;                     -- FIFO_IN clock source for the top level  (only when FCLK_G_DOMAIN=FALSE)
-- WRITE FIFO CLK
    FCLK_WR         : in  std_logic;                     -- FIFO_OUT clock to be used in this module (only when FCLK_G_DOMAIN=FALSE)
    SRC_FCLK_WR     : out std_logic;                     -- FIFO_OUT clock source for the top level  (only when FCLK_G_DOMAIN=FALSE)
-- COMMON FIFO CLK
    FCLK_G          : in  std_logic;                     -- Common FIFO clock to be used in this module (only when FCLK_G_DOMAIN=TRUE)
    SRC_FCLK_G      : out std_logic;                     -- Common FIFO clock source for the top level  (only when FCLK_G_DOMAIN=TRUE)
-- FIFO IN / HE_RD_6F interface
    INFIFO_READ_REQ : out std_logic_vector(5 downto 0);  -- FIFO_IN Read Request
    INFIFO_DVALID   : in  std_logic_vector(5 downto 0);  -- FIFO_IN Data Valid
    INFIFO_SINGLE   : in  std_logic_vector(5 downto 0);  -- FIFO_IN Single data item available
    INFIFO_BURST    : in  std_logic_vector(5 downto 0);  -- FIFO_IN Block of data available
    INFIFO0_D       : in  std_logic_vector(31 downto 0); -- FIFO_IN Data for FIFO 0
    INFIFO1_D       : in  std_logic_vector(31 downto 0); -- FIFO_IN Data for FIFO 1
    INFIFO2_D       : in  std_logic_vector(31 downto 0); -- FIFO_IN Data for FIFO 2
    INFIFO3_D       : in  std_logic_vector(31 downto 0); -- FIFO_IN Data for FIFO 3
    INFIFO4_D       : in  std_logic_vector(31 downto 0); -- FIFO_IN Data for FIFO 4
    INFIFO5_D       : in  std_logic_vector(31 downto 0); -- FIFO_IN Data for FIFO 5
-- FIFO OUT / HE_WR_6F interface
    OUTFIFO_READY   : in  std_logic_vector(5 downto 0);  -- FIFO_OUT Ready Flag
    OUTFIFO_WRITE   : out std_logic_vector(5 downto 0);  -- FIFO_OUT Write Command
    OUTFIFO_D       : out std_logic_vector(31 downto 0); -- FIFO_OUT Data
-- SERIAL BUS / HE_USER Message interface
    MSG_CLK         : out std_logic;                     -- HSB clock source for the top level
    MSG_DIN         : in  std_logic_vector(7 downto 0);  -- data received from HSB
    MSG_ADDR        : in  std_logic_vector(7 downto 0);  -- HSB address
    MSG_AEN         : in  std_logic;                     -- address enable
    MSG_WEN         : in  std_logic;                     -- write access
    MSG_REN         : in  std_logic;                     -- read access
    MSG_DONE        : in  std_logic;                     -- message successfully transmitted
    MSG_COUNT       : in  std_logic;                     -- user's counter enable
    MSG_CLEAR       : in  std_logic;                     -- user's counter clear
    MSG_READY       : out std_logic;                     -- to acknowledge an access
    MSG_SEND_MSG    : out std_logic;                     -- message send command
    MSG_CE          : out std_logic;                     -- to control speed operation
    MSG_DOUT        : out std_logic_vector(7 downto 0);  -- data sent to HSB
    MSG_SEND_ID     : out std_logic;                     -- ID information send command
    MSG_LAST_BYTE   : out std_logic;                     -- to indicate when the last byte is presented
  -- CONNECTOR DIGITAL IO
    CONN_A_IN       : in  std_logic_vector(14 downto 0); -- Data input on Connector A
    CONN_A_OUT      : out std_logic_vector(14 downto 0); -- Data output on Connector A
    CONN_A_EN       : out std_logic_vector(14 downto 0); -- Output enables for Connector A
    CONN_B_IN       : in  std_logic_vector(14 downto 0); -- Data input on Connector B
    CONN_B_OUT      : out std_logic_vector(14 downto 0); -- Data output on Connector B
    CONN_B_EN       : out std_logic_vector(14 downto 0); -- Output enables for Connector B
    CONN_C_IN       : in  std_logic_vector(14 downto 0); -- Data input on Connector C
    CONN_C_OUT      : out std_logic_vector(14 downto 0); -- Data output on Connector C
    CONN_C_EN       : out std_logic_vector(14 downto 0); -- Output enables for Connector C
  -- SDRAM Clock
    SDCLK           : in  std_logic;                     -- SDRAM Clock
  -- SDRAM Write Port
    SDRAM_WR_DATA   : out std_logic_vector(31 downto 0); -- SDRAM Write port data
    SDRAM_WR_ADDR   : out std_logic_vector(25 downto 0); -- SDRAM Write port address
    SDRAM_WR_AEN    : out std_logic;                     -- SDRAM Write port address enable
    SDRAM_WR_READY  : out std_logic;                     -- SDRAM Write port data ready
    SDRAM_WR_BURST  : out std_logic;                     -- SDRAM Write port burst ready
    SDRAM_WR_DONE   : in  std_logic;                     -- SDRAM Write port data transfer
  -- SDRAM Read Port
    SDRAM_RD_DATA   : in  std_logic_vector(31 downto 0); -- SDRAM Read port data
    SDRAM_RD_ADDR   : out std_logic_vector(25 downto 0); -- SDRAM Read port address
    SDRAM_RD_AEN    : out std_logic;                     -- SDRAM Read port address enable
    SDRAM_RD_READY  : out std_logic;                     -- SDRAM Read port data ready
    SDRAM_RD_BURST  : out std_logic;                     -- SDRAM Read port burst ready
    SDRAM_RD_DONE   : in  std_logic;                     -- SDRAM Read port data transfer
    SDRAM_RD_DVAL   : in  std_logic                      -- SDRAM Read port data valid
  );
end USER_AP;


architecture EXAMPLE2 of USER_AP is

  component HSB1
    port (
      RST         : in  std_logic;
      CLK         : in  std_logic;
      DIN         : in  std_logic_vector(7 downto 0);
      ADDR        : in  std_logic_vector(7 downto 0);
      AEN         : in  std_logic;
      WEN         : in  std_logic;
      REN         : in  std_logic;
      DONE        : in  std_logic;
      COUNT       : in  std_logic;
      CLEAR       : in  std_logic;
      READY       : out std_logic;
      SEND_MSG    : out std_logic;
      CE          : out std_logic;
      DOUT        : out std_logic_vector(7 downto 0);
      SEND_ID     : out std_logic;
      LAST_BYTE   : out std_logic;
      --
      INFIFO_SEL  : out std_logic_vector (5 downto 0);
      OUTFIFO_SEL : out std_logic_vector (5 downto 0)
    );
  end component;

  -- CoreGen Asynchronous FIFO
  component fifo15x32
    port (
      din          : in  std_logic_vector(31 downto 0);
      wr_en        : in  std_logic;
      wr_clk       : in  std_logic;
      rd_en        : in  std_logic;
      rd_clk       : in  std_logic;
      ainit        : in  std_logic;
      dout         : out std_logic_vector(31 downto 0);
      full         : out std_logic;
      empty        : out std_logic;
      almost_full  : out std_logic;
      almost_empty : out std_logic
    );
  end component;

  -- Synchronous FIFO
  component FIFO16x32
    port (
      RESET    : in  std_logic;
      CLOCK    : in  std_logic;
      DIN      : in  std_logic_vector(31 downto 0);
      WR_EN    : in  std_logic;
      RD_EN    : in  std_logic;
      DOUT     : out std_logic_vector(31 downto 0);
      FULL     : out std_logic;
      EMPTY    : out std_logic;
      WR_COUNT : out std_logic_vector(4 downto 0);
      RD_COUNT : out std_logic_vector(4 downto 0)
    );
  end component;

  component BUFG
    port (
      I : in  std_logic;
      O : out std_logic
    );
  end component;

  type state_type is (idle, write_mem, write_wait, read_mem, read_wait);

  signal state         : state_type;

  signal MCLK          : std_logic;
  signal HSB_CLK       : std_logic;
  signal HSB_CLOCK     : std_logic;
  signal LEDcount      : unsigned(24 downto 0);

  signal IN_FIFO_DATA  : std_logic_vector(31 downto 0);
  signal IN_FIFO_RD    : std_logic;
  signal IN_FIFO_EF    : std_logic;
  signal IN_FIFO_AF    : std_logic;

  signal OUT_FIFO_RD   : std_logic;
  signal OUT_FIFO_EF   : std_logic;
  signal OUT_FIFO_AF   : std_logic;

  signal NEW_COMMAND   : std_logic;
  signal DO_MEM_WRITE  : std_logic;
  signal DO_MEM_READ   : std_logic;

  signal IDLE_STATE    : std_logic;
  signal ADDR_EN       : std_logic;
  signal WRITE_EN      : std_logic;
  signal READ_EN       : std_logic;
  signal WRITE_DONE    : std_logic;
  signal READ_DONE     : std_logic;
  signal WR_DELAY      : std_logic;
  signal RD_DELAY      : std_logic;

  signal DATA_SEL      : std_logic_vector(3 downto 0);
  signal TEST_PATTERN  : std_logic_vector(31 downto 0);

  signal WR_FIFO_WR_CT : std_logic_vector(4 downto 0);
  signal WR_FIFO_RD_CT : std_logic_vector(4 downto 0);
  signal WR_FIFO_WR    : std_logic;
  signal WR_FIFO_EF    : std_logic;
  signal WR_FIFO_AE    : std_logic;
  signal WR_FIFO_AF    : std_logic;

  signal RD_FIFO_OUT   : std_logic_vector(31 downto 0);
  signal RD_FIFO_WR_CT : std_logic_vector(4 downto 0);
  signal RD_FIFO_RD    : std_logic;
  signal RD_FIFO_EF    : std_logic;
  signal RD_FIFO_FF    : std_logic;
  signal RD_FIFO_AF    : std_logic;

  signal WRITE_DATA    : std_logic_vector(31 downto 0);
  signal READ_DATA     : std_logic_vector(31 downto 0);
  signal WRITE_SHIFT   : std_logic;
  signal READ_SHIFT    : std_logic;
  signal WRITE_COUNT   : std_logic_vector(26 downto 0);
  signal READ_COUNT    : std_logic_vector(26 downto 0);
  signal WRITE_ZERO    : std_logic;
  signal READ_TOP_ZERO : std_logic;
  signal READ_ZERO     : std_logic;
  signal READ_ZERO_DEL : std_logic_vector(4 downto 0);
  signal READ_ONE      : std_logic;

  signal DATA_REG      : std_logic_vector(31 downto 0);
  signal DATA_COMPARE  : std_logic;
  signal DATA_CHECK    : std_logic;
  signal DATA_ERROR    : std_logic;

begin

  -- ==========================================================
  -- ==========================================================
  --       <<<<<<<   INSERT YOUR CODE HERE    >>>>>>>>>
  -- ==========================================================
  -- ==========================================================


  -------------------------------------------------------------
  -- OSC3 clock

  MCLK <= OSC3;

  -------------------------------------------------------------
  -- Drive the UDP-RESET high (inactive)

  UDPRES <= '1';

  -------------------------------------------------------------
  -- Drive the BOOT ENABLE signal high (inactive)

  BOOTEN <= '1';

  -------------------------------------------------------------
  -- Drive unused signals

  UMI_OUT <= (others=>'0');
  UMI_EN  <= (others=>'1');

  T1IN_A      <= '0';
  T2IN_A      <= '0';
  RS485_232_A <= '0';
  HDPLX_A     <= '0';
  FAST_A      <= '0';

  -------------------------------------------------------------
  -- FIFO Clocks

  SRC_FCLK_RD <= '0';   -- clock source for READ FIFO
  SRC_FCLK_WR <= '0';   -- clock source for WRITE FIFO
  SRC_FCLK_G  <= MCLK;  -- global clock source for both FIFOs

  -------------------------------------------------------------
  -- Test Control Data from HERON Input FIFO

  INFIFO_READ_REQ(0) <= (NOT IN_FIFO_AF);
  INFIFO_READ_REQ(1) <= '0';
  INFIFO_READ_REQ(2) <= '0';
  INFIFO_READ_REQ(3) <= '0';
  INFIFO_READ_REQ(4) <= '0';
  INFIFO_READ_REQ(5) <= '0';

  iIN_FIFO : fifo15x32
    port map (
      din          => INFIFO0_D,
      wr_en        => INFIFO_DVALID(0),
      wr_clk       => FCLK_G,
      rd_en        => IN_FIFO_RD,
      rd_clk       => SDCLK,
      ainit        => RESET,
      dout         => IN_FIFO_DATA,
      full         => open,
      empty        => IN_FIFO_EF,
      almost_full  => IN_FIFO_AF,
      almost_empty => open );

  IN_FIFO_RD <= IDLE_STATE AND (NOT IN_FIFO_EF) AND (NOT NEW_COMMAND) AND (NOT ADDR_EN);

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      NEW_COMMAND <= '0';
      ADDR_EN     <= '0';
    elsif rising_edge(SDCLK) then
      NEW_COMMAND <= IN_FIFO_RD;
      ADDR_EN     <= NEW_COMMAND;
    end if;
  end process;

  SDRAM_WR_AEN <= ADDR_EN;
  SDRAM_RD_AEN <= ADDR_EN;

  SDRAM_WR_ADDR <= (others=>'0');
  SDRAM_RD_ADDR <= (others=>'0');

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      DATA_SEL <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if NEW_COMMAND='1' then
        DATA_SEL <= IN_FIFO_DATA(3 downto 0);
      end if;
    end if;
  end process;

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      DO_MEM_WRITE <= '0';
      DO_MEM_READ  <= '0';
    elsif rising_edge(SDCLK) then
      DO_MEM_WRITE <= '0';
      DO_MEM_READ  <= '0';
      if (NEW_COMMAND='1' and IN_FIFO_DATA(4)='1') or (DO_MEM_WRITE='1' and WRITE_DONE='0') then
        DO_MEM_WRITE <= '1';
      end if;
      if (NEW_COMMAND='1' and IN_FIFO_DATA(5)='1') or (DO_MEM_READ='1' and READ_DONE='0') then
        DO_MEM_READ  <= '1';
      end if;
    end if;
  end process;

  -------------------------------------------------------------
  -- SDRAM Memory Read and Write Control State Machine

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      state <= idle;
    elsif rising_edge(SDCLK) then
      case state is
        when idle       => if    DO_MEM_WRITE='1'                   then state <= write_mem;
                           elsif DO_MEM_READ='1'                    then state <= read_mem;
                           else                                          state <= idle;
                           end if;
        when write_mem  =>                                               state <= write_wait;
        when write_wait => if    WRITE_DONE='1' and DO_MEM_READ='1' then state <= read_mem;
                           elsif WRITE_DONE='1'                     then state <= idle;
                           else                                          state <= write_wait;
                           end if;
        when read_mem   =>                                               state <= read_wait;
        when read_wait  => if    READ_DONE='1'                      then state <= idle;
                           else                                          state <= read_wait;
                           end if;
      end case;
    end if;
  end process;

  process(state)
  begin
    case state is
      when idle       => IDLE_STATE <= '1'; WRITE_EN <= '0'; READ_EN <= '0';
      when write_mem  => IDLE_STATE <= '0'; WRITE_EN <= '1'; READ_EN <= '0';
      when write_wait => IDLE_STATE <= '0'; WRITE_EN <= '0'; READ_EN <= '0';
      when read_mem   => IDLE_STATE <= '0'; WRITE_EN <= '0'; READ_EN <= '1';
      when read_wait  => IDLE_STATE <= '0'; WRITE_EN <= '0'; READ_EN <= '0';
    end case;
  end process;

  process(SDCLK)
  begin
    if rising_edge(SDCLK) then
      WR_DELAY <= WRITE_ZERO;
      RD_DELAY <= READ_ZERO_DEL(0);
    end if;
  end process;

  WRITE_DONE <= (NOT WR_DELAY) AND WRITE_ZERO;
  READ_DONE  <= (NOT RD_DELAY) AND READ_ZERO_DEL(0);

  -------------------------------------------------------------
  -- SDRAM Test Data Generation

  process (DATA_SEL(1 downto 0))
  begin
    case DATA_SEL(1 downto 0) is
      when "00"   => TEST_PATTERN <= "00000000000000000000000000000000";
      when "01"   => TEST_PATTERN <= "11111111111111111111111111111111";
      when "10"   => TEST_PATTERN <= "01010101010101010101010101010101";
      when "11"   => TEST_PATTERN <= "10101010101010101010101010101010";
      when others => TEST_PATTERN <= (others=>'0');
    end case;
  end process;

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      WRITE_DATA  <= (others=>'0');
      WRITE_SHIFT <= '0';
    elsif rising_edge(SDCLK) then
      if ADDR_EN='1' then
        WRITE_DATA <= TEST_PATTERN;
        WRITE_SHIFT <= '1';
      elsif WR_FIFO_WR='1' then
        if DATA_SEL(2)='1' then
          WRITE_DATA <= WRITE_DATA + 1;
        elsif DATA_SEL(3)='1' then
          WRITE_DATA <= WRITE_DATA(30 downto 0) & WRITE_SHIFT;
          WRITE_SHIFT <= WRITE_DATA(31);
        else
          WRITE_DATA <= WRITE_DATA;
        end if;
      end if;
    end if;
  end process;

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      READ_DATA  <= (others=>'0');
      READ_SHIFT <= '0';
    elsif rising_edge(SDCLK) then
      if ADDR_EN='1' then
        READ_DATA  <= TEST_PATTERN;
        READ_SHIFT <= '1';
      elsif DATA_CHECK='1' then
        if DATA_SEL(2)='1' then
          READ_DATA <= READ_DATA + 1;
        elsif DATA_SEL(3)='1' then
          READ_DATA <= READ_DATA(30 downto 0) & READ_SHIFT;
          READ_SHIFT <= READ_DATA(31);
        else
          READ_DATA <= READ_DATA;
        end if;
      end if;
    end if;
  end process;

  -------------------------------------------------------------
  -- SDRAM Write Data Count (to write all 256 Mbytes of memory)

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      WRITE_COUNT <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if WRITE_EN='1' then
        WRITE_COUNT <= "100000000000000000000000000";
      elsif WR_FIFO_WR='1' and WRITE_ZERO='0' then
        WRITE_COUNT <= WRITE_COUNT - 1;
      end if;
    end if;
  end process;

  WRITE_ZERO <= '1' when WRITE_COUNT="000000000000000000000000000" else '0';

  -------------------------------------------------------------
  -- SDRAM Read Data Count (to read all 256 Mbytes of memory)

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      READ_COUNT <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if READ_EN='1' then
        READ_COUNT <= "100000000000000000000000000";
      elsif SDRAM_RD_DONE='1' and READ_ZERO='0' then
        READ_COUNT <= READ_COUNT - 1;
      end if;
    end if;
  end process;

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      READ_TOP_ZERO <= '1';
    elsif rising_edge(SDCLK) then
      READ_TOP_ZERO <= '0';
      if READ_COUNT(26 downto 3)="000000000000000000000000" then
        READ_TOP_ZERO <= '1';
      end if;
    end if;
  end process;

  READ_ZERO <= READ_TOP_ZERO AND (NOT READ_COUNT(2)) AND (NOT READ_COUNT(1)) AND (NOT READ_COUNT(0));
  READ_ONE  <= READ_TOP_ZERO AND (NOT READ_COUNT(2)) AND (NOT READ_COUNT(1)) AND (    READ_COUNT(0));

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      READ_ZERO_DEL <= (others=>'0');
    elsif rising_edge(SDCLK) then
      READ_ZERO_DEL <= READ_ZERO & READ_ZERO_DEL(4 downto 1);
    end if;
  end process;

  -------------------------------------------------------------
  -- SDRAM Write Interface Data FIFO

  WR_FIFO_WR <= WR_FIFO_AF AND (NOT WRITE_ZERO);

  WR_FIFO_AF <= '0' when WR_FIFO_WR_CT(4 downto 1)="0000" else '1';
  WR_FIFO_AE <= '0' when WR_FIFO_RD_CT(4 downto 1)="0000" else '1';

  iWR_FIFO : FIFO16x32
    port map (
      RESET    => RESET,
      CLOCK    => SDCLK,
      DIN      => WRITE_DATA,
      WR_EN    => WR_FIFO_WR,
      RD_EN    => SDRAM_WR_DONE,
      DOUT     => SDRAM_WR_DATA,
      FULL     => open,
      EMPTY    => WR_FIFO_EF,
      WR_COUNT => WR_FIFO_WR_CT,
      RD_COUNT => WR_FIFO_RD_CT );

  SDRAM_WR_READY <= WR_FIFO_EF;
  SDRAM_WR_BURST <= WR_FIFO_AE;

  -------------------------------------------------------------
  -- SDRAM Read Interface Data FIFO

  iRD_FIFO : FIFO16x32
    port map (
      RESET    => RESET,
      CLOCK    => SDCLK,
      DIN      => SDRAM_RD_DATA,
      WR_EN    => SDRAM_RD_DVAL,
      RD_EN    => RD_FIFO_RD,
      DOUT     => RD_FIFO_OUT,
      FULL     => open,
      EMPTY    => RD_FIFO_EF,
      WR_COUNT => RD_FIFO_WR_CT,
      RD_COUNT => open );

  SDRAM_RD_READY <= (NOT READ_ZERO) AND RD_FIFO_FF;
  SDRAM_RD_BURST <= (NOT READ_ONE)  AND RD_FIFO_AF;

  -- Indicate almost full when 7 words to write or less
  process(RESET, SDCLK)
  begin
    if RESET='1' then
      RD_FIFO_AF <= '1';
    elsif rising_edge(SDCLK) then
      RD_FIFO_AF <= '1';
      if (RD_FIFO_WR_CT(4 downto 3)="00") then
        RD_FIFO_AF <= '0';
      end if;
    end if;
  end process;

  -- Indicate full when 6 words to write or less
  process(RESET, SDCLK)
  begin
    if RESET='1' then
      RD_FIFO_FF <= '1';
    elsif rising_edge(SDCLK) then
      RD_FIFO_FF <= '1';
      if (RD_FIFO_WR_CT(4 downto 3)="00") and
        ((RD_FIFO_WR_CT(2)='0') or
         (RD_FIFO_WR_CT(1)='0') or
         (RD_FIFO_WR_CT(0)='0')) then
        RD_FIFO_FF <= '0';
      end if;
    end if;
  end process;

  -------------------------------------------------------------
  -- SDRAM Data Check

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      DATA_REG <= (others=>'0');
    elsif rising_edge(SDCLK) then
      if SDRAM_RD_DVAL='1' then
        DATA_REG <= SDRAM_RD_DATA;
      end if;
    end if;
  end process;

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      DATA_CHECK <= '0';
    elsif rising_edge(SDCLK) then
      DATA_CHECK <= SDRAM_RD_DVAL;
    end if;
  end process;

  DATA_COMPARE <= '1' when READ_DATA=DATA_REG else '0';

  process(RESET, SDCLK)
  begin
    if RESET='1' then
      DATA_ERROR <= '0';
    elsif rising_edge(SDCLK) then
      if DATA_COMPARE='0' and DATA_CHECK='1' then
        DATA_ERROR <= '1';
      end if;
    end if;
  end process;

  -------------------------------------------------------------
  -- Transfer SDRAM Data to HERON Output FIFO

  RD_FIFO_RD <= (NOT OUT_FIFO_AF) AND RD_FIFO_EF;

  iFIFO6 : fifo15x32
    port map (
      din          => RD_FIFO_OUT,
      wr_en        => RD_FIFO_RD,
      wr_clk       => SDCLK,
      rd_en        => OUT_FIFO_RD,
      rd_clk       => FCLK_G,
      ainit        => RESET,
      dout         => OUTFIFO_D,
      full         => open,
      empty        => OUT_FIFO_EF,
      almost_full  => OUT_FIFO_AF,
      almost_empty => open );

  OUT_FIFO_RD <= OUTFIFO_READY(0) AND (NOT OUT_FIFO_EF);

  process(RESET, FCLK_G)
  begin
    if RESET='1' then
      OUTFIFO_WRITE(0) <= '0';
    elsif rising_edge(FCLK_G) then
      OUTFIFO_WRITE(0) <= OUT_FIFO_RD;
    end if;
  end process;

  OUTFIFO_WRITE(1) <= '0';
  OUTFIFO_WRITE(2) <= '0';
  OUTFIFO_WRITE(3) <= '0';
  OUTFIFO_WRITE(4) <= '0';
  OUTFIFO_WRITE(5) <= '0';

  -------------------------------------------------------------
  -- Serial Bus

  iHSB_BUFG : BUFG port map (I=>HSB_CLK, O=>HSB_CLOCK);

  process (OSC3)
  begin
    if rising_edge(OSC3) then
      HSB_CLK <= NOT HSB_CLK;
    end if;
  end process;

  MSG_CLK <= HSB_CLOCK;  -- clock for Message Interface

  iHSB1 : HSB1
    port map (
      RST         => RESET,
      CLK         => HSB_CLOCK,
      DIN         => MSG_DIN,
      ADDR        => MSG_ADDR,
      AEN         => MSG_AEN,
      WEN         => MSG_WEN,
      REN         => MSG_REN,
      DONE        => MSG_DONE,
      COUNT       => MSG_COUNT,
      CLEAR       => MSG_CLEAR,
      READY       => MSG_READY,
      SEND_MSG    => MSG_SEND_MSG,
      CE          => MSG_CE,
      DOUT        => MSG_DOUT,
      SEND_ID     => MSG_SEND_ID,
      LAST_BYTE   => MSG_LAST_BYTE,
      INFIFO_SEL  => open,
      OUTFIFO_SEL => open );

  -------------------------------------------------------------
  -- Connector Digital IO (not used in this example)

  CONN_A_OUT <= (others=>'0');
  CONN_A_EN  <= (others=>'1');

  CONN_B_OUT <= (others=>'0');
  CONN_B_EN  <= (others=>'1');

  CONN_C_OUT <= (others=>'0');
  CONN_C_EN  <= (others=>'1');

  -------------------------------------------------------------
  -- Led Flasher : dividing F_CLKG clock by 2**24

  process(RESET, FCLK_G)
  begin
    if RESET='1' then
      LEDcount <= (others=>'0');
    elsif rising_edge(FCLK_G) then
      LEDcount <= LEDcount + 1;
    end if;
  end process;

  LED(4) <= LEDcount(24);

  LED(0) <= NOT (DATA_SEL(0) OR DATA_ERROR);
  LED(1) <= NOT (DATA_SEL(1) OR DATA_ERROR);
  LED(2) <= NOT (DATA_SEL(2) OR DATA_ERROR);
  LED(3) <= NOT (DATA_SEL(3) OR DATA_ERROR);

end EXAMPLE2;
